using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;
using UnityEngine.XR;
using Leap;

namespace JayJayWon
{
    public class BrowserAssist : MVRScript
    {
        public static BrowserUI mainBrowserUI { get; set; } =null;
        public static CallbackBrowserUI callbackBrowserUI { get; set; } = null;

        public static Vector3 mainBrowserVAMUIOffset = new Vector3();

        SuperController.OnAtomUIDRename atomRenameHandle;
        SuperController.OnAtomRemoved atomRemoveHandle;
        SuperController.OnAtomAdded atomAddedHandle;
        SuperController.OnSceneLoaded sceneLoadedHandle;

        private Canvas launchButtonCanvas;
        private Canvas topScreenLaunchButtonCanvas;

        private UIDynamicButton vamUILaunchButton;
        private UIDynamicButton vamTopMenuLaunchButton;

        private GameObject moreOptionsGameObject;

        private Button vamSceneBrowserButton;

        private bool isEnabled = false;
        public override void Init()
        {
            try
            {
                Globals.mvrScript = this;
                
                if (XRSettings.enabled) Globals.isVREnabled = true;
                else Globals.isVREnabled = false;

#if VAM_GT_1_21
                Globals.isFileCreateModifiedDateAvailable = true;
                Globals.isRealClothingAvailable = true;
#endif

                if (containingAtom.name == "CoreControl" && containingAtom.type == "SessionPluginManager")
                {                                       
                    if (baPluginCount < 2)
                    {
                        if (legacyMetaDataExists)
                        {
                            string message = "BrowserAssist has detected a previous installation containing legacy meta data. This version of BrowserAssist does not support upgrading from that legacy meta data format. To uprgade the meta data to a compatible format you will need to run BrowserAssist v1.11 first and then re-run this version of Browser Assist.";
                            JSONStorableString text = new JSONStorableString("text", message);
                            CreateTextField(text);
                            text.dynamicText.height = 500f;
                            SuperController.LogMessage(message);
                        }
                        else
                        {
                            RegisterVAMHandlers();
                            SetupPluginIntegrations();

                            Utils.CreatePluginDataFolder();
                            Globals.caseSensitiveFileSystem = Utils.IsCaseSensitiveFileSystem(Const.BAScriptDataFolder);

                            var settingsJC = Settings.GetSettingsJSON();
                            Settings.EarlyLoadSettings(settingsJC);

                            var lastSupportedResourceTypes = Settings.GetLastSupportedResourceTypes(settingsJC);

                            HashSet<int> firstScanResourceTypes = new  HashSet<int> (ResourceType.currentSupportedResourceTypes.Except(lastSupportedResourceTypes));

                            bool saveSettings = false;
                            if (firstScanResourceTypes.Count > 0) saveSettings = true;

                            DateTime scanTime = DateTime.Now;
                            VARPackageManifest.Init(scanTime, firstScanResourceTypes);
                            ResourceManifest.LocalResourcesInit(scanTime, firstScanResourceTypes);
                            ResourceManifest.CachePrimaryClothingItemPresetRVGEs();

                            if (FileManagerSecure.FileExists(Const.BAScriptDataFolder + "\\" + "BAOverride.cfg"))
                            {
                                var overrridesettingsJC = SuperController.singleton.LoadJSON(Const.BAScriptDataFolder + "\\" + "BAOverride.cfg").AsObject;
                                if (overrridesettingsJC != null)
                                {
                                    if (overrridesettingsJC["startupFailureOverride"].AsBool) Globals.overrideFailedStartupHalt = true;

                                }                                
                            }

                            if (Globals.validStartup)
                            {
                                if (Settings.restoreNameJSS.val != "")
                                {
                                    SuperController.LogMessage("Successful restore from BrowserAssist backup '" + Settings.restoreNameJSS.val+ "' was detected.");
                                    Settings.restoreNameJSS.val = "";
                                    saveSettings= true;
                                }                             

                                BAPluginUI.InitUI();
                                Settings.LoadSettings(settingsJC);

                                BAPluginUI.RePopulateResourceTypeSelector();
           

                                mainBrowserUI = new BrowserUI();
                                callbackBrowserUI = new CallbackBrowserUI();

                                CreateVAMUILaunchButton();
                                CreateTopScreenLaunchButton();

                                CopySymLinkBatScripts();
                                CheckSymLinkAccess();

                                DateTime initCompleteTime = DateTime.Now;

                                if (Globals.debugLoadTimingEnabled) SuperController.LogMessage("TOTAL PLUGIN STARTUP TIME TAKEN " + (initCompleteTime - scanTime).TotalSeconds.ToString("0.00") + " seconds");

                                if (saveSettings) Settings.SaveSettings();

                                isEnabled = true;
                            }
                            else
                            {
                                isEnabled = false;
                                if (Settings.restoreNameJSS.val != "")
                                {
                                    SuperController.LogMessage("A restore from BrowserAssist backup '" + Settings.restoreNameJSS.val + "' was detected. The restore failed due to a corruption detected in the backup. The backup will be marked as corrupt. Try restoring from a different restore point or contact JayJayWon on Discord if no valid backups are available.");
                                    MarkBackupAsCorrupt(Settings.restoreNameJSS.val);
                                    
                                }
                                else
                                {
                                    SuperController.LogMessage("An unexpected error occured during BrowserAssist startup relating to loading of the BA stored meta data (e.g. Tags etc). To preserve the integrity of that data, BrowserAssist will not operate until the issue is resolved. Previous backups can be restored from the Plugin UI to resolve this issue. If no valid backups are available then please report the issue to JayJayWon on Discord.");
                                }

                                BAPluginUI.InitUI();
                                Settings.LoadSettings(settingsJC);

                                if (Settings.restoreNameJSS.val != "")
                                {
                                    Settings.restoreNameJSS.val = "";
                                    Settings.SaveSettings();
                                }
                            }
                            
                        }
                        
                    }
                    else
                    {
                        JSONStorableString text = new JSONStorableString("text", "Only one instance of Browser Assist can be active at a time.");
                        CreateTextField(text);
                    }
                    
                }
                else
                {
                    JSONStorableString text = new JSONStorableString("text", "Browser Assist must be loaded as a Session Plugin");
                    CreateTextField(text);
                }

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(Init)}: {e}");
                OnDestroy();
            }
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            jc["BAMajorVersion"].AsInt = Const.BAMajorVersion;
            jc["BAMinorVersion"].AsInt = Const.BAMinorVersion;
            jc["BAFixVersion"].AsInt = Const.BAFixVersion;
            jc["BAInteropVersion"].AsInt = Const.BAInteropVersion;
            jc["BAPatreonVersion"].AsBool = PC.patreonContentEnabled;

            return jc;
        }

        private int baPluginCount { get
            {
                int browserAssistCount = 0;
                foreach (var plugin in PluginUtils.GetSessionPlugins())
                {
                    if (plugin.name.EndsWith("_JayJayWon.BrowserAssist")) browserAssistCount++;

                }
                return browserAssistCount;
            } }

        private bool legacyMetaDataExists { get
            {
                bool legacyVARMetaDataExists = FileManagerSecure.DirectoryExists(Const.VARLegacyUserDataFolder);
                bool newVARMetaDataExists = FileManagerSecure.DirectoryExists(Const.VARGroupCoreDataFilesFolder);
                if (legacyVARMetaDataExists && FileManagerSecure.GetFiles(Const.VARLegacyUserDataFolder).Count() > 0 && !newVARMetaDataExists) return true;
                else return false;

            } }

        private void MarkBackupAsCorrupt(string backupName)
        {
            try
            {
                string unusedCorruptBackupFolderName = backupName;
                int nextUniqueID = 1;

                while (FileManagerSecure.DirectoryExists(Const.MetaDataBackupFolder + "\\" + unusedCorruptBackupFolderName + " CORRUPTED"))
                {
                    unusedCorruptBackupFolderName = backupName + "(" + nextUniqueID + ")";
                    nextUniqueID++;
                }

                string corruptBackupFolder = Const.MetaDataBackupFolder + "\\" + unusedCorruptBackupFolderName + " CORRUPTED";
            
                FileManagerSecure.MoveFile(Const.MetaDataBackupFolder + "\\" + backupName, corruptBackupFolder);
            }
            catch (Exception e)
            {
                SuperController.LogError("BA.BrowserAssist.MarkBackupAsCorrupt: Exception occured whilst trying to mark a corrup backup folder '" + backupName+"' as CORRUPTED. " + e);
                SuperController.LogMessage("WARNING: BrowserAssist detected that a restore from backup '"+ backupName+"' was in progress. A data corruption was detected in the restored backup. BrowserAssist was attempting to mark the backup as corrupt, but an error occured during that process (see Error Log window). The backup should not be used as a valid restore point.");
            }
        }

        private void RegisterVAMHandlers()
        {
            atomRenameHandle = new SuperController.OnAtomUIDRename(this.OnAtomNameUpdate);
            SuperController.singleton.onAtomUIDRenameHandlers += atomRenameHandle;

            atomRemoveHandle = new SuperController.OnAtomRemoved(this.OnAtomRemoved);
            SuperController.singleton.onAtomRemovedHandlers += atomRemoveHandle;

            sceneLoadedHandle = new SuperController.OnSceneLoaded(OnSceneLoaded);
            SuperController.singleton.onSceneLoadedHandlers += sceneLoadedHandle;

            atomAddedHandle = new SuperController.OnAtomAdded(OnAtomAdded);
            SuperController.singleton.onAtomAddedHandlers += atomAddedHandle;
        }

        private void SetupPluginIntegrations()
        {
            // Initiate integration with Acidbubbles KeyBindings
            SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);

            SuperController.singleton.BroadcastMessage("OnBrowserAssistAvailable", this, SendMessageOptions.DontRequireReceiver);
        }
        private void CreateTopScreenLaunchButton()
        {
            GameObject canvasLaunchButtonObject = new GameObject();
            topScreenLaunchButtonCanvas = canvasLaunchButtonObject.AddComponent<Canvas>();
            topScreenLaunchButtonCanvas.renderMode = RenderMode.WorldSpace;
            topScreenLaunchButtonCanvas.pixelPerfect = false;
            canvasLaunchButtonObject.layer = SuperController.singleton.topWorldUI.gameObject.layer;

            topScreenLaunchButtonCanvas.transform.SetParent(SuperController.singleton.topWorldUI, false);
            CanvasScaler topScreenLaunchButtonCS = canvasLaunchButtonObject.AddComponent<CanvasScaler>();
            topScreenLaunchButtonCS.scaleFactor = 100f;
            topScreenLaunchButtonCS.dynamicPixelsPerUnit = 1f;
            canvasLaunchButtonObject.AddComponent<GraphicRaycaster>();

            vamTopMenuLaunchButton = UI.CreateUIButton(topScreenLaunchButtonCanvas.gameObject, 100f, 100f, "Browser Assist", 40);
            vamTopMenuLaunchButton.buttonColor = Color.white;
            string thumbnailPath = Utils.GetLatestVARPath("JayJayWon.BrowserAssist.latest:/Custom/Scripts/JayJayWon/BrowserAssist/BrowserAssist.jpg");
            if (!FileManagerSecure.FileExists(thumbnailPath))
            {
                thumbnailPath = Utils.GetLatestVARPath("JayJayWon.BrowserAssist(Free).latest:/Custom/Scripts/JayJayWon/BrowserAssist/BrowserAssist(Free).jpg");
            }
            ImageUtils.QueueLoadTexture(thumbnailPath, TopScreenLaunchButtonImageLoadedCallback, true);
            vamTopMenuLaunchButton.button.onClick.AddListener(delegate
            {
                mainBrowserUI.ToggleActive();
            });

            SuperController.singleton.AddCanvas(topScreenLaunchButtonCanvas);

            topScreenLaunchButtonCanvas.transform.localPosition = new Vector3(320f, -800f, 0f);

            ToggleTopMenuLanchButton();
        }
        private void TopScreenLaunchButtonImageLoadedCallback(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi.tex != null)
            {
                
                Texture2D tex = qi.tex;
                vamTopMenuLaunchButton.button.image.sprite = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.5f, 0.5f), 100.0f);
                vamTopMenuLaunchButton.buttonText.text = "";
            }
        }
        public void ToggleTopMenuLanchButton()
        {
            if (Settings.enableTopScreenLaunchButtonJSB.val) topScreenLaunchButtonCanvas.gameObject.SetActive(true);
            else topScreenLaunchButtonCanvas.gameObject.SetActive(false);
        }

        private void CreateVAMUILaunchButton()
        {
            var vamSceneBrowserTransform = SuperController.singleton.mainHUD.Find("LowerHUDPivot/LowerHUDFlip/Scene Control Canvas/Toolbar/AutoLayoutArea/ButtonOpenSceneFullDialog");
            vamSceneBrowserButton = vamSceneBrowserTransform.GetComponent<Button>();
            

            var moreOptionstransform = SuperController.singleton.mainHUD.Find("LowerHUDPivot/LowerHUDFlip/Scene Control Canvas/Panel/ClickForMoreOptions/OtherOptionsPopup");

            moreOptionsGameObject = moreOptionstransform.gameObject;

            GameObject canvasLaunchButtonObject = new GameObject();
            launchButtonCanvas = canvasLaunchButtonObject.AddComponent<Canvas>();
            launchButtonCanvas.renderMode = RenderMode.WorldSpace;
            launchButtonCanvas.pixelPerfect = false;

            canvasLaunchButtonObject.layer = SuperController.singleton.mainHUD.gameObject.layer;
            launchButtonCanvas.transform.SetParent(SuperController.singleton.mainHUD, false);
            CanvasScaler launchButtonCS = canvasLaunchButtonObject.AddComponent<CanvasScaler>();
            launchButtonCS.scaleFactor = 100f;
            launchButtonCS.dynamicPixelsPerUnit = 1f;
            GraphicRaycaster gr = canvasLaunchButtonObject.AddComponent<GraphicRaycaster>();

            vamUILaunchButton = UI.CreateUIButton(launchButtonCanvas.gameObject, 38f, 36f, "Browser Assist", 20);
            vamUILaunchButton.buttonColor = Color.white;
            string thumbnailPath = Utils.GetLatestVARPath("JayJayWon.BrowserAssist.latest:/Custom/Scripts/JayJayWon/BrowserAssist/BrowserAssist.jpg");
            if (!FileManagerSecure.FileExists(thumbnailPath))
            {
                thumbnailPath = Utils.GetLatestVARPath("JayJayWon.BrowserAssist(Free).latest:/Custom/Scripts/JayJayWon/BrowserAssist/BrowserAssist(Free).jpg");
            }
            ImageUtils.QueueLoadTexture(thumbnailPath, LaunchButtonImageLoadedCallback, true);
            vamUILaunchButton.button.onClick.AddListener(delegate
            {
                mainBrowserUI.ToggleActive();
            });
            vamUILaunchButton.button.gameObject.AddComponent <LaunchButtonPointerEnterExitHelp>();
            SuperController.singleton.AddCanvas(launchButtonCanvas);

            launchButtonCanvas.transform.localScale = new Vector3(0.001f, 0.001f, 0.001f);
            RePositionLaunchButton();

        }

        public void RePositionLaunchButton()
        {
            vamSceneBrowserButton.interactable = true;
            launchButtonCanvas.gameObject.SetActive(true);
            switch (Settings.launchButtonPlacementJSESC.val)
            {
                case VAMUILaunchButtonPosition.toolBarGap:
                    launchButtonCanvas.transform.localPosition = new Vector3(-0.131f, -0.074f, 0f);
                    break;
                case VAMUILaunchButtonPosition.sceneBrowserSub:
                    launchButtonCanvas.transform.localPosition = new Vector3(0.2205f, -0.074f, 0f);
                    vamSceneBrowserButton.interactable = false;
                    break;
                case VAMUILaunchButtonPosition.toolBarLeftInnerUpper:
                    launchButtonCanvas.transform.localPosition = new Vector3(0.37f, -0.074f, 0f);
                    break;
                case VAMUILaunchButtonPosition.toolBarRightInnerUpper:
                    launchButtonCanvas.transform.localPosition = new Vector3(-0.37f, -0.074f, 0f);
                    break;
                case VAMUILaunchButtonPosition.disabled:
                    launchButtonCanvas.gameObject.SetActive(false);
                    break;
            }

            if (Globals.isVREnabled)
            {
                launchButtonCanvas.transform.localEulerAngles = new Vector3(32, 180, 0);
                launchButtonCanvas.transform.Translate(new Vector3(0f, -0.01f * mainBrowserUI.worldScaleAdjust, -0.039f * mainBrowserUI.worldScaleAdjust));
            }
            else
            {
                launchButtonCanvas.transform.localEulerAngles = new Vector3(0, 180, 0);
            }
        }

        private void LaunchButtonImageLoadedCallback(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi.tex != null)
            {
                Texture2D tex = qi.tex;
                vamUILaunchButton.button.image.sprite = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.5f, 0.5f), 100.0f);
                vamUILaunchButton.buttonText.text = "";
            }
        }

        private void CopySymLinkBatScripts()
        {
            try
            {
                string pluginSlot = name.Substring(0, name.IndexOf('_'));               
                JSONClass pluginsJC = manager.GetJSON(true, true, true).AsObject["plugins"].AsObject;

                string pluginPath = pluginsJC[pluginSlot];
                if (pluginPath.Contains(":"))
                {
                    string varName = pluginPath.Substring(0, pluginPath.IndexOf(':'));
                    string createSymLinksSourceFileName = FileManagerSecure.NormalizePath(varName + ":\\" + Const.BAScriptFolder + "\\CreateBASymLinks.bat");
                    string removeSymLinksSourceFileName = FileManagerSecure.NormalizePath(varName + ":\\" + Const.BAScriptFolder + "\\RemoveBASymLinks.bat");

                    if (FileManagerSecure.FileExists(createSymLinksSourceFileName))
                    {
                        FileManagerSecure.DeleteFile(Const.BAScriptDataFolder + "\\CreateBASymLinks.bat");
                        FileManagerSecure.CopyFile(createSymLinksSourceFileName, Const.BAScriptDataFolder + "\\CreateBASymLinks.bat");
                    }
                    if (FileManagerSecure.FileExists(removeSymLinksSourceFileName))
                    {
                        FileManagerSecure.DeleteFile(Const.BAScriptDataFolder + "\\RemoveBASymLinks.bat");
                        FileManagerSecure.CopyFile(removeSymLinksSourceFileName, Const.BAScriptDataFolder + "\\RemoveBASymLinks.bat");
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(CopySymLinkBatScripts)}: {e}");
            }
        }
        private void CheckSymLinkAccess()
        {
            try
            {
                string guid = SuperController.singleton.GetTempUID();
                if (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\Custom"))
                {
                    FileManagerSecure.WriteAllBytes(Const.SymLinksFolder + "\\Custom\\" + guid, new byte[1]);
                    if (FileManagerSecure.FileExists("Custom\\" + guid))
                    {
                        Globals.customFolderAccess = true;
                        FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\Custom\\" + guid);
                    }
                }
                if (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\Saves\\scene"))
                {
                    FileManagerSecure.WriteAllBytes(Const.SymLinksFolder + "\\Saves\\scene\\" + guid, new byte[1]);
                    if (FileManagerSecure.FileExists("Saves\\scene\\" + guid))
                    {
                        Globals.scenesFolderAccess = true;
                        FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\Saves\\scene\\" + guid);
                    }
                }
                if (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\AddonPackagesFilePrefs"))
                {
                    FileManagerSecure.WriteAllBytes(Const.SymLinksFolder + "\\AddonPackagesFilePrefs\\" + guid, new byte[1]);
                    if (FileManagerSecure.FileExists("AddonPackagesFilePrefs\\" + guid))
                    {
                        Globals.addonPackageFilePrefsFolderAccess = true;
                        FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\AddonPackagesFilePrefs\\" + guid);
                    }
                }
                if (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\AddonPackages"))
                {
                    FileManagerSecure.WriteAllBytes(Const.SymLinksFolder + "\\AddonPackages\\" + guid, new byte[1]);
                    if (FileManagerSecure.FileExists("AddonPackages\\" + guid))
                    {
                        Globals.addonPackagesAccess = true;
                        FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\AddonPackages\\" + guid);
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(CheckSymLinkAccess)}: {e}");
            }
            
        }

        public void OnBrowserAssistConsumerAvailable(MVRScript consumer)
        {
            if (isEnabled) consumer.SendMessage("OnBrowserAssistAvailable", this, SendMessageOptions.DontRequireReceiver);
        }

        public void OnGetTagNames(object[] objects)
        {
            if (isEnabled)
            {
                BAInteropServer.OnGetTagNames(objects);
            }

        }

        public void OnGetCreatorNames(object[] objects)
        {
            if (isEnabled)
            {
                BAInteropServer.OnGetCreatorNames(objects);
            }

        }

        public void OnGetFilteredResourceFilenames(object[] objects)
        {
            if (isEnabled)
            {
                BAInteropServer.OnGetFilteredResourceFilenames(objects);
            }

        }

        public void OnGetOrderedFilteredResourceFilenames(object[] objects)
        {
            if (isEnabled)
            {
                BAInteropServer.OnGetOrderedFilteredResourceFilenames(objects);
            }

        }

        public void OnSelectResourceFilename(object[] objects)
        {
            if (isEnabled)
            {
                BAInteropServer.OnSelectResourceFilename(objects);
            }

        }

        public void OnSetMainResourceFilters(object[] objects)
        {
            if (isEnabled)
            {
                BAInteropServer.OnSetMainResourceFilters(objects);
            }

        }

        public void OnGetSupportedResourceTypes(object[] objects)
        {
            if (isEnabled)
            {
                BAInteropServer.OnGetSupportedResourceTypes(objects);
            }

        }

        public void OnGetSupportedAndConcealedResourceTypes(object[] objects)
        {
            if (isEnabled)
            {
                BAInteropServer.OnGetSupportedAndConcealedResourceTypes(objects);
            }

        }

        public void OnGetExistingAtomPresetResourceTypes(object[] objects)
        {
            if (isEnabled)
            {
                BAInteropServer.OnGetExistingAtomPresetResourceTypes(objects);
            }

        }
        public void OnGetExistingAndConcealedAtomPresetResourceTypes(object[] objects)
        {
            if (isEnabled)
            {
                BAInteropServer.OnGetExistingAndConcealedAtomPresetResourceTypes(objects);
            }

        }


        public void OnCloseVAMUIAttachedUI()
        {
            if (isEnabled)
            {
                CloseVAMUIAttachedUI(mainBrowserUI);
                CloseVAMUIAttachedUI(callbackBrowserUI);
            }
        }

        private void CloseVAMUIAttachedUI(BrowserUI browserUI)
        {
            if (browserUI.attachedVAMUI && browserUI.isActive && browserUI.isVisible) browserUI.SetInactive();
        }

        public void OnBindingsListRequested(List<object> bindings)
        {
            try
            {
                int pluginSlot = int.Parse(this.name.Substring(7, this.name.IndexOf('_') - 7));

                bindings.Add(new JSONStorableAction("Open UI", () =>
                {
                    mainBrowserUI.SetActive();
                }
                ));

                bindings.Add(new JSONStorableAction("Close UI", () =>
                {
                    mainBrowserUI.SetInactive();
                }
                ));

                

                bindings.Add(new JSONStorableAction("Toggle UI Open/Close", () =>
                {
                    mainBrowserUI.ToggleActive();
                }
                ));

                bindings.Add(new JSONStorableAction("Open UI: Scenes", () =>
                {
                    mainBrowserUI.SetActive();
                    mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.scene;
                }
                ));

                bindings.Add(new JSONStorableAction("Open UI: Presets Appearance", () =>
                {
                    mainBrowserUI.SetActive();
                    mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.presetAppearance;
                }
                ));

                if (PC.patreonContentEnabled)
                {
                    bindings.Add(new JSONStorableAction("Open UI: Presets Breast Physics", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.presetBreastPhys;
                    }
               ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Clothing", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.presetClothing;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets General", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.presetGeneral;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Glute Phys", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.presetGlutePhysics;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Hair", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.presetHair;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Morphs", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.presetMorphs;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Plugins", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.presetPlugins;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Pose", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.presetPose;
                    }
                    ));
                    bindings.Add(new JSONStorableAction("Open UI: Presets Skin", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.presetSkin;
                    }
                    ));

                    bindings.Add(new JSONStorableAction("Open UI: Plugins", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.plugins;
                    }
                    ));

                    bindings.Add(new JSONStorableAction("Open UI: SubScenes", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.subScenes;
                    }
                    ));

                    bindings.Add(new JSONStorableAction("Open UI: Clothing (Female)", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.femaleClothingItems;
                    }
                    ));

                    bindings.Add(new JSONStorableAction("Open UI: Clothing (Male)", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.maleClothingItems;
                    }
                    ));

                    bindings.Add(new JSONStorableAction("Open UI: Scene & Session Plugin Presets", () =>
                    {
                        mainBrowserUI.SetActive();
                        mainBrowserUI.browserFiltersUI.currentResourceType = ResourceType.presetSceneSessionPlugins;
                    }
                    ));

                }
               

                bindings.Add(new JSONStorableAction("Toggle 'Only Favorites' filter", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.browserFiltersUI.onlyFavouritesJSB.val = !mainBrowserUI.browserFiltersUI.onlyFavouritesJSB.val;
                }
                ));

                bindings.Add(new JSONStorableAction("Toggle 'Show Hidden' filter", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.browserFiltersUI.showHiddenJSB.val = !mainBrowserUI.browserFiltersUI.showHiddenJSB.val;
                }
                ));

                bindings.Add(new JSONStorableAction("Toggle 'Only New' filter", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.browserFiltersUI.onlyNewJSB.val = !mainBrowserUI.browserFiltersUI.onlyNewJSB.val;
                }
                ));

                bindings.Add(new JSONStorableAction("First Results Page", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.resourceBrowserUI.FirstPageResults();
                }
                ));

                bindings.Add(new JSONStorableAction("Previous Results Page", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.resourceBrowserUI.PrevPageResults();
                }
                ));

                bindings.Add(new JSONStorableAction("Next Results Page", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.resourceBrowserUI.NextPageResults();
                }
                ));

                bindings.Add(new JSONStorableAction("Last Results Page", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.resourceBrowserUI.LastPageResults();
                }
                ));

                bindings.Add(new JSONStorableAction("Increase Thumbnail Size", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.resourceBrowserUI.IncreaseButtonSizes();
                }
                ));

                bindings.Add(new JSONStorableAction("Decrease Thumbnail Size", () =>
                {
                    if (mainBrowserUI.isVisible) mainBrowserUI.resourceBrowserUI.DecreaseButtonSizes();
                }
                ));
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnBindingsListRequested)}: {e}");
            }
        }

        
        private void OnSceneLoaded()
        {
            try
            {
                if (mainBrowserUI!=null && mainBrowserUI.isVREnabled) mainBrowserUI.SetInactive();
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnSceneLoaded)}: {e}");
            }
        }
        private void OnAtomAdded(Atom atom)
        {
         
            try
            {
                if (mainBrowserUI != null) mainBrowserUI.resourceActionsUI.OnAtomAdded(atom);

                AtomUtils.OnAtomAdded(atom);
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnAtomAdded)}: {e}");
            }
        }

        private void OnAtomNameUpdate(string oldName, string newName)
        {           
            try
            {
                if (mainBrowserUI != null) mainBrowserUI.resourceActionsUI.AtomNameUpdated(oldName, newName);               
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnAtomNameUpdate)}: {e}");
            }
        }

        private void OnAtomRemoved(Atom atom)
        {
            try
            {
                if (mainBrowserUI != null) mainBrowserUI.resourceActionsUI.OnAtomRemoved(atom);

                AtomUtils.OnAtomAdded(atom);
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnAtomRemoved)}: {e}");
            }
        }

        public void OnEnable()
        {
            try
            {

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnEnable)}: {e}");
            }
        }

        public void OnDisable()
        {
            try
            {

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnDisable)}: {e}");
            }
        }

        public void OnDestroy()
        {
            try
            {
                if (isEnabled)
                {
                    SuperController.singleton.onAtomUIDRenameHandlers -= atomRenameHandle;
                    SuperController.singleton.onAtomRemovedHandlers -= atomRemoveHandle;
                    SuperController.singleton.onAtomAddedHandlers -= atomAddedHandle;
                    SuperController.singleton.onSceneLoadedHandlers -= sceneLoadedHandle;

                    foreach (Atom atom in SuperController.singleton.GetAtoms()) AtomUtils.OnAtomRemoved(atom);

                    //Deregisters with Acidbubbles KeyBindings plugin
                    SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);

                    SuperController.singleton.BroadcastMessage("OnBrowserAssistDestroyed", SendMessageOptions.DontRequireReceiver);

                    if (SuperController.singleton != null)
                    {
                        SuperController.singleton.RemoveCanvas(launchButtonCanvas);
                        SuperController.singleton.RemoveCanvas(topScreenLaunchButtonCanvas);
                    }
                    if (launchButtonCanvas != null) launchButtonCanvas.transform.SetParent(null, false);
                    if (launchButtonCanvas.gameObject != null) GameObject.Destroy(launchButtonCanvas.gameObject);
                    if (topScreenLaunchButtonCanvas != null) topScreenLaunchButtonCanvas.transform.SetParent(null, false);
                    if (topScreenLaunchButtonCanvas.gameObject != null) GameObject.Destroy(topScreenLaunchButtonCanvas.gameObject);

                    vamSceneBrowserButton.interactable = true;

                    if (mainBrowserUI != null) mainBrowserUI.OnDestroy();
                    if (callbackBrowserUI != null) callbackBrowserUI.OnDestroy();

                    ResourceManifest.OnDestoy();
                    Resources.UnloadUnusedAssets();
                    GC.Collect();
                }

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnDestroy)}: {e}");
            }
        }


        void Start()
        {
            try
            {
                RegisterFloat(Settings.uiScaleJSF);
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(Start)}: {e}");
            }
        }


        private DateTime lastDCICacheClearTime= DateTime.MinValue;

        void Update()
        {
            try
            {

                if (isEnabled && (Globals.failureOnLocalMetaDataLoad || Globals.failureOnVARMetaDataLoad))
                {
                    if (mainBrowserUI.isActive) SuperController.LogMessage("An unexpected error occured when scanning the BrowserAssist meta data for VAR packages or Local Files. To preserve the integrity of the meta data, BrowserAssist will not operate until this issue is resolved. Please contact JayJayWon on his Discord channel for assistance. Its expected that you will get Error messages in the red VAM Error Log window on BA startup - please capture these and share with JayJayWon on Discord.");
                    mainBrowserUI.SetInactive();
                }
                else
                {
                    if (isEnabled)
                    {
                        if (vamUILaunchButton != null && moreOptionsGameObject != null)
                        {
                            if (moreOptionsGameObject.activeSelf) vamUILaunchButton.gameObject.SetActive(false);
                            else vamUILaunchButton.gameObject.SetActive(true);
                        }

                        if (mainBrowserUI != null)
                        {
                            mainBrowserUI.Update();

                            /*
                            var diffInSeconds = (DateTime.Now - lastDCICacheClearTime).TotalSeconds;
                            if (diffInSeconds > 1f)
                            {
                                ResourcePrimaryActionUI.ClearCachedDCIs();
                                if (mainBrowserUI.resourceActionsUI != null && mainBrowserUI.resourceActionsUI.primaryActionUITab != null) mainBrowserUI.resourceActionsUI.primaryActionUITab.RefreshActiveClothing();

                                // Refresh the Clothing Item Browser on clothing item being worn or removed
                                if (mainBrowserUI.resourceActionsUI!=null && mainBrowserUI.resourceActionsUI.currentActionUITabType == ActionUITabType.clothingPresetTab)
                                {
                                    var filtersResourceType = mainBrowserUI.browserFiltersUI.currentResourceType;
                                    var filtersAtomPresetResourceType = mainBrowserUI.browserFiltersUI.currentAtomPresetResourceType;
                                    var targetAtomJSMultiEnum = mainBrowserUI.resourceActionsUI.primaryActionUITab.targetAtomJSMultiEnum;
                                    var primaryRVGE = mainBrowserUI.resourceSelectionsUI.GetFirstSelectedRVGE(filtersResourceType, filtersAtomPresetResourceType);
                                    var resourceClothingPresetsActionUI = mainBrowserUI.resourceActionsUI.currentActionUITab as ResourceClothingPresetsActionUI;

                                    if (targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag && AtomUtils.GetClothingActive(SuperController.singleton.GetAtomByUid(targetAtomJSMultiEnum.mainVal), primaryRVGE).val)
                                    {
                                        if (!resourceClothingPresetsActionUI.isNewPresetActive) resourceClothingPresetsActionUI.RefreshUI();
                                    }
                                    else if (resourceClothingPresetsActionUI.isNewPresetActive)
                                    {
                                        resourceClothingPresetsActionUI.RefreshUI();
                                    }
                                }
                                lastDCICacheClearTime = DateTime.Now; 
                            }
                            
                            */
                        }
                        if (callbackBrowserUI != null) callbackBrowserUI.Update();

                        DataFileManifest.Update();
                        ResourceManifest.Update();
                        ResourceVersionGroupEntry.Update();
                        
                    }
                }

                
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(Update)}: {e}");
            }

        }
    }

}
