// BrowserAssistInterop.cs  v1.3.1
// This file contains classes that can be used to integrate with the BrowserAssist plugin by JayJayWon. The main class 'BAInterop' has a range of static methods and properties that can be used once a reference to Browser Assist has been established.This reference can be set using the static property 'BAPlugin'. There is a standard code pattern available to establish the BrowserAssist reference. Other static classes are included that provide unique enumerations for various parameters that can be passed to BrowserAssist.
// To use this file you will need to change the namespace to one that matches the namespace used by a plugin wanting to interact with BrowserAssist. This file will also need to be referenced in the VAM cslist file along with any other .cs files. NO OTHER CHANGES SHOULD BE REQUIRED TO THIS FILE

// This integration obviously requires BrowserAssist to be running as a session plugin when its active. Different API calls are supported by different BrowserAssist versions:
//      GetTagNames requires BrowserAssist v1.10 (or greater)
//      GetCreatorNames requires BrowserAssist v1.10 (or greater)
//      GetSupportedResourceTypes requires BrowserAssist v1.10 (or greater)
//      GetFilteredResourceFilenames requires BrowserAssist v1.10 (or greater)
//      GetOrderedFilteredResourceFilenames requires BrowserAssist v1.10 (or greater)
//      GetSupportedResourceTypes requires BrowserAssist v1.10 (or greater)
//      SelectResourceFilename requires BrowserAssist v1.11 (or greater)
//      SetMainResourceFilters requires BrowserAssist v1.24.1 (or greater)
//      GetExistingAtomPresetResourceTypes requires BrowserAssist v1.24.1 (or greater)
//      GetConcealedResourceTypes requires BrowserAssist v1.24.1 (or greater)
//      GetConcealedAtomPresetResourceTypes requires BrowserAssist v1.24.1 (or greater)

// A number of API calls can be used to pre-determine if BrowserAssist is available:
//      isBAEnabled will return true if BrowserAssist v1.10 (or greater) is available
//      isSelectResourceFilenameAvailable will return true if BrowserAssist v1.11 (or greater) is available

using SimpleJSON;
using System;
using System.Collections.Generic;
using UnityEngine;

namespace JayJayWon
{
    public class BAInterop
    {
        private static MVRScript _BAPlugin;
        public static bool enableFailureLogging { get; set; } = false;

        // 'BAInteropVersion' will return the version of the interop specification used by the currently connected BrowserAssist plugin. If no BrowserAssist plugin is connected the property will be set to -1.
        public static int BAInteropVersion { get; private set; } = -1;

        // 'BAPlugin' must be set to reference the BrowserAssist plugin to be connected to. There is a standard pattern to do this which requires specific methods to be impelmented in 
        public static MVRScript BAPlugin
        {
            get
            {
                return _BAPlugin;
            }
            set
            {
                if (value == null)
                {
                    _BAPlugin = null;
                    BAInteropVersion = -1;
                }
                else
                {
                    _BAPlugin = value;
                    var jc = _BAPlugin.GetJSON();
                    if (jc["BAInteropVersion"] != null) BAInteropVersion = jc["BAInteropVersion"].AsInt;
                    else
                    {
                        BAInteropVersion = -1;
                        if (enableFailureLogging) SuperController.LogMessage("BAInterop.BAPlugin: The specidied MVRScript is not a plugin version that supports  BrowserAssist integration.");
                    }
                }

            }
        }

        // 'isBAenabled' will return true if a connection is established with BrowserAssist and that plugin is not currently disabled.
        public static bool isBAenabled
        {
            get
            {
                return BAPlugin != null && BAPlugin.enabled;
            }
        }


        public delegate void CallbackFail(string failureMessage);
        public delegate void CallbackException(string exceptionMessage);

        public delegate void GetStringListCallback(List<string> stringList);
        public delegate void GetIntHashsetCallback(HashSet<int> intHashet);
        public delegate void GetTwoIntHashsetCallback(HashSet<int> intHashet, HashSet<int> intHashet2);
        public delegate void GetStringHashsetCallback(HashSet<string> stringHashset);
        public delegate void GetTwoStringHashsetCallback(HashSet<string> stringHashset, HashSet<string> stringHashset2);
        public delegate void GetStringCallback(string stringCallback);

        // 'GetTagNames' will obtain a list of all Tag Names for the specified Resource Type. The callback method specified will be triggered with the list of tag names as a parameter. Alternative callbacks can optionally be specified if the function fails or triggers an exception. The atomPresetResourceName parameter only needs to be set if resourceType is atomPreset (for non person Atom Presets). atomPresetResourceName should be set to  a string containing the name of the type of atom required (e.g. 'Empty' or 'InvisibleLight').
        public static void GetTagNames(int resourceTypeEnum, GetStringListCallback callback, CallbackFail callbackFail = null, CallbackException callbackException = null, string atomPresetResourceName="")
        {
            try
            {
                if (isBAenabled)
                {
                    var tagsList = new List<string>();

                    JSONStorableAction callbackActionSuccess = new JSONStorableAction("successCallback", delegate
                    {
                        if (callback != null) callback(tagsList);
                    });

                    JSONStorableString callbackActionFail = new JSONStorableString("failCallback", "", delegate (string failureMessage)
                    {
                        if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetTagNames: " + failureMessage);
                        if (callbackFail != null) callbackFail(failureMessage);
                    });

                    JSONStorableString callbackExceptionMessage = new JSONStorableString("exceptionCallback", "", delegate (string exceptionMessage)
                    {
                        if (callbackException != null) callbackException(exceptionMessage);
                        else SuperController.LogError("BAInterop.GetTagNames exception callback from BA: " + exceptionMessage);
                    });

                    var objects = new object[6];
                    objects[0] = resourceTypeEnum;
                    objects[1] = tagsList;
                    objects[2] = callbackActionSuccess;
                    objects[3] = callbackActionFail;
                    objects[4] = callbackExceptionMessage;
                    objects[5] = atomPresetResourceName;

                    BAPlugin.SendMessage("OnGetTagNames", objects);
                }
                else if (callbackFail != null)
                {
                    string failureMessage = "A request to retrieve all Tag Names from BrowswerAssist failed due to the plugin not being available.";
                    if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetTagNames: " + failureMessage);
                    callbackFail(failureMessage);

                }
            }
            catch (Exception e)
            {
                if (callbackException != null) callbackException(e.ToString());
                else SuperController.LogError("BAInterop.GetTagNames exception: " + e);
            }
        }

        // 'GetCreatorNames' will obtain a list of all Creator Names for the specified Resource Type. The callback method specified will be triggered with the list of Creator names as a parameter. Alternative callbacks can optionally be specified if the function fails or triggers an exception. The atomPresetResourceName parameter only needs to be set if resourceType is atomPreset (for non person Atom Presets). atomPresetResourceName should be set to  a string containing the name of the type of atom required (e.g. 'Empty' or 'InvisibleLight').
        public static void GetCreatorNames(int resourceTypeEnum, GetStringListCallback callback, CallbackFail callbackFail = null, CallbackException callbackException = null, string atomPresetResourceName = "")
        {
            try
            {
                if (isBAenabled)
                {
                    var creatorsList = new List<string>();

                    JSONStorableAction callbackActionSuccess = new JSONStorableAction("successCallback", delegate
                    {
                        if (callback != null) callback(creatorsList);
                    });

                    JSONStorableString callbackActionFail = new JSONStorableString("failCallback", "", delegate (string failureMessage)
                    {
                        if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetCreatorNames: " + failureMessage);
                        if (callbackFail != null) callbackFail(failureMessage);
                    });

                    JSONStorableString callbackExceptionMessage = new JSONStorableString("exceptionCallback", "", delegate (string exceptionMessage)
                    {
                        if (callbackException != null) callbackException(exceptionMessage);
                        else SuperController.LogError("BAInterop.GetCreatorNames exception callback from BA: " + exceptionMessage);
                    });

                    var objects = new object[6];
                    objects[0] = resourceTypeEnum;
                    objects[1] = creatorsList;
                    objects[2] = callbackActionSuccess;
                    objects[3] = callbackActionFail;
                    objects[4] = callbackExceptionMessage;
                    objects[5] = atomPresetResourceName;

                    BAPlugin.SendMessage("OnGetCreatorNames", objects);
                }
                else if (callbackFail != null)
                {
                    string failureMessage = "A request to retrieve all Creator Names from BrowswerAssist failed due to the plugin not being available.";
                    if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetCreatorNames: " + failureMessage);
                    callbackFail(failureMessage);

                }
            }
            catch (Exception e)
            {
                if (callbackException != null) callbackException(e.ToString());
                else SuperController.LogError("BAInterop.GetCreatorNames exception: " + e);
            }
        }

        
        // 'GetSupportedResourceTypes' will obtain a list of all Resource Types that are supported. The callback method specified will be triggered with the list of resource type as a parameter.  Alternative callbacks can optionally be specified if the function fails or triggers an exception 
        public static void GetSupportedResourceTypes(GetIntHashsetCallback callback, CallbackFail callbackFail = null, CallbackException callbackException = null)
        {
            try
            {
                if (isBAenabled)
                {
                    var resourceTypesList = new HashSet<int>();

                    JSONStorableAction callbackActionSuccess = new JSONStorableAction("successCallback", delegate
                    {
                        if (callback != null) callback(resourceTypesList);
                    });

                    JSONStorableString callbackActionFail = new JSONStorableString("failCallback", "", delegate (string failureMessage)
                    {
                        if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetSupportedResourceTypes: " + failureMessage);
                        if (callbackFail != null) callbackFail(failureMessage);
                    });

                    JSONStorableString callbackExceptionMessage = new JSONStorableString("exceptionCallback", "", delegate (string exceptionMessage)
                    {
                        if (callbackException != null) callbackException(exceptionMessage);
                        else SuperController.LogError("BAInterop.GetSupportedResourceTypes exception callback from BA: " + exceptionMessage);
                    });

                    var objects = new object[5];
                    objects[0] = resourceTypesList;
                    objects[1] = callbackActionSuccess;
                    objects[2] = callbackActionFail;
                    objects[3] = callbackExceptionMessage;

                    BAPlugin.SendMessage("OnGetSupportedResourceTypes", objects);
                }
                else if (callbackFail != null)
                {
                    string failureMessage = "A request to retrieve all supported Resource Types from BrowserAssist failed due to the plugin not being available.";
                    if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetSupportedResourceTypes: " + failureMessage);
                    callbackFail(failureMessage);

                }
            }
            catch (Exception e)
            {
                if (callbackException != null) callbackException(e.ToString());
                else SuperController.LogError("BAInterop.GetSupportedResourceTypes exception: " + e);
            }
        }
        public static void GetSupportedAndConcealedResourceTypes(GetTwoIntHashsetCallback callback, CallbackFail callbackFail = null, CallbackException callbackException = null)
        {
            try
            {
                if (isBAenabled)
                {
                    var supportedResourceTypesList = new HashSet<int>();
                    var concealedResourceTypesList = new HashSet<int>();

                    JSONStorableAction callbackActionSuccess = new JSONStorableAction("successCallback", delegate
                    {
                        if (callback != null) callback(supportedResourceTypesList, concealedResourceTypesList);
                    });

                    JSONStorableString callbackActionFail = new JSONStorableString("failCallback", "", delegate (string failureMessage)
                    {
                        if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetSupportedAndConcealedResourceTypes: " + failureMessage);
                        if (callbackFail != null) callbackFail(failureMessage);
                    });

                    JSONStorableString callbackExceptionMessage = new JSONStorableString("exceptionCallback", "", delegate (string exceptionMessage)
                    {
                        if (callbackException != null) callbackException(exceptionMessage);
                        else SuperController.LogError("BAInterop.GetSupportedAndConcealedResourceTypes exception callback from BA: " + exceptionMessage);
                    });

                    var objects = new object[5];
                    objects[0] = supportedResourceTypesList;
                    objects[1] = concealedResourceTypesList;
                    objects[2] = callbackActionSuccess;
                    objects[3] = callbackActionFail;
                    objects[4] = callbackExceptionMessage;

                    BAPlugin.SendMessage("OnGetSupportedAndConcealedResourceTypes", objects);
                }
                else if (callbackFail != null)
                {
                    string failureMessage = "A request to retrieve all supported & concealed Resource Types from BrowserAssist failed due to the plugin not being available.";
                    if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetSupportedAndConcealedResourceTypes: " + failureMessage);
                    callbackFail(failureMessage);

                }
            }
            catch (Exception e)
            {
                if (callbackException != null) callbackException(e.ToString());
                else SuperController.LogError("BAInterop.GetSupportedAndConcealedResourceTypes exception: " + e);
            }
        }

        public static void GetExistingAtomPresetResourceTypes(GetStringHashsetCallback callback, CallbackFail callbackFail = null, CallbackException callbackException = null)
        {
            try
            {
                if (isBAenabled)
                {
                    var resourceTypesList = new HashSet<string>();

                    JSONStorableAction callbackActionSuccess = new JSONStorableAction("successCallback", delegate
                    {
                        if (callback != null) callback(resourceTypesList);
                    });

                    JSONStorableString callbackActionFail = new JSONStorableString("failCallback", "", delegate (string failureMessage)
                    {
                        if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetExistingAtomPresetResourceTypes: " + failureMessage);
                        if (callbackFail != null) callbackFail(failureMessage);
                    });

                    JSONStorableString callbackExceptionMessage = new JSONStorableString("exceptionCallback", "", delegate (string exceptionMessage)
                    {
                        if (callbackException != null) callbackException(exceptionMessage);
                        else SuperController.LogError("BAInterop.GetExistingAtomPresetResourceTypes exception callback from BA: " + exceptionMessage);
                    });

                    var objects = new object[5];
                    objects[0] = resourceTypesList;
                    objects[1] = callbackActionSuccess;
                    objects[2] = callbackActionFail;
                    objects[3] = callbackExceptionMessage;

                    BAPlugin.SendMessage("OnGetExistingAtomPresetResourceTypes", objects);
                }
                else if (callbackFail != null)
                {
                    string failureMessage = "A request to retrieve all existing Atom Preset Resource Types from BrowserAssist failed due to the plugin not being available.";
                    if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetExistingAtomPresetResourceTypes: " + failureMessage);
                    callbackFail(failureMessage);

                }
            }
            catch (Exception e)
            {
                if (callbackException != null) callbackException(e.ToString());
                else SuperController.LogError("BAInterop.GetExistingAtomPresetResourceTypes exception: " + e);
            }
        }
        public static void GetExistingAndConcealedAtomPresetResourceTypes(GetTwoStringHashsetCallback callback, CallbackFail callbackFail = null, CallbackException callbackException = null)
        {
            try
            {
                if (isBAenabled)
                {
                    var existingResourceTypesList = new HashSet<string>();
                    var concealedResourceTypesList = new HashSet<string>();

                    JSONStorableAction callbackActionSuccess = new JSONStorableAction("successCallback", delegate
                    {
                        if (callback != null) callback(existingResourceTypesList, concealedResourceTypesList);
                    });

                    JSONStorableString callbackActionFail = new JSONStorableString("failCallback", "", delegate (string failureMessage)
                    {
                        if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetExistingAndConcealedAtomPresetResourceTypes: " + failureMessage);
                        if (callbackFail != null) callbackFail(failureMessage);
                    });

                    JSONStorableString callbackExceptionMessage = new JSONStorableString("exceptionCallback", "", delegate (string exceptionMessage)
                    {
                        if (callbackException != null) callbackException(exceptionMessage);
                        else SuperController.LogError("BAInterop.GetExistingAndConcealedAtomPresetResourceTypes exception callback from BA: " + exceptionMessage);
                    });

                    var objects = new object[5];
                    objects[0] = existingResourceTypesList;
                    objects[1] = concealedResourceTypesList;
                    objects[2] = callbackActionSuccess;
                    objects[3] = callbackActionFail;
                    objects[4] = callbackExceptionMessage;

                    BAPlugin.SendMessage("OnGetExistingAndConcealedAtomPresetResourceTypes", objects);
                }
                else if (callbackFail != null)
                {
                    string failureMessage = "A request to retrieve all existing and concealed Atom Preset Resource Types from BrowserAssist failed due to the plugin not being available.";
                    if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetExistingAndConcealedAtomPresetResourceTypes: " + failureMessage);
                    callbackFail(failureMessage);

                }
            }
            catch (Exception e)
            {
                if (callbackException != null) callbackException(e.ToString());
                else SuperController.LogError("BAInterop.GetExistingAndConcealedAtomPresetResourceTypes exception: " + e);
            }
        }


        public static void GetFilteredResourceFilenames(BAResourceFilter filter, GetStringHashsetCallback callback, CallbackFail callbackFail = null, CallbackException callbackException = null)
        {
            try
            {
                if (isBAenabled)
                {
                    var resourceFileNames = new HashSet<string>();

                    JSONStorableAction callbackActionSuccess = new JSONStorableAction("successCallback", delegate
                    {
                        if (callback != null) callback(resourceFileNames);
                    });


                    JSONStorableString callbackActionFail = new JSONStorableString("failCallback", "", delegate (string failureMessage)
                    {
                        if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetFilteredResourceFilenames: " + failureMessage);
                        if (callbackFail != null) callbackFail(failureMessage);
                    });

                    JSONStorableString callbackExceptionMessage = new JSONStorableString("exceptionCallback", "", delegate (string exceptionMessage)
                    {
                        if (callbackException != null) callbackException(exceptionMessage);
                        else SuperController.LogError("BAInterop.GetFilteredResourceFilenames exception callback from BA: " + exceptionMessage);
                    });

                    var filterDict = new Dictionary<int, object>();
                    PackFilterDict(filter, filterDict);

                    var objects = new object[5];
                    objects[0] = filterDict;
                    objects[1] = resourceFileNames;
                    objects[2] = callbackActionSuccess;
                    objects[3] = callbackActionFail;
                    objects[4] = callbackExceptionMessage;

                    BAPlugin.SendMessage("OnGetFilteredResourceFilenames", objects);
                }
                else if (callbackFail != null)
                {
                    string failureMessage = "A request to retrieve a filtered list of resources from BrowswerAssist failed due to the plugin not being available.";
                    if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetFilteredResourceFilenames: " + failureMessage);
                    callbackFail(failureMessage);
                }
            }
            catch (Exception e)
            {
                if (callbackException != null) callbackException(e.ToString());
                else SuperController.LogError("BAInterop.GetFilteredResourceFilenames exception: " + e);
            }
        }

        public static void GetOrderedFilteredResourceFilenames(BAOrderedResourceFilter filter, GetStringListCallback callback, CallbackFail callbackFail = null, CallbackException callbackException = null)
        {
            try
            {
                if (isBAenabled)
                {
                    var resourceFileNames = new List<string>();

                    JSONStorableAction callbackActionSuccess = new JSONStorableAction("successCallback", delegate
                    {
                        if (callback != null) callback(resourceFileNames);

                    });


                    JSONStorableString callbackActionFail = new JSONStorableString("failCallback", "", delegate (string failureMessage)
                    {
                        if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetOrderedFilteredResourceFilenames: " + failureMessage);
                        if (callbackFail != null) callbackFail(failureMessage);
                    });

                    JSONStorableString callbackExceptionMessage = new JSONStorableString("exceptionCallback", "", delegate (string exceptionMessage)
                    {
                        if (callbackException != null) callbackException(exceptionMessage);
                        else SuperController.LogError("BAInterop.GetOrderedFilteredResourceFilenames exception callback from BA: " + exceptionMessage);
                    });

                    var filterDict = new Dictionary<int, object>();
                    PackFilterDict(filter, filterDict);

                    var objects = new object[5];
                    objects[0] = filterDict;
                    objects[1] = resourceFileNames;
                    objects[2] = callbackActionSuccess;
                    objects[3] = callbackActionFail;
                    objects[4] = callbackExceptionMessage;

                    BAPlugin.SendMessage("OnGetOrderedFilteredResourceFilenames", objects);
                }
                else if (callbackFail != null)
                {
                    string failureMessage = "A request to retrieve an ordered filtered list of resources from BrowswerAssist failed due to the plugin not being available.";
                    if (enableFailureLogging) SuperController.LogMessage("BAInterop.GetOrderedFilteredResourceFilenames: " + failureMessage);
                    callbackFail(failureMessage);
                }
            }
            catch (Exception e)
            {
                if (callbackException != null) callbackException(e.ToString());
                else SuperController.LogError("BAInterop.GetOrderedFilteredResourceFilenames exception: " + e);
            }
        }


        public static bool isSelectResourceFilenameAvailable { get {
                if (isBAenabled && BAInteropVersion >= 2) return true;
                return false; } }

        public static bool isSetMainResourceFiltersAvailable { get
            {
                if (isBAenabled && BAInteropVersion >= 4) return true;
                return false;
            }
        }

        public static void SelectResourceFilename(int resourceType, string atomPresetResourceType, GetStringCallback callback, int desktopDisplayMode=BADesktopDisplayModeEnum.rightDocked, string browserTitle = "Select Resource", int baKeepOpenOption = BAKeepOpenOptionEnum.unavailable, CallbackFail callbackFail = null, CallbackException callbackException = null)
        {

            Quaternion headFacingDirection = SuperController.singleton.lookCamera.transform.rotation;
            Vector3 offsetPosition = new Vector3(0f, 0f, 1.8f * SuperController.singleton.worldScale);
            Vector3 headPosition = SuperController.singleton.lookCamera.transform.position;

            Vector3 vrBrowserPosition = headPosition + headFacingDirection * offsetPosition;

            
            SelectResourceFilename(resourceType, atomPresetResourceType, vrBrowserPosition, callback, desktopDisplayMode, browserTitle, baKeepOpenOption, callbackFail, callbackException);
        }
        public static void SelectResourceFilename(int resourceType, string atomPresetResourceType, Vector3 vrBrowserPosition, GetStringCallback callback, int desktopDisplayMode = BADesktopDisplayModeEnum.rightDocked, string browserTitle = "Select Resource", int baKeepOpenOption = BAKeepOpenOptionEnum.unavailable, CallbackFail callbackFail = null, CallbackException callbackException = null)
        {
            BAOrderedResourceFilter filter = new BAOrderedResourceFilter(resourceType, atomPresetResourceType);
            filter.SetAllParamsToUserDefault();
            SelectResourceFilename(filter, vrBrowserPosition, callback, desktopDisplayMode, browserTitle, baKeepOpenOption, callbackFail, callbackException);
        }
        

        public static void SetMainResourceFilters(BAOrderedResourceFilter filter, int selectMode = BASelectModeEnum.single, int desktopDisplayMode = BADesktopDisplayModeEnum.noChange, CallbackFail callbackFail = null, CallbackException callbackException = null)
        {
            try
            {
                if (isBAenabled && isSetMainResourceFiltersAvailable)
                {
                    JSONStorableString callbackActionFail = new JSONStorableString("failCallback", "", delegate (string failureMessage)
                    {
                        if (enableFailureLogging) SuperController.LogMessage("BAInterop.SetMainResourceFilters: " + failureMessage);
                        if (callbackFail != null) callbackFail(failureMessage);
                    });

                    JSONStorableString callbackExceptionMessage = new JSONStorableString("exceptionCallback", "", delegate (string exceptionMessage)
                    {
                        if (callbackException != null) callbackException(exceptionMessage);
                        else SuperController.LogError("BAInterop.SetMainResourceFilters exception callback from BA: " + exceptionMessage);
                    });

                    var filterDict = new Dictionary<int, object>();
                    PackFilterDict(filter, filterDict);
                    filterDict.Add(selectModeKey, selectMode);

                    var objects = new object[4];
                    objects[0] = filterDict;
                    objects[1] = desktopDisplayMode;
                    objects[2] = callbackActionFail;
                    objects[3] = callbackExceptionMessage;

                    BAPlugin.SendMessage("OnSetMainResourceFilters", objects);
                }
                else if (callbackFail != null)
                {
                    string failureMessage = "A request to set resource filters in BrowswerAssist failed due to the plugin not being available.";
                    if (isBAenabled) failureMessage = "A request to set resource filters from BrowswerAssist failed due to the loaded BrowserAssist plugin not being the correct version 1.14.1 or higher.";
                    if (enableFailureLogging) SuperController.LogMessage("BAInterop.OnSetMainResourceFilters: " + failureMessage);
                    callbackFail(failureMessage);
                }
            }
            catch (Exception e)
            {
                if (callbackException != null) callbackException(e.ToString());
                else SuperController.LogError("BAInterop.OnSetMainResourceFilters exception: " + e);
            }
        }
        public static void SelectResourceFilename(BAOrderedResourceFilter filter, Vector3 vrBrowserPosition, GetStringCallback callback, int desktopDisplayMode = BADesktopDisplayModeEnum.rightDocked, string browserTitle = "Select Resource", int baKeepOpenOption=BAKeepOpenOptionEnum.unavailable, CallbackFail callbackFail = null, CallbackException callbackException = null)
        {
            try
            {
                if (isBAenabled && isSelectResourceFilenameAvailable)
                {
                    JSONStorableString callbackJSS = new JSONStorableString("successCallback", "*", delegate (string callbackValue)
                    {
                        callback(callbackValue);
                    });

                    JSONStorableString callbackActionFail = new JSONStorableString("failCallback", "", delegate (string failureMessage)
                    {
                        if (enableFailureLogging) SuperController.LogMessage("BAInterop.SelectResourceFilename: " + failureMessage);
                        if (callbackFail != null) callbackFail(failureMessage);
                    });

                    JSONStorableString callbackExceptionMessage = new JSONStorableString("exceptionCallback", "", delegate (string exceptionMessage)
                    {
                        if (callbackException != null) callbackException(exceptionMessage);
                        else SuperController.LogError("BAInterop.SelectResourceFilename exception callback from BA: " + exceptionMessage);
                    });

                    var filterDict = new Dictionary<int, object>();
                    PackFilterDict(filter, filterDict);

                    var objects = new object[8];
                    objects[0] = filterDict;
                    objects[1] = vrBrowserPosition;
                    objects[2] = callbackJSS;
                    objects[3] = callbackActionFail;
                    objects[4] = callbackExceptionMessage;
                    objects[5] = browserTitle;
                    objects[6] = desktopDisplayMode;
                    objects[7] = baKeepOpenOption;

                    BAPlugin.SendMessage("OnSelectResourceFilename", objects);
                }
                else if (callbackFail != null)
                {
                    string failureMessage = "A request to select a resource from BrowswerAssist failed due to the plugin not being available.";
                    if (isBAenabled) failureMessage = "A request to select a resource from BrowswerAssist failed due to the loaded BrowserAssist plugin not being the correct version 1.11 or higher.";
                    if (enableFailureLogging) SuperController.LogMessage("BAInterop.SelectResourceFilename: " + failureMessage);
                    callbackFail(failureMessage);
                }
            }
            catch (Exception e)
            {
                if (callbackException != null) callbackException(e.ToString());
                else SuperController.LogError("BAInterop.SelectResourceFilename exception: " + e);
            }
        }


        private static void PackFilterDict(BAResourceFilter filter, Dictionary<int, object> filterDict)
        {
            filterDict.Add(resourceTypeKey, filter.resourceType);
            filterDict.Add(atomPresetResourceNameKey, filter.atomPresetResourceType);
            filterDict.Add(tagGroupTypeKey, filter.tagGroupType);
            if (filter.tagGroupType == BATagGroupTypeEnum.userTag) filterDict.Add(userTagNameNameKey, filter.userTagName);
            filterDict.Add(showHiddenKey, filter.showHidden);
            filterDict.Add(onlyFavKey, filter.onlyFav);
            filterDict.Add(onlyNewKey, filter.onlyNew);
            filterDict.Add(creatorTypeKey, filter.creatorType);
            if (filter.creatorType == BACreatorTypeEnum.specificCreator) filterDict.Add(creatorNameKey, filter.creatorName);
            if (filter.creatorType == BACreatorTypeEnum.localFolders && filter.resourceType != BAResourceTypeEnum.femaleClothingItems && filter.resourceType != BAResourceTypeEnum.maleClothingItems)
            {
                filterDict.Add(localFolderKey, filter.localFolder);
                filterDict.Add(localFolderIncSubFolderKey, filter.includeSubFolders);
            }
            else if (filter.resourceType == BAResourceTypeEnum.femaleClothingItems || filter.resourceType == BAResourceTypeEnum.maleClothingItems)
            {
                filterDict.Add(allClothingCreatorsKey, filter.allClothingCreator);
                filterDict.Add(clothingCreatorNameKey, filter.clothingCreatorName);
            }
            if (filter.creatorType == BACreatorTypeEnum.allPackageCreators || filter.creatorType == BACreatorTypeEnum.specificCreator) filterDict.Add(varLicenseKey, filter.varLicenseType);
            filterDict.Add(resourceNameFilterModeKey, filter.resourceNameFilterMode);
            filterDict.Add(resourceNameKey, filter.resourceName);
        }
        private static void PackFilterDict(BAOrderedResourceFilter filter, Dictionary<int, object> filterDict)
        {
            PackFilterDict(filter as BAResourceFilter, filterDict);
            filterDict.Add(orderByKey, filter.resourceOrder);
        }

        protected const int resourceTypeKey = 1;
        protected const int tagGroupTypeKey = 2;
        protected const int userTagNameNameKey = 3;
        protected const int showHiddenKey = 4;
        protected const int onlyFavKey = 5;
        protected const int onlyNewKey = 6;
        protected const int creatorTypeKey = 7;
        protected const int creatorNameKey = 8;
        protected const int localFolderKey = 9;
        protected const int localFolderIncSubFolderKey = 10;
        protected const int varLicenseKey = 11;
        protected const int resourceNameFilterModeKey = 12;
        protected const int resourceNameKey = 13;
        protected const int orderByKey = 14;
        protected const int selectModeKey = 15;
        protected const int atomPresetResourceNameKey = 16;
        protected const int allClothingCreatorsKey = 17;
        protected const int clothingCreatorNameKey = 18;
    }

    public class BAOrderedResourceFilter : BAResourceFilter
    {
        public int resourceOrder { get; set; } = BAResultsSortModeEnum.AtoZ;
        public BAOrderedResourceFilter(int resourceType, string atomPresetResourceType) : base(resourceType, atomPresetResourceType)
        {

        }

        public override JSONClass GetJSON()
        {
            var jc = base.GetJSON();
            jc["resourceOrder"].AsInt = resourceOrder;

            return jc;
        }

        public override void RestoreJSON(JSONClass jc)
        {
            base.RestoreJSON(jc);
            if (jc["resourceOrder"] != null) resourceOrder = jc["resourceOrder"].AsInt;
        }

        public override void SetAllParamsToUserDefault()
        {
            base.SetAllParamsToUserDefault();
            resourceOrder = BAResultsSortModeEnum.userDefault;
        }
    }

    public class BAResourceFilter : BACreatorFilter
    {
        // 'creatorType' must always be set to one of the integer enumerations in 'BACreatorType'. If set to 'BACreatorType.userDefault' then the filter will be set to the BrowserAssist default value.
        public int creatorType { get; set; } = BACreatorTypeEnum.allLocalAndCreators;

        // 'creatorName' only needs to be set if creatorType is set to 'BACreatorType.specificCreator'. When set, must be one of the strings returned by BAInterop.GetCreatorNames()
        public string creatorName { get; set; } = "";

        // 'localFolder' only needs to be set if creatorType is set to 'BACreatorType.localFolders' and resourceType is not set to BAResourceTypeEnum.maleClothingItems or BAResourceTypeEnum.femaleClothingItems. When set, should be a folder within the VAM storage location for that resource type. If set to "", then then the root folder for that resource type will be used.
        public string localFolder { get; set; } = "";

        // 'includeSubFolders' only needs to be set if creatorType is set to 'BACreatorType.localFolders' and resourceType is not set to BAResourceTypeEnum.maleClothingItems or BAResourceTypeEnum.femaleClothingItems. When set, will indicate if resources with in sub folders of the local folder are included.
        public bool includeSubFolders { get; set; } = false;
        
        // 'allClothingCreator' only needs to be set if resourceType is set to BAResourceTypeEnum.maleClothingItems or BAResourceTypeEnum.femaleClothingItems and creatorType is NOT set to 'BACreatorType.allPackageCreators' or 'BACreatorType.specificCreator'
        public bool allClothingCreator { get; set; } = true;

        // 'clothingCreatorName' will filter clothing resources that have creators which dont match the specified string. It only needs to be set if resourceType is set to BAResourceTypeEnum.maleClothingItems or BAResourceTypeEnum.femaleClothingItems and creatorType is NOT set to 'BACreatorType.allPackageCreators' or 'BACreatorType.specificCreator' and allClothingCreator is set to false.
        public string clothingCreatorName { get; set; } = "";

        // 'varLicenseType' only needs to be set if creatorType is set to 'BACreatorType.allPackageCreators' or 'BACreatorType.specificCreator'. When set, should be one of the integer enumerations in 'BALicenseType'.
        public int varLicenseType { get; set; } = BALicenseTypeEnum.Any;

        // 'resourceNameFilterMode' will determine what filter search logic is applied when filtering by the 'resourceName' field.
        public int resourceNameFilterMode { get; set; } = BAResourceNameFilterModeEnum.ContainsFullPath;

        // 'resourceName' will filter resources that have names which dont match the specified string. The comparison logic is determined by the 'resourceNameFilterMode'. If 'resourceName' is set to an empty string (e.g. resourceName =""), then no resources will be filtered based on the resource's name.
        public string resourceName { get; set; } = "";

        public BAResourceFilter(int resourceType, string atomPresetResourceType) : base(resourceType, atomPresetResourceType)
        {

        }

        public override JSONClass GetJSON()
        {
            var jc = base.GetJSON();
            jc["creatorType"].AsInt = creatorType;
            jc["creatorName"] = creatorName;
            jc["localFolder"] = localFolder;
            jc["allClothingCreator"].AsBool = allClothingCreator;
            jc["clothingCreatorName"] = clothingCreatorName;
            jc["resourceName"] = resourceName;
            jc["varLicenseType"].AsInt = varLicenseType;
            jc["resourceNameFilterMode"].AsInt = resourceNameFilterMode;
            jc["resourceName"] = resourceName;

            return jc;
        }

        public override void RestoreJSON(JSONClass jc)
        {
            base.RestoreJSON(jc);
            if (jc["creatorType"] != null) creatorType = jc["creatorType"].AsInt;
            if (jc["creatorName"] != null) creatorName = jc["creatorName"];
            if (jc["localFolder"] != null) localFolder = jc["localFolder"];
            if (jc["includeSubFolders"] != null) includeSubFolders = jc["includeSubFolders"].AsBool;
            if (jc["clothingCreatorName"] != null) clothingCreatorName = jc["clothingCreatorName"];
            if (jc["allClothingCreator"] != null) allClothingCreator = jc["allClothingCreator"].AsBool;

            if (jc["varLicenseType"] != null) varLicenseType = jc["varLicenseType"].AsInt;
            if (jc["resourceNameFilterMode"] != null) resourceNameFilterMode = jc["resourceNameFilterMode"].AsInt;
            if (jc["resourceName"] != null) resourceName = jc["resourceName"];
        }

        public override void SetAllParamsToUserDefault()
        {
            base.SetAllParamsToUserDefault();
            creatorType = BACreatorTypeEnum.userDefault;
            creatorName = "";
            localFolder = "";
            allClothingCreator = true;
            clothingCreatorName = "";
            includeSubFolders = false;
            varLicenseType = BALicenseTypeEnum.userDefault;
            resourceNameFilterMode = BAResourceNameFilterModeEnum.userDefault;
            resourceName = "";
        }
    }

    public class BACreatorFilter
    {
        // 'resourceType' must always be set to one of the integer enumerations in 'BAResourceType'. It is set as part of the constructor of this class
        public int resourceType { get; protected set; } = BAResourceTypeEnum.scene;

        // 'atomPresetResourceType' must be set if resourceType is set to AtomPreset. It is set as part of the constructor of this class
        public string atomPresetResourceType { get; protected set; } = "";

        // 'tagGroupType' must always be set to one of the integer enumerations in 'BATagGroupType'. If set to 'BATagGroupType.userDefault' then the filter will be set to the BrowserAssist default value.
        public int tagGroupType { get; set; } = BATagGroupTypeEnum.anyTags;

        // 'userTagName' only needs to be set if tagGroupType is set to 'BATagGroupType.userTag'. When set, must be one of the strings returned by BAInterop.GetTagNames()
        public string userTagName { get; set; } = "";

        // 'showHidden' must always be set to one of the integer enumerations in 'BAShowHidden'. If set to 'BAShowHidden.userDefault' then the filter will be set to the BrowserAssist default value.
        public int showHidden { get; set; } = BAShowHiddenEnum.excludeHidden;

        // 'onlyFav' must always be set to one of the integer enumerations in 'BAOnlyFavorite'. If set to 'BAOnlyFavorite.userDefault' then the filter will be set to the BrowserAssist default value.
        public int onlyFav { get; set; } = BAOnlyFavoriteEnum.includeNotFavorites;

        // 'onlyNew' must always be set to one of the integer enumerations in 'BAOnlyNew'. If set to 'BAOnlyNew.userDefault' then the filter will be set to the BrowserAssist default value.
        public int onlyNew { get; set; } = BAOnlyNewEnum.includeNotNew;
        public BACreatorFilter(int resourceType, string atomPresetResourceType)
        {
            SetResourceType(resourceType,atomPresetResourceType);

        }

        public virtual void SetResourceType(int resourceType, string atomPresetResourceType)
        {
            this.resourceType = resourceType;

            this.atomPresetResourceType = atomPresetResourceType;

            if (resourceType == BAResourceTypeEnum.scene || resourceType == BAResourceTypeEnum.presetAppearance || resourceType == BAResourceTypeEnum.presetAnimation || resourceType == BAResourceTypeEnum.presetBreastPhys || resourceType == BAResourceTypeEnum.presetClothing || resourceType == BAResourceTypeEnum.presetGeneral || resourceType == BAResourceTypeEnum.presetGlutePhysics || resourceType == BAResourceTypeEnum.presetHair || resourceType == BAResourceTypeEnum.presetMorphs || resourceType == BAResourceTypeEnum.presetPlugins || resourceType == BAResourceTypeEnum.presetPose || resourceType == BAResourceTypeEnum.presetSkin) this.atomPresetResourceType = "Person";

        }

        public virtual JSONClass GetJSON()
        {
            var jc = new JSONClass();
            jc["resourceType"].AsInt = resourceType;
            jc["atomPresetResourceType"] = atomPresetResourceType;
            jc["tagGroupType"].AsInt = tagGroupType;
            jc["userTagName"] = userTagName;
            jc["showHidden"].AsInt = showHidden;
            jc["onlyFav"].AsInt = onlyFav;
            jc["onlyNew"].AsInt = onlyNew;

            return jc;
        }

        public virtual void RestoreJSON(JSONClass jc)
        {
            if (jc["resourceType"] !=null) resourceType = jc["resourceType"].AsInt;
            if (jc["atomPresetResourceType"] != null) atomPresetResourceType = jc["atomPresetResourceType"];
            else atomPresetResourceType = "";
            if (jc["tagGroupType"] != null) tagGroupType = jc["tagGroupType"].AsInt;
            if (jc["userTagName"] != null) userTagName = jc["userTagName"];
            if (jc["showHidden"] != null) showHidden = jc["showHidden"].AsInt;
            if (jc["onlyFav"] != null) onlyFav = jc["onlyFav"].AsInt;
            if (jc["onlyNew"] != null) onlyNew = jc["onlyNew"].AsInt;
        }

        public virtual void SetAllParamsToUserDefault()
        {
            tagGroupType = BATagGroupTypeEnum.userDefault;
            userTagName = "";
            showHidden = BAShowHiddenEnum.userDefault;
            onlyFav = BAOnlyFavoriteEnum.userDefault;
            onlyNew = BAOnlyNewEnum.userDefault;
        }
    }

    

    public static class BAResourceTypeEnum
    {
        public const int scene = 1;
        public const int presetAppearance = 2;
        public const int presetAnimation = 3;
        public const int presetBreastPhys = 4;
        public const int presetClothing = 5;
        public const int presetGeneral = 6;
        public const int presetGlutePhysics = 7;
        public const int presetHair = 8;
        public const int presetMorphs = 9;
        public const int presetPlugins = 10;
        public const int presetPose = 11;
        public const int presetSkin = 12;
        public const int personTexture = 13;
        public const int atomTexture = 14;
        public const int plugins = 15;
        public const int presetAtom = 16;
        public const int subScenes = 17;
        public const int femaleClothingItems = 18;
        public const int maleClothingItems = 19;
        public const int cua = 20;
        public const int clothingItemPresets =21;
        public const int presetSceneSessionPlugins = 22;
    }

    public static class BATagGroupTypeEnum
    {
        public const int userDefault = -1;
        public const int anyTags = 1;
        public const int hiddenOnly = 2;
        public const int noTags = 3;
        public const int noUserTags = 4;
        public const int wornClothingOnly = 5;
        public const int userTag = 100;
    }

    public static class BAShowHiddenEnum
    {
        public const int userDefault = -1;
        public const int includeHidden = 1;
        public const int excludeHidden = 2;
    }

    public static class BAOnlyFavoriteEnum
    {
        public const int userDefault = -1;
        public const int onlyFavorites = 1;
        public const int includeNotFavorites = 2;
    }

    public static class BAOnlyNewEnum
    {
        public const int userDefault = -1;
        public const int onlyNew = 1;
        public const int includeNotNew = 2;
    }

    public static class BACreatorTypeEnum
    {
        public const int userDefault = -1;
        public const int allPackageCreators = 1;
        public const int localFolders = 2;
        public const int allLocalAndCreators = 3;
        public const int specificCreator = 4;
    }

    public static class BAClothingCreatorTypeEnum
    {
        public const int userDefault = -1;
        public const int allCreators = 1;
    }

    public static class BALicenseTypeEnum
    {
        public const int userDefault = -1;
        public const int Unknown = 0;
        public const int Any = 1;
        public const int PC = 2;
        public const int PCEA = 3;
        public const int Questionable = 4;
        public const int FC = 5;
        public const int CCBY = 6;
        public const int CCBYSA = 7;
        public const int CCBYND = 8;
        public const int CCBYNC = 9;
        public const int CCBYNCSA = 10;
        public const int CCBYNCND = 11;

    }

    public static class BAResourceNameFilterModeEnum
    {
        public const int userDefault = -1;
        public const int StartsWith = 1;
        public const int Contains = 2;
        public const int ContainsFullPath = 3;
    }

    public static class BAResultsSortModeEnum
    {
        public const int userDefault = -1;
        public const int AtoZ = 1;
        public const int ZtoA = 2;
        public const int packageFolder = 3;
        public const int newDateTime = 4;
        public const int reverseNewDateTime = 5;
        public const int updatedDateTime = 6;
        public const int reverseUpdatedDateTime = 7;
    }

    public static class BASelectModeEnum
    {
        public const int userDefault = -1;
        public const int instant = 1;
        public const int single = 2;
        public const int multi = 3;
    }

    public static class BADesktopDisplayModeEnum
    {
        public const int userDefault = -1;
        public const int noChange = 0;
        public const int rightDocked = 1;
        public const int fullWidth = 2;
        public const int leftDocked = 3;
    }

    public static class BAKeepOpenOptionEnum
    {
        public const int unavailable = 1;
        public const int availableActive = 2;
        public const int availableInactive = 3;
    }
}
