﻿
using Leap.Unity;
using Leap.Unity.Query;
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using UnityEngine;

namespace JayJayWon
{
    public class VARPackageManifest
    {
        public static Dictionary<string, VARPackageManifestEntry> varPackages = new Dictionary<string, VARPackageManifestEntry>();

        public static Dictionary<string, VARPackageVersionGroupEntry> varPackageVersionGroups = new Dictionary<string, VARPackageVersionGroupEntry>();

        public static void Init(DateTime scanTime, HashSet<int> firstScanResourceTypes)
        {
            Globals.failureOnVARMetaDataLoad = false;
            DateTime varMetaFilesProcessCompleteTime=DateTime.Now;
            DateTime varUserFilesProcessCompleteTime = DateTime.Now;
            string[] varGroupCoreFiles;
            string[] varResourceCoreFiles;
            string[] varGroupUserFiles = null;
            string[] varResourceUserFiles = null;
            HashSet<string> varFilePaths = null;
            try
            {
                varGroupCoreFiles = FileManagerSecure.GetFiles(Const.VARGroupCoreDataFilesFolder, "*" + Const.CoreDataFileExt);
                varResourceCoreFiles = FileManagerSecure.GetFiles(Const.VARResourcesCoreDataFilesFolder, "*" + Const.CoreDataFileExt);
                varGroupUserFiles = FileManagerSecure.GetFiles(Const.VARGroupUserDataFilesFolder, "*" + Const.UserDataFileExt);
                varResourceUserFiles = FileManagerSecure.GetFiles(Const.VARResourcesUserDataFilesFolder, "*" + Const.UserDataFileExt);

                varFilePaths = Utils.GetFilesAtPathRecursive("AddonPackages", "*.var", true, true);

                DateTime varFilesScanCompleteTime = DateTime.Now;
                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" VAR files found in ", varFilePaths.Count, (varFilesScanCompleteTime - scanTime).TotalSeconds);

                Dictionary<string, string> varFileNamesToPaths = new Dictionary<string, string>();
                HashSet<string> varFileNames = new HashSet<string>();
                foreach (string varFilePath in varFilePaths)
                {
                    if (FileManagerSecure.DirectoryExists(varFilePath + ":"))
                    {
                        string varFileName = FileManagerSecure.GetFileName(varFilePath);
                        varFileNames.Add(varFileName);
                        varFileNamesToPaths.Add(varFileName, varFilePath);
                    }
                }

                ReloadVARManifest(varGroupCoreFiles, varResourceCoreFiles, varFileNames, scanTime, varFileNamesToPaths);
                varMetaFilesProcessCompleteTime = DateTime.Now;
                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" BA VAR Manifest files processed in ", varResourceCoreFiles.Count() + varGroupCoreFiles.Count(), (varMetaFilesProcessCompleteTime - varFilesScanCompleteTime).TotalSeconds);

                ReloadGroupVARUserData(varGroupUserFiles);
                ResourceManifest.ReloadResourceVARUserData(varResourceUserFiles);
                varUserFilesProcessCompleteTime = DateTime.Now;
                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" VAR User files processed in ", varResourceUserFiles.Count(), (varUserFilesProcessCompleteTime - varMetaFilesProcessCompleteTime).TotalSeconds);

            }
            catch (Exception e)
            {
                Globals.failureOnVARMetaDataLoad = true;
                DataFileManifest.CancelAllVARSaveRequests();
                SuperController.LogError("BA.VARPackageManifest.Init: "+e);
            }


            RefreshManifestWithNewVARs(varFilePaths, scanTime, firstScanResourceTypes);
            DateTime varFilesProcessCompleteTime = DateTime.Now;
            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" VAM VAR files processed in ", varFilePaths.Count(), (varFilesProcessCompleteTime - varUserFilesProcessCompleteTime).TotalSeconds);

            RefreshVARUserFilePrefs();
            DateTime vaFilePrefsProcessCompleteTime = DateTime.Now;
            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" VAR Prefs files processed in ", 0, (vaFilePrefsProcessCompleteTime - varFilesProcessCompleteTime).TotalSeconds);
        }

        public static string GetVPMEKey(string creatorName, string packageName, int varVersion)
        {
            return creatorName + "." + packageName + "." + varVersion.ToString();
        }
        public static string GetVPMEKey(VARPackageManifestEntry vpme)
        {
            return GetVPMEKey(vpme.creatorName, vpme.packageName, vpme.varVersion);


        }
        private static string GetVPMEKey(VARPackageVersionGroupEntry vpvge, int varVersion)
        {
            return GetVPMEKey(vpvge.creatorName, vpvge.packageName, varVersion);

        }

        public static VARPackageManifestEntry GetVPME(VARPackageVersionGroupEntry vpvge, int varVersion)
        {
            return varPackages[ GetVPMEKey(vpvge, varVersion)];
        }

        public static string GetVPVGEKey(string creatorName, string packgageName)
        {
            return creatorName + "." + packgageName;
        }
        public static string GetVPVGEKey(VARPackageManifestEntry vpme)
        {
            return GetVPVGEKey(vpme.creatorName, vpme.packageName);
        }

        public static string GetVARPath(PackagedRVGE packagedRVGE, int version)
        {
            string vpmeKey = GetVPMEKey(packagedRVGE.creatorName, packagedRVGE.packageName, version);
            if (varPackages.ContainsKey(vpmeKey))
            {
                return varPackages[vpmeKey].varFilePath;
            }
            return null;
        }
        public static void ReloadVARManifest(string[] varGroupCoreFileNames, string[] varResourceCoreFileNames, HashSet<string> varFileNames, DateTime scanTime, Dictionary<string, string> varFileNamesToPaths)
        {
            HashSet<string> varGroupCoreFileNamesToSave = new HashSet<string>();
            HashSet<string> varGroupCoreFileNamesToDelete = new HashSet<string>();

            foreach (string varGroupCoreFileName in varGroupCoreFileNames)
            {
                try
                {
                    JSONClass varGroupCoreFileJC = SuperController.singleton.LoadJSON(varGroupCoreFileName).AsObject;
                    if (varGroupCoreFileJC["VARManifestStoreFormat"].AsInt == Const.VARManifestStoreFormat)
                    {
                        JSONArray varGroupsJA = varGroupCoreFileJC["varGroups"].AsArray;

                        if (varGroupsJA.Count==0) varGroupCoreFileNamesToDelete.Add(varGroupCoreFileName);
                        else
                        {
                            foreach (JSONClass varGroupJC in varGroupsJA)
                            {
                                JSONArray versionsJA = varGroupJC["varVersions"].AsArray;

                                string creatorName = varGroupJC["creatorName"];
                                string packageName = varGroupJC["packageName"];
                                string vpvgeKey = GetVPVGEKey(creatorName, packageName);

                                HashSet<int> missingVARversions = new HashSet<int>();
                                JSONArray currentVARVersionsJA = new JSONArray();
                                foreach (JSONClass jc in versionsJA)
                                {
                                    int version = jc["varVersion"].AsInt;
                                    string varFileNameFromMetaFile = vpvgeKey + "." + version.ToString() + ".var";
                                    if (varFileNames.Contains(varFileNameFromMetaFile) || Settings.disableMetaPurgeOnMissingVARJSB.val)
                                    {
                                        currentVARVersionsJA.Add(jc);
                                        if (!varFileNames.Contains(varFileNameFromMetaFile)) missingVARversions.Add(version);
                                    }
                                    else varGroupCoreFileNamesToSave.Add(varGroupCoreFileName);
                                }

                                // Check for duplicate VPVGE keys and de-dupe where possible
                                if (varPackageVersionGroups.ContainsKey(vpvgeKey) && Globals.isFileCreateModifiedDateAvailable)
                                {
                                    var existingVPVGE = varPackageVersionGroups[vpvgeKey];
                                    var dataFile = existingVPVGE.coreDataFile;
                                    DateTime existingVPVGEModifiedTime = Utils.FileLastWriteTime(dataFile.fileName);
                                    DateTime currentVPVGEModifiedTime = Utils.FileLastWriteTime(varGroupCoreFileName);

                                    if (existingVPVGEModifiedTime < currentVPVGEModifiedTime)
                                    {
                                        dataFile.UnRegisterVARGroup(existingVPVGE);
                                        varPackageVersionGroups.Remove(vpvgeKey);
                                    }
                                    else continue;
                                }

                                if (!varPackageVersionGroups.ContainsKey(vpvgeKey))
                                {
                                    if (currentVARVersionsJA.Count > 0)
                                    {
                                        VARPackageVersionGroupEntry vpvge = new VARPackageVersionGroupEntry(creatorName, packageName, currentVARVersionsJA, scanTime, varFileNamesToPaths, missingVARversions);
                                        varPackageVersionGroups.Add(vpvgeKey, vpvge);
                                        DataFileManifest.RegisterExisitingDFCoreVARGroup(vpvge, varGroupCoreFileName);
                                        CreatorManifest.RegisterCreator(vpvge.creatorName);
                                    }
                                    else varGroupCoreFileNamesToSave.Add(varGroupCoreFileName);
                                }
                                else
                                {

                                    throw new Exception("Unexpected VPVG duplicate key: '" + vpvgeKey + "'. Conflicting VPVG is in " + varPackageVersionGroups[vpvgeKey].coreDataFile.fileName.ToString());
                                }
                            }

                        }

                    }
                    else varGroupCoreFileNamesToSave.Add(varGroupCoreFileName);
                }
                catch (Exception e) { throw new Exception("Exception caught in BA.VARPackageManifest.ReloadVARManifest whilst processing Group Meta File " + varGroupCoreFileName + ": " + e); }
            }
            

            HashSet<string> varResourceCoreFileNamesToSave = new HashSet<string>();
            HashSet<string> varResourceCoreFileNamesToDelete = new HashSet<string>();

            foreach (string varResourceCoreFileName in varResourceCoreFileNames)
            {
                try
                {
                    JSONClass varResourcesCoreFileJC = SuperController.singleton.LoadJSON(varResourceCoreFileName).AsObject;
                    if (varResourcesCoreFileJC["VARManifestStoreFormat"].AsInt == Const.VARManifestStoreFormat)
                    {

                        JSONArray varResourcesJA = varResourcesCoreFileJC["resources"].AsArray;

                        if (varResourcesJA.Count == 0) varResourceCoreFileNamesToDelete.Add(varResourceCoreFileName);
                        else
                        {
                            foreach (JSONClass varResourceJC in varResourcesJA)
                            {
                                string creatorName = varResourceJC["creatorName"];
                                string packageName = varResourceJC["packageName"];
                                string varGroupName = GetVPVGEKey(creatorName, packageName);
                                if (varPackageVersionGroups.ContainsKey(varGroupName))
                                {
                                    bool result = ResourceManifest.RegisterVARResourceFromManifest(varPackageVersionGroups[varGroupName], varResourceJC, varResourceCoreFileName);
                                    if (!result) varResourceCoreFileNamesToSave.Add(varResourceCoreFileName);
                                }
                                else varResourceCoreFileNamesToSave.Add(varResourceCoreFileName);
                            }

                        }

                    }
                    else varResourceCoreFileNamesToSave.Add(varResourceCoreFileName);

                }
                catch (Exception e) { throw new Exception("Exception in BA.VARPackageManifest.ReloadVARManifest for Resource Meta File " + varResourceCoreFileName + ": " + e); }
            }

            foreach (string varGroupCoreFileName in varGroupCoreFileNamesToSave) DataFileManifest.RegisterVARGroupCoreSaveRequest(varGroupCoreFileName);

            foreach (string varResourceCoreFileName in varResourceCoreFileNamesToSave) DataFileManifest.RegisterVARResourceCoreSaveRequest(varResourceCoreFileName);

            foreach (string varGroupCoreFileName in varGroupCoreFileNamesToDelete) FileManagerSecure.DeleteFile(varGroupCoreFileName);

            foreach (string varResourceCoreFileName in varResourceCoreFileNamesToDelete) FileManagerSecure.DeleteFile(varResourceCoreFileName);
        }

        public static void RescanPackages()
        {
            DateTime scanTime = DateTime.Now;
            SuperController.singleton.RescanPackages();
            HashSet<string> varFiles = Utils.GetFilesAtPathRecursive("AddonPackages", "*.var",true,true);
            RefreshMissingVARs(varFiles);
            RefreshManifestWithNewVARs(varFiles, scanTime, new HashSet<int>());
            RefreshVARUserFilePrefs();
        }

        public static void RescanAllVARResource()
        {
            RescanPackages();
            foreach (var vpvgeKVP in varPackageVersionGroups)
            {
                var vpvge = vpvgeKVP.Value;
                foreach (int varVersion in vpvge.varVersions)
                {
                    if (!vpvge.IsVARMissing(varVersion))
                    {
                        RescanVARResources(vpvge,varVersion, new HashSet<int>());
                    }
                }                
            }
            BrowserAssist.mainBrowserUI.RefreshFilterResults(true);
            if (BrowserAssist.callbackBrowserUI.browserFiltersUI.currentFilterSettings != null) BrowserAssist.callbackBrowserUI.RefreshFilterResults(true);
        }

        public static void RescanVARResources(VARPackageVersionGroupEntry vpvge, int varVersion, HashSet<int> firstScanResourceTypes)
        {
            VARPackageManifestEntry vpme = vpvge.GetVPME(varVersion);
            if (vpme.vpvge == null) vpme.vpvge = vpvge;

            if (FileManagerSecure.DirectoryExists(vpme.varFilePath + ":") || FileManagerSecure.DirectoryExists(vpme.varFilePath))
            {
                HashSet<string> contentFiles = Utils.GetFilesAtPathRecursive(vpme.varFilePath + ":", "*.*");
                HashSet<string> contentVARInternalPaths = new HashSet<string>();
                HashSet<string> existingFiles = vpvge.GetResourcePaths(varVersion);

                int length = vpme.varFilePath.Length + 2;

                HashSet<string> hideFiles = new HashSet<string>();
                HashSet<string> favFiles = new HashSet<string>();

                foreach (var contentFile in contentFiles)
                {
                    if (contentFile.EndsWith(".hide")) hideFiles.Add(contentFile);
                    else if (contentFile.EndsWith(".fav")) favFiles.Add(contentFile);

                    contentVARInternalPaths.Add(contentFile.Substring(length));
                }

                var removedFiles = existingFiles.Except(contentVARInternalPaths);
                foreach (var removedFile in removedFiles)
                {
                    try
                    {
                        var rvge = ResourceManifest.GetRVGE(ResourceManifest.GetRVGKey(vpme, removedFile)) as PackagedRVGE;
                        vpvge.PurgeRVGEVersion(rvge, varVersion);
                    }
                    catch(Exception ex) { SuperController.LogError("BA.VARPackageManifest.RescanVARResources: Error removing file "+removedFile+" - "+ex); }
                    
                }

                var newFiles = contentVARInternalPaths.Except(existingFiles);
                foreach (var newResource in newFiles)
                {
                    try
                    {
                        int resourceType = ResourceManifest.GetResourceType(newResource);
                        int resourceCategory = ResourceType.GetResourceCategory(resourceType);

                        bool forceNew = false;
                        if (resourceType != ResourceType.unknown)
                        {
                            if (firstScanResourceTypes.Contains(resourceType)) forceNew = true;

                            ResourceManifest.RegisterNewVARResourceOfType(vpme, newResource, resourceType, forceNew);
                        }
                    }
                    catch (Exception ex) { SuperController.LogError("BA.VARPackageManifest.RescanVARResources: Error scanning new file " + newResource + " - " + ex); }
                    
                }

                HashSet<string> csListFiles = new HashSet<string>(contentFiles.Where(fileName => fileName.EndsWith(".cslist") && fileName.Contains(":\\Custom\\Scripts")));
                ResourceManifest.ProcessVARCSListReferences(vpme, csListFiles);

                ResourceManifest.RegisterCreatorHideFlags(vpme, hideFiles);
                ResourceManifest.RegisterCreatorFavFlags(vpme, favFiles);

            }
            else
            {
                SuperController.LogMessage("BA: Missing VAR file detected '" + vpme.varFilePath+"'");
            }

        }

        public static void RefreshMissingVARs(HashSet<string> varFiles)
        {
            var existingVPMEKeys = new HashSet<string>();
            foreach (string varFile in varFiles)
            {
                try
                {
                    if (FileManagerSecure.DirectoryExists(varFile + ":") || FileManagerSecure.DirectoryExists(varFile))
                    {
                        VARPackageManifestEntry vpme = new VARPackageManifestEntry(varFile);
                        string vpmeKey = GetVPMEKey(vpme);
                        existingVPMEKeys.Add(vpmeKey);
                    }
              
                }
                catch(Exception e)
                {
                    SuperController.LogError("BA.VARPackageManifest.RefreshMissingVARs: Exception caught processing VAR file " + varFile);
                }
            }
            var varPackagesKeys =  new HashSet<string>( varPackages.Keys);
            var missingVPMEKeys = varPackagesKeys.Except(existingVPMEKeys);
            foreach (string vpmeKey in missingVPMEKeys)
            {
                var vpme = varPackages[vpmeKey];
                string vpvgeKey = GetVPVGEKey(vpme);
                varPackageVersionGroups[vpvgeKey].SetVARMissing(vpme.varVersion, true);
                varPackages.Remove(vpmeKey);
            }
        }

        public static void RefreshManifestWithNewVARs(HashSet<string> varFiles, DateTime scanTime, HashSet<int> firstScanResourceTypes)
        {
            bool firstScan = firstScanResourceTypes.Count> 0;
            foreach (string varFile in varFiles)
            {
                try
                {
                    if (FileManagerSecure.DirectoryExists(varFile + ":") || FileManagerSecure.DirectoryExists(varFile))
                    {
                        VARPackageManifestEntry vpme = new VARPackageManifestEntry(varFile);
                        string vpmeKey = GetVPMEKey(vpme);

                        bool existingVAR = varPackages.ContainsKey(vpmeKey);
                        string vpvgeKey = GetVPVGEKey(vpme.creatorName, vpme.packageName);

                        if (!existingVAR)
                        {
                            varPackages.Add(vpmeKey, vpme);
                            if (!varPackageVersionGroups.ContainsKey(vpvgeKey) || !varPackageVersionGroups[vpvgeKey].varVersionLicenses.Keys.Contains(vpme.varVersion))
                            {
                                ProcessNewVARFile(vpme, scanTime, firstScanResourceTypes);
                                vpme.vpvge = varPackageVersionGroups[vpvgeKey];
                            }
                            else
                            {
                                // VAR that has been scanned previously
                                var vpge = varPackageVersionGroups[vpvgeKey];
                                vpme.vpvge = vpge;
                                vpge.SetVARMissing(vpme.varVersion, false);
                                if (firstScan) RescanVARResources(vpge, vpme.varVersion, firstScanResourceTypes);
                            }
                            
                        }
                        else
                        {
                            varPackages[vpmeKey] = vpme;
                        }

                        
                    }
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.VARPackageManifest.RefreshManifestWithNewVARs for " + varFile + ": " + e); }
            }
        }

        public static void PurgeVARFile(string creatorName, string packageName, int varVersion)
        {
            string vpmeKey = GetVPMEKey(creatorName, packageName, varVersion);
            if (varPackages.ContainsKey(vpmeKey))
            {
                string vpvgeKey = GetVPVGEKey(creatorName, packageName);
                if (varPackageVersionGroups.ContainsKey(vpvgeKey))
                {
                    var vpvge = varPackageVersionGroups[vpvgeKey];
                    vpvge.PurgeVARVersion(varVersion);

                    if (vpvge.versionsCount == 0)
                    {
                        varPackageVersionGroups.Remove(vpvgeKey);
                        vpvge.coreDataFile.UnRegisterVARGroup(vpvge);
                        vpvge.userDataFile.UnRegisterVARGroup(vpvge);
                    }
                }
                varPackages.Remove(vpmeKey);
            }
        }

        public static void ReloadGroupVARUserData(string[] varGroupUserFiles)
        {
            HashSet<string> varGroupUserFileNamesToSave = new HashSet<string>();
            HashSet<string> varGroupUserFileNamesToDelete = new HashSet<string>();

            foreach (string varGroupUserFileName in varGroupUserFiles)
            {
                try
                {
                    JSONClass jc = SuperController.singleton.LoadJSON(varGroupUserFileName).AsObject;

                    JSONArray varGroupsJA = jc["varGroups"].AsArray;

                    if (varGroupsJA.Count==0) varGroupUserFileNamesToDelete.Add(varGroupUserFileName);
                    else
                    {
                        foreach (JSONClass varGroupJC in varGroupsJA)
                        {
                            string creatorName = varGroupJC["creatorName"];
                            string packageName = varGroupJC["packageName"];

                            string vpvgeKey = GetVPVGEKey(creatorName, packageName);

                            if (varPackageVersionGroups.ContainsKey(vpvgeKey))
                            {
                                VARPackageVersionGroupEntry vpvge = varPackageVersionGroups[vpvgeKey];
                                DataFileManifest.RegisterExisitingDFUserVARGroup(vpvge, varGroupUserFileName);

                                JSONArray varVersionJA = varGroupJC["varVersions"].AsArray;

                                JSONArray currentVARVersionsJA = new JSONArray();
                                foreach (JSONClass varVersionJC in varVersionJA)
                                {
                                    int version = varVersionJC["varVersion"].AsInt;
                                    if (vpvge.varVersionLicenses.Keys.Contains(version)) currentVARVersionsJA.Add(varVersionJC);
                                    else varGroupUserFileNamesToSave.Add(varGroupUserFileName);
                                }
                                vpvge.ReloadVARVersionsUserData(currentVARVersionsJA);
                            }
                            else varGroupUserFileNamesToSave.Add(varGroupUserFileName);
                        }
                    }                      
                }
                catch (Exception e) { throw new Exception("Exception in BA.VARPackageManifest.ReloadGroupVARUserData for " + varGroupUserFileName + ": " + e); }
            }

            foreach (string varGroupUserFileNameToSave in varGroupUserFileNamesToSave) DataFileManifest.RegisterVARGroupUserSaveRequest(varGroupUserFileNameToSave);

            foreach (string varGroupUserFileNameToDelete in varGroupUserFileNamesToDelete) FileManagerSecure.DeleteFile(varGroupUserFileNameToDelete);
        }


        public static void RefreshVARUserFilePrefs()
        {
            ResourceManifest.ClearAllUserFilePrefs();
            foreach (var vpme in varPackages)
            {
                if (FileManagerSecure.DirectoryExists("AddonPackagesFilePrefs\\" + vpme.Value.varFileNameNoExtension))
                {
                    string searchFolder = "AddonPackagesFilePrefs\\" + vpme.Value.varFileNameNoExtension;
                    HashSet<string> varFilePrefs = Utils.GetFilesAtPathRecursive(searchFolder, "*.*");
                    foreach (string varFilePref in varFilePrefs)
                    {
                        try
                        {
                            int fileExtIndex = varFilePref.LastIndexOf('.');
                            int filePathLength = varFilePref.Length;

                            string rvgeKey = ResourceManifest.GetRVGKey(vpme.Value, varFilePref.Substring(searchFolder.Length + 1, fileExtIndex - searchFolder.Length - 1));
                            if (varFilePref.EndsWith(".hide")) ResourceManifest.UpdateVARUserHideFlag(rvgeKey, vpme.Value.varVersion, true);
                            if (varFilePref.EndsWith(".fav")) ResourceManifest.UpdateVARUserFavFlag(rvgeKey, vpme.Value.varVersion, true);
                        }
                        catch (Exception e) { SuperController.LogError("Exception in BA.RefreshVARUserFilePrefs for VAR " + vpme.Value.varFileNameNoExtension + "and file " + varFilePref + ": " + e); }

                    }
                }

            }
        }

        private static void ProcessNewVARFile(VARPackageManifestEntry vpme, DateTime scanTime, HashSet<int> firstScanResourceTypes)
        {
            string vpvgeKey = GetVPVGEKey(vpme);
            VARPackageVersionGroupEntry vpvge;
            if (!varPackageVersionGroups.ContainsKey(vpvgeKey))
            {
                vpvge = new VARPackageVersionGroupEntry(vpme, scanTime);
                varPackageVersionGroups.Add(vpvgeKey, vpvge);
                CreatorManifest.RegisterCreator(vpvge.creatorName);
                DataFileManifest.RegisterNewDFCoreVARGroup(vpvge);
            }
            else
            {
                vpvge = varPackageVersionGroups[vpvgeKey];
                if (vpvge.varVersionLicenses.Keys.Contains(vpme.varVersion)) return;
                vpvge.AddVARVersion(vpme.varVersion, vpme.varFilePath, scanTime);
            }

            vpvge.SetPersistVARCoreData();
            vpvge.SetPersistVARGroupUserData();
            vpme.vpvge = vpvge;
            ResourceManifest.RegisterNewVARResources(vpme, firstScanResourceTypes);

        }
    }

    public class VARPackageManifestEntry
    {
        public string varFilePath { get; private set; }
        public string varFileNameNoExtension { get; private set; }
        public string creatorName { get; private set; }
        public string packageName { get; private set; }
        public int varVersion { get; private set; }

        public VARPackageVersionGroupEntry vpvge { get; set; }

        public VARPackageManifestEntry(string varFile)
        {
            varFilePath = varFile;
            string varFileName = FileManagerSecure.GetFileName(varFilePath);

            if (!varFileName.EndsWith(".var")) { throw new Exception("BA.VARPackageManifestEntry.contructor: VAR Package '"+varFileName+"' doesnt end with .var"); }
            varFileNameNoExtension = varFileName.Substring(0, varFileName.Length - 4);

            string[] parts = varFileNameNoExtension.Split('.');

            if (parts.Length !=3 ) { throw new Exception("BA.VARPackageManifestEntry.contructor: VAR Package name '"+varFileName+"' doesnt contain 3 parts"); }
            creatorName = parts[0];
            packageName = parts[1];

            int tempVarVersion;
            bool validVer = int.TryParse(parts[2], out tempVarVersion);
            if (!validVer) { throw new Exception("BA.VARPackageManifestEntry.constructor: VAR Package name '"+varFileName+"' has an invalid version format"); }
            else varVersion = tempVarVersion;
        }

        public HashSet<string> resourcePaths { get
            {
                if (vpvge == null)
                {
                    if (VARPackageManifest.varPackageVersionGroups.ContainsKey(VARPackageManifest.GetVPVGEKey(this)))
                    {
                        vpvge = VARPackageManifest.varPackageVersionGroups[VARPackageManifest.GetVPVGEKey(this)];
                    }
                    else return null;
                }
                return vpvge.GetResourcePaths(varVersion);
            }

        }
    }
    public class VARPackageVersionGroupEntry
    {
        public string creatorName { get; private set; }
        public string packageName { get; private set; }

        public string packageUID
        {
            get { return creatorName + "." + packageName; }
        }

        public string varGroupName { get { return creatorName + "." + packageName; } }

        private HashSet<int> _varVersions = new HashSet<int>();

        public HashSet<int> varVersions
        {
            get
            {
                return new HashSet<int>(_varVersions);
            }
        }
        public bool IsVARMissing(int varVersion)
        {
            return missingVARVersions.Contains(varVersion);
        }
        private HashSet<int> missingVARVersions = new HashSet<int>();
        public Dictionary<int, int> varVersionLicenses = new Dictionary<int, int>();

        private Dictionary<int, DateTime> varVersionFirstScanTimes = new Dictionary<int, DateTime>();
        private Dictionary<int, DateTime> varVersionFirstCacheTimes = new Dictionary<int, DateTime>();

        private Dictionary<int, DateTime> varVersionCreatedTimes = new Dictionary<int, DateTime>();
        private Dictionary<int, DateTime> varVersionModifiedTimes = new Dictionary<int, DateTime>();

        private VARGroupUserDataFile _userDataFile = null;

        public VARGroupCoreDataFile coreDataFile { get; private set; }
        public VARGroupUserDataFile userDataFile {
            get
            {
                if (_userDataFile == null) DataFileManifest.RegisterNewDFUserVARGroup(this);
                return _userDataFile;
            }
            private set { _userDataFile = value; } }

        public DateTime firstCataloguedDateTime
        {
            get
            {
                DateTime firstScanDateTime = varVersionFirstScanTimes.Values.Min(d => d);
                if (Globals.isFileCreateModifiedDateAvailable)
                {
                    if (varVersionCreatedTimes.Count == 0) return firstScanDateTime;
                    DateTime earliestCreateDate = varVersionCreatedTimes.Values.Min(d => d);
                    return new DateTime(Math.Min(firstScanDateTime.Ticks, earliestCreateDate.Ticks));
                }
                else
                {
                    if (varVersionFirstCacheTimes.Count == 0) return firstScanDateTime;
                    DateTime firstCacheDateTime = varVersionFirstCacheTimes.Values.Min(d => d);
                    return new DateTime(Math.Min(firstScanDateTime.Ticks, firstCacheDateTime.Ticks));
                }

            }
        }
        public DateTime lastCataloguedDateTime
        {
            get
            {
                DateTime lastScanDateTime = varVersionFirstScanTimes[highestVARVersion];

                if (Globals.isFileCreateModifiedDateAvailable)
                {
                    if (!varVersionCreatedTimes.ContainsKey(highestVARVersion)) return lastScanDateTime;
                    DateTime createDateOfHighestVAR = varVersionCreatedTimes[highestVARVersion];
                    return new DateTime(Math.Min(lastScanDateTime.Ticks, createDateOfHighestVAR.Ticks));
                }
                else
                {
                    if (!varVersionFirstCacheTimes.ContainsKey(highestVARVersion)) return lastScanDateTime;
                    DateTime lastCacheDateTime = varVersionFirstCacheTimes[highestVARVersion];
                    return new DateTime(Math.Min(lastScanDateTime.Ticks, lastCacheDateTime.Ticks));
                }

            }
        }


        public int highestVARVersion { get; private set; }
        public int versionsCount { get { return _varVersions.Count; } }

        private HashSet<PackagedRVGE> packageRVGEs = new HashSet<PackagedRVGE>();
        private Dictionary<string, PackagedRVGE> packageRVGEDict = new Dictionary<string, PackagedRVGE>();

        public static bool anyPersistVARMetaDataPending { get; private set; } = false;
        public static bool anyPersistVARUserDataPending { get; private set; } = false;

        public VARPackageVersionGroupEntry(VARPackageManifestEntry vpme, DateTime scanTime)
        {
            creatorName = vpme.creatorName;
            packageName = vpme.packageName;

            _varVersions.Add(vpme.varVersion);
            varVersionLicenses.Add(vpme.varVersion, GetLicenseTypeFromMeta(vpme.varFilePath));
            varVersionFirstScanTimes.Add(vpme.varVersion, scanTime);
            highestVARVersion = vpme.varVersion;

            CheckFirstCachedDate(vpme.varVersion);

            if (Globals.isFileCreateModifiedDateAvailable)
            {
                var createdTime = Utils.FileCreationTime(vpme.varFilePath);
                varVersionCreatedTimes[vpme.varVersion] = createdTime;
                var modifiedTime = Utils.FileLastWriteTime(vpme.varFilePath);
                varVersionModifiedTimes[vpme.varVersion] = modifiedTime;
            }

        }

        public VARPackageVersionGroupEntry(string _creatorName, string _packageName, JSONArray varVersionsJA, DateTime scanTime, Dictionary<string, string> varFileNamesToPaths, HashSet<int> currentMissingVARVersions)
        {
            creatorName = _creatorName;
            packageName = _packageName;
            highestVARVersion = 0;
            foreach (JSONClass jc in varVersionsJA)
            {
                int varVersion = jc["varVersion"].AsInt;
                string varFileNameFromMetaFile = varGroupName + "." + varVersion.ToString() + ".var";
                if (varFileNamesToPaths.ContainsKey(varFileNameFromMetaFile))
                {
                    string varFilePath = varFileNamesToPaths[varFileNameFromMetaFile];
                    if (Globals.isFileCreateModifiedDateAvailable && FileManagerSecure.FileExists(varFilePath))
                    {
                        var createdTime = Utils.FileCreationTime(varFilePath);
                        varVersionCreatedTimes[varVersion] = createdTime;
                        var modifiedTime = Utils.FileLastWriteTime(varFilePath);
                        varVersionModifiedTimes[varVersion] = modifiedTime;
                    }

                }

                _varVersions.Add(varVersion);
                if (currentMissingVARVersions.Contains(varVersion)) missingVARVersions.Add(varVersion);
                varVersionLicenses.Add(varVersion, EnumManifest.GetEnumValFromStoreVal(LicenseType.enumManifestName, jc["varLicense"]));
                varVersionFirstScanTimes.Add(varVersion, scanTime);

                if (highestVARVersion < varVersion) highestVARVersion = varVersion;
            }

        }

        public VARPackageManifestEntry GetVPME(int varVersion)
        {
            return VARPackageManifest.GetVPME(this, varVersion);
        }

        public bool AnyExistingVARVersions(JSONArray varResourcesVersionsJA)
        {
            foreach (JSONNode versionJN in varResourcesVersionsJA) if (_varVersions.Contains(versionJN.AsInt)) return true;
            return false;
        }

        public bool AnyMismatchedVARVersions(JSONArray varResourcesVersionsJA)
        {
            foreach (JSONNode versionJN in varResourcesVersionsJA) if (!_varVersions.Contains(versionJN.AsInt)) return true;
            return false;
        }

        public void RegisterCoreDataFile(VARGroupCoreDataFile coreDataFile)
        {
            this.coreDataFile = coreDataFile;
        }

        public void RegisterUserDataFile(VARGroupUserDataFile userDataFile)
        {
            this.userDataFile = userDataFile;
        }

        private void CheckFirstCachedDate(int varVersion)
        {
            string cacheMetaFileName = "Cache\\PackageJSON\\" + varGroupName + "." + varVersion.ToString() + ".vamcachemeta";
            DateTime? firstCachedDateTime = null;

            if (FileManagerSecure.FileExists(cacheMetaFileName))
            {
                JSONClass cacheMetaFileJC = SuperController.singleton.LoadJSON(cacheMetaFileName).AsObject;
                if (cacheMetaFileJC["date"] != null) firstCachedDateTime = DateTime.Parse(cacheMetaFileJC["date"], new CultureInfo("en-US", false));
                SetPersistVARGroupUserData();
            }
            if (firstCachedDateTime != null)
            {
                if (varVersionFirstCacheTimes.ContainsKey(varVersion)) varVersionFirstCacheTimes[varVersion] = (DateTime)firstCachedDateTime;
                else varVersionFirstCacheTimes.Add(varVersion, (DateTime)firstCachedDateTime);
            }
        }

        private int GetLicenseTypeFromMeta(string varFileName)
        {
            int varLicenseType = LicenseType.Unknown;
            if (FileManagerSecure.FileExists(varFileName + ":\\meta.json"))
            {
                JSONClass varMetaJSON = null;
                try
                {
                    varMetaJSON = SuperController.singleton.LoadJSON(varFileName + ":\\meta.json").AsObject;
                }
                catch (Exception e) { SuperController.LogError("BA.VARPackageVersionGroupEntry.GetLicenseTypeFromMeta: Error loading meta.json in '" + varFileName + "'. Possibly malformed JSON format. Exception message - " + e); }

                if (varMetaJSON != null && varMetaJSON["licenseType"] != null)
                {
                    string varLicenseTypeStoreVal = varMetaJSON["licenseType"];
                    if (EnumManifest.GetStoreValEnumsDict(LicenseType.enumManifestName).Keys.Contains(varLicenseTypeStoreVal))
                    {
                        varLicenseType = EnumManifest.GetEnumValFromStoreVal(LicenseType.enumManifestName, varLicenseTypeStoreVal);
                    }
                    else
                    {
                        SuperController.LogMessage(varFileName + " contains unrecognised License Type. Possibly caused by creator manually editing VAR meta.json");
                    }

                }
            }
            else SuperController.LogMessage("VARPackageVersionGroupEntry.GetLicenseTypeFromMeta: Meta file not found in: " + varFileName);
            return varLicenseType;
        }
        public void SetPersistVARCoreData()
        {
            DataFileManifest.RegisterVARGroupCoreSaveRequest(coreDataFile);
        }

        public void SetPersistVARGroupUserData()
        {
            DataFileManifest.RegisterVARGroupUserSaveRequest(userDataFile);
        }

        public void AddVARVersion(int varVersion, string varFilePath, DateTime scanTime)
        {
            _varVersions.Add(varVersion);
            varVersionLicenses.Add(varVersion, GetLicenseTypeFromMeta(varFilePath));
            if (varVersion > highestVARVersion) highestVARVersion = varVersion;
            varVersionFirstScanTimes.Add(varVersion, scanTime);
            CheckFirstCachedDate(varVersion);
        }

        public void SetVARMissing(int varVersion, bool missing)
        {
            if (missing) missingVARVersions.Add(varVersion);
            else missingVARVersions.Remove(varVersion);
        }

        public void UnhideResources(HashSet<ResourceVersionGroupEntry> hiddenRVGEs, int varVersion)
        {
            foreach (PackagedRVGE rvge in packageRVGEs)
            {
                if (!hiddenRVGEs.Contains(rvge)) rvge.SetHideCreatorVARVersion(varVersion, false);
            }
        }

        public void UnfavResource(HashSet<ResourceVersionGroupEntry> favRVGEs, int varVersion)
        {
            foreach (PackagedRVGE rvge in packageRVGEs)
            {
                if (!favRVGEs.Contains(rvge)) rvge.SetFavCreatorVARVersion(varVersion, false);
            }
        }

        public void PurgeRVGEVersion(PackagedRVGE rvge, int varVersion)
        {
            rvge.PurgeVARVersion(varVersion);
            if (rvge.versionsCount == 0)
            {
                ResourceManifest.PurgePackageRVGE(rvge);
                string rvgeKey = ResourceManifest.GetRVGKey(rvge);
                packageRVGEDict.Remove(rvgeKey);
                packageRVGEs.Remove(rvge);
                rvge.SetUserDataPersist();
                rvge.coreDataFile.UnRegisterVARResource(rvge);
                rvge.userDataFile.UnRegisterVARResource(rvge);
            }
            rvge.SetCoreDataPersist();
        }

        public void PurgeVARVersion(int varVersion)
        {
            HashSet<PackagedRVGE> tempPackageRVGEs = new HashSet<PackagedRVGE>(packageRVGEs);
            foreach (PackagedRVGE rvge in tempPackageRVGEs)
            {
                PurgeRVGEVersion(rvge, varVersion);
            }
            _varVersions.Remove(varVersion);
            missingVARVersions.Remove(varVersion);
            varVersionFirstCacheTimes.Remove(varVersion);
            varVersionFirstScanTimes.Remove(varVersion);
            varVersionLicenses.Remove(varVersion);
            if (highestVARVersion == varVersion)
            {
                highestVARVersion = -1;
                foreach (int remainingVersion in _varVersions)
                {
                    if (remainingVersion > highestVARVersion) highestVARVersion = remainingVersion;
                }
            }

            SetPersistVARCoreData();
            SetPersistVARGroupUserData();
        }

        public void RegisterRVGE(PackagedRVGE packagedRVGE)
        {
            packageRVGEs.Add(packagedRVGE);
            packageRVGEDict.Add(ResourceManifest.GetRVGKey(this, packagedRVGE.resourceFolderPathFileName), packagedRVGE);
        }

        private void ReloadVARVersionUserData(JSONClass varVersionJC, int varVersion)
        {
            DateTime firstScanTime = DateTime.Parse(varVersionJC["firstScanDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);
            if (varVersionFirstScanTimes.ContainsKey(varVersion)) varVersionFirstScanTimes[varVersion] = firstScanTime;
            else varVersionFirstScanTimes.Add(varVersion, firstScanTime);

            string cacheMetaFileName = "Cache\\PackageJSON\\" + varGroupName + "." + varVersion.ToString() + ".vamcachemeta";
            if (varVersionJC["firstCacheDateTime"] != null)
            {
                DateTime firstCachedDateTime = DateTime.Parse(varVersionJC["firstCacheDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);
                if (varVersionFirstCacheTimes.ContainsKey(varVersion)) varVersionFirstCacheTimes[varVersion] = firstCachedDateTime;
                else varVersionFirstCacheTimes.Add(varVersion, firstCachedDateTime);

            }
            else CheckFirstCachedDate(varVersion);
        }
        public void ReloadVARVersionsUserData(JSONArray varVersionsJA)
        {
            HashSet<int> previouslyScannedVARVersions = new HashSet<int>();

            foreach (JSONClass varVersionJC in varVersionsJA)
            {
                int varVersion = varVersionJC["varVersion"].AsInt;
                previouslyScannedVARVersions.Add(varVersion);
                ReloadVARVersionUserData(varVersionJC, varVersion);
            }

            HashSet<int> newVARVersions = new HashSet<int>(_varVersions.Except(previouslyScannedVARVersions));
            if (newVARVersions.Count > 0)
            {
                foreach (ResourceVersionGroupEntry rvge in packageRVGEs)
                {
                    if (!rvge.localFileResource) (rvge as PackagedRVGE).NewVARVersions(newVARVersions);
                }

            }
        }
        public void ReloadLegacyVARUserData(JSONClass varUserDataJC)
        {
            int varUserDataStoreFormat = varUserDataJC["VARUserDataStoreFormat"].AsInt;
            JSONArray varVersionsJA = varUserDataJC["varVersions"].AsArray;
            JSONArray resourcesJA = varUserDataJC["resources"].AsArray;

            HashSet<int> previouslyScannedVARVersions = new HashSet<int>();

            if (varUserDataStoreFormat == 1)
            {
                foreach (JSONNode varVersionJC in varVersionsJA)
                {
                    int varVersion = varVersionJC.AsInt;
                    previouslyScannedVARVersions.Add(varVersion);
                }
                foreach (JSONClass resourceJC in resourcesJA)
                {
                    JSONArray resourceVARVersionsJA = resourceJC["varVersions"].AsArray;
                    foreach (JSONClass resourceVARVersionJC in resourceVARVersionsJA)
                    {
                        int varVersion = resourceVARVersionJC["varVersion"].AsInt;
                        DateTime firstScanTime = DateTime.Parse(resourceVARVersionJC["firstScanDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);

                        if (varVersionFirstScanTimes.ContainsKey(varVersion)) varVersionFirstScanTimes[varVersion] = firstScanTime;
                        else varVersionFirstScanTimes.Add(varVersion, firstScanTime);
                    }
                }
                SetPersistVARGroupUserData();
            }
            else
            {
                foreach (JSONClass varVersionJC in varVersionsJA)
                {
                    int varVersion = varVersionJC["varVersion"].AsInt;
                    previouslyScannedVARVersions.Add(varVersion);

                    ReloadVARVersionUserData(varVersionJC, varVersion);
                }
            }

            HashSet<int> newVARVersions = new HashSet<int>(_varVersions.Except(previouslyScannedVARVersions));
            if (newVARVersions.Count > 0)
            {
                foreach (ResourceVersionGroupEntry rvge in packageRVGEs)
                {
                    if (!rvge.localFileResource) (rvge as PackagedRVGE).NewVARVersions(newVARVersions);
                }

            }

            foreach (JSONClass resourceJC in resourcesJA)
            {
                string rvgeKey = ResourceManifest.GetRVGKey(creatorName, packageName, resourceJC["resourceFullFileName"]);
                if (packageRVGEDict.ContainsKey(rvgeKey))
                {
                    ResourceVersionGroupEntry rvge = packageRVGEDict[rvgeKey];
                    if (!rvge.localFileResource)
                    {
                        (rvge as PackagedRVGE).RestoreVARUserData(resourceJC);
                        (rvge as PackagedRVGE).SetUserDataPersist();
                    }
                }
            }

        }
        public JSONClass GetVARGroupUserDataJSON()
        {
            JSONClass varJC = new JSONClass();
            varJC["creatorName"] = creatorName;
            varJC["packageName"] = packageName;

            JSONArray varVersionsJA = new JSONArray();
            varJC["varVersions"] = varVersionsJA;
            foreach (var varVersionFirstScanTime in varVersionFirstScanTimes)
            {
                JSONClass varVersionJC = new JSONClass();
                varVersionJC["varVersion"].AsInt = varVersionFirstScanTime.Key;
                varVersionJC["firstScanDateTime"] = varVersionFirstScanTime.Value.ToString("O");
                if (varVersionFirstCacheTimes.ContainsKey(varVersionFirstScanTime.Key))
                {
                    varVersionJC["firstCacheDateTime"] = varVersionFirstCacheTimes[varVersionFirstScanTime.Key].ToString("O");
                }

                varVersionsJA[-1] = varVersionJC;
            }

            return varJC;
        }

        public JSONClass GetVARGroupMetaDataJSON()
        {
            JSONClass jc = new JSONClass();


            jc["creatorName"] = creatorName;
            jc["packageName"] = packageName;
            //            jc["varGroupName"] = varGroupName;
            JSONArray varVersionsJA = new JSONArray();
            jc["varVersions"] = varVersionsJA;
            foreach (var varVersion in varVersionLicenses)
            {
                JSONClass varVersionJC = new JSONClass();
                varVersionJC["varVersion"].AsInt = varVersion.Key;
                varVersionJC["varLicense"] = EnumManifest.GetStoreVal(LicenseType.enumManifestName, varVersion.Value);
                varVersionsJA[-1] = varVersionJC;
            }
            return jc;
        }

        public HashSet<string> GetResourcePaths(int varVersion)
        {
            HashSet<string> resourcePaths = new HashSet<string>();

            foreach (var rvge in packageRVGEs)
            {
                if (rvge.IsInVersion(varVersion)) resourcePaths.Add(rvge.resourceFolderPathFileName);
            }

            return resourcePaths;
        }
    }

    public class ResourceManifest
    {
        private static Dictionary<string, ResourceVersionGroupEntry> resourceVersionGroupsDict = new Dictionary<string, ResourceVersionGroupEntry>();

        private static HashSet<ResourceVersionGroupEntry> resourceVersionGroups = new HashSet<ResourceVersionGroupEntry>();

        private static HashSet<LocalRVGE> localResourceVersionGroups = new HashSet<LocalRVGE>();
        private static HashSet<PackagedRVGE> varResourceVersionGroups = new HashSet<PackagedRVGE>();

        private static Dictionary<int, HashSet<ResourceVersionGroupEntry>> rvgeByResourceType = new Dictionary<int, HashSet<ResourceVersionGroupEntry>>();

        private static Dictionary<string, HashSet<ResourceVersionGroupEntry>> rvgeByAtomPresetResourceType = new Dictionary<string, HashSet<ResourceVersionGroupEntry>>();


        private static HashSet<ResourceVersionGroupEntry> cupUpdates = new HashSet<ResourceVersionGroupEntry>();

        private static HashSet<ResourceVersionGroupEntry> cachePrimaryClothingItemPresetRVGEUpdates = new HashSet<ResourceVersionGroupEntry>();

        public static void Update()
        {
            foreach (var pcRVGE in cupUpdates)
            {
                pcRVGE.PersistCUP();
            }
            cupUpdates.Clear();
        }

        public static void CachePrimaryClothingItemPresetRVGEs()
        {
            foreach (var rvge in cachePrimaryClothingItemPresetRVGEUpdates)
            {               
                rvge.CachePrimaryClothingItemPresetRVGE();
            }
            cachePrimaryClothingItemPresetRVGEUpdates.Clear();
        }

        public static void RegisterCUPUpdate(ResourceVersionGroupEntry pcRVGE)
        {
            cupUpdates.Add(pcRVGE);
        }

        public static void RegisterPrimaryClothingItemPresetRVGE(ResourceVersionGroupEntry rvge)
        {
            cachePrimaryClothingItemPresetRVGEUpdates.Add(rvge);
        }

        public static HashSet<ResourceVersionGroupEntry> GetMatchingClothingRVGEs(HashSet<DAZClothingItem> dciUIDs, int resourceType)
        {
            var matchingRVGEs = new HashSet<ResourceVersionGroupEntry>();

            if (ResourceType.GetResourceCategory(resourceType) == ResourceCategory.clothing)
            {
                var clothingRVGEs = resourceVersionGroups.Where(rvge => rvge.resourceType == resourceType).ToList();

                foreach (var clothingRVGE in clothingRVGEs)
                {

                    if (!clothingRVGE.localFileResource)
                    {
                        PackagedClothingRVGE packagedClothingRVGE = clothingRVGE as PackagedClothingRVGE;
                        if (dciUIDs.Where(dci => dci.internalUid == packagedClothingRVGE.clothingUID && dci.packageUid== packagedClothingRVGE.packageVersionUID).Any()) matchingRVGEs.Add(clothingRVGE);
                    }
                    else
                    {
                        if (dciUIDs.Where(dci => dci.internalUid == clothingRVGE.clothingUID).Any()) matchingRVGEs.Add(clothingRVGE);
                    }
                }
            }

            return matchingRVGEs;
        }

        public static void OnDestoy()
        {
            foreach (ResourceVersionGroupEntry rvge in resourceVersionGroups)
            {
                if (rvge.resourceSprite != null)
                {
                    GameObject.Destroy(rvge.resourceSprite.texture);
                    GameObject.Destroy(rvge.resourceSprite);
                }

            }
            resourceVersionGroupsDict.Clear();
            resourceVersionGroupsDict = null;
            resourceVersionGroups.Clear();
            resourceVersionGroups = null;
            localResourceVersionGroups.Clear();
            localResourceVersionGroups = null;
            varResourceVersionGroups.Clear();
            varResourceVersionGroups = null;
        }

        private static void AddRVGE(string rvgeKey, ResourceVersionGroupEntry rvge)
        {
            resourceVersionGroupsDict.Add(rvgeKey, rvge);
            resourceVersionGroups.Add(rvge);
            if (rvge.localFileResource) localResourceVersionGroups.Add(rvge as LocalRVGE);
            else
            {
                varResourceVersionGroups.Add(rvge as PackagedRVGE);
                CreatorManifest.RegisterResource(rvge);
            }
            if (rvge.resourceCategory == ResourceCategory.clothing) CreatorManifest.RegisterClothingResource(rvge);
            if (!rvgeByResourceType.ContainsKey(rvge.resourceType)) rvgeByResourceType.Add(rvge.resourceType, new HashSet<ResourceVersionGroupEntry>());
            rvgeByResourceType[rvge.resourceType].Add( rvge);

            
            if (rvge.resourceType == ResourceType.presetAtom)
            {
                if (!rvgeByAtomPresetResourceType.ContainsKey(rvge.atomPresetResourceType)) rvgeByAtomPresetResourceType.Add(rvge.atomPresetResourceType, new HashSet<ResourceVersionGroupEntry>());
                rvgeByAtomPresetResourceType[rvge.atomPresetResourceType].Add(rvge);
            }
        }

        public static void PurgeLocalRVGE(LocalRVGE localRVGE)
        {
            RemoveRVGE(localRVGE);
        }

        public static void PurgePackageRVGE(PackagedRVGE packagedRVGE)
        {
            RemoveRVGE(packagedRVGE);
        }
        private static void RemoveRVGE(ResourceVersionGroupEntry rvge, bool keepSceneCache = false)
        {
            resourceVersionGroups.Remove(rvge);
            if (rvge.localFileResource)
            {
                var localRVGE = rvge as LocalRVGE;
                resourceVersionGroupsDict.Remove(rvge.resourceFolderPathFileName);
                localResourceVersionGroups.Remove(localRVGE);
                if (localRVGE.userDataFile != null) localRVGE.userDataFile.PurgeLocalRVGE(localRVGE);
            }
            else
            {
                varResourceVersionGroups.Remove(rvge as PackagedRVGE);
                resourceVersionGroupsDict.Remove(GetRVGKey(rvge as PackagedRVGE));
                CreatorManifest.RemoveRVGE(rvge as PackagedRVGE);
            }
            if (rvge.resourceCategory == ResourceCategory.clothing) CreatorManifest.RemoveClothingRVGE(rvge);

            rvgeByResourceType[rvge.resourceType].Remove(rvge);
            if (rvge.resourceType == ResourceType.presetAtom)
            {
                rvgeByAtomPresetResourceType[rvge.atomPresetResourceType].Remove(rvge);
            }

            if (rvge.resourceType == ResourceType.scene && !keepSceneCache) SceneImportCache.RemoveSceneCache(rvge);
        }

        public static int GetResourceTypeCount(int resourceType, string atomPresetResourceType)
        {
            if (!rvgeByResourceType.ContainsKey(resourceType)) return 0;
            if (resourceType != ResourceType.presetAtom) return rvgeByResourceType[resourceType].Where(o=>o.resourcePresent).Count();
            else
            {
                if (!rvgeByAtomPresetResourceType.ContainsKey(atomPresetResourceType)) return 0;
                return rvgeByAtomPresetResourceType[atomPresetResourceType].Where(o => o.resourcePresent).Count();
            }
        }

        public static void ResetRVGEFolderPath(LocalRVGE rvge)
        {
            string oldRVGEKey = rvge.resourceFolderPathFileName;
            rvge.ResetFolderPath();
            RekeyRVGE(rvge, oldRVGEKey);
        }

        public static void RenameRVGEFilename(LocalRVGE rvge, string newFullPathFileName)
        {
            string oldRVGEKey = rvge.resourceFolderPathFileName;
            rvge.RenameRVGEFilename(newFullPathFileName);
            RekeyRVGE(rvge, oldRVGEKey);
        }

        public static void MoveRenameRVGELocalFolderPath(LocalRVGE rvge, string newFileName, Folder targetFolder)
        {
            string oldRVGEKey = rvge.resourceFolderPathFileName;

            rvge.MoveRenameFolderPath(newFileName, targetFolder);
            RekeyRVGE(rvge, oldRVGEKey);
        }


        public static void MoveRVGELocalFolderPath(LocalRVGE rvge, Folder targetFolder)
        {
            string oldRVGEKey = rvge.resourceFolderPathFileName;
            rvge.MoveFolderPath(targetFolder);
            RekeyRVGE(rvge, oldRVGEKey);
        }

        private static void RekeyRVGE(LocalRVGE rvge, string oldRVGEKey)
        {
            string newRVGEKey = rvge.resourceFolderPathFileName;
            resourceVersionGroupsDict.Remove(oldRVGEKey);
            resourceVersionGroupsDict.Add(newRVGEKey, rvge);

            if (rvge.resourceType == ResourceType.scene) SceneImportCache.UpdateSceneImportLocalCacheFileName(oldRVGEKey, newRVGEKey);
        }


        public static void LocalResourcesInit(DateTime scanTime, HashSet<int> firstScanResourceTypes)
        {
            Globals.failureOnLocalMetaDataLoad = false;

            var localRVGEs = new HashSet<LocalRVGE>(localResourceVersionGroups);

            foreach (LocalRVGE existingLocalRVGE in localRVGEs)
            {
                RemoveRVGE(existingLocalRVGE, true);
            }
            localResourceVersionGroups.Clear();
            DataFileManifest.ClearLocalResourceUserDataFile();

            bool success = ReloadLocalDataFiles();

            if (success)
            {
                RefreshLocalSceneManifest(scanTime, firstScanResourceTypes.Contains(ResourceType.scene));
                RefreshLocalSubSceneManifest(scanTime, firstScanResourceTypes.Contains(ResourceType.subScenes));
                RefreshLocalCUAManifest(scanTime, firstScanResourceTypes.Contains(ResourceType.cua));

                RefreshLocalPresetManifest(ResourceType.presetAppearance, "", scanTime, firstScanResourceTypes.Contains(ResourceType.presetAppearance));
                RefreshLocalPresetManifest(ResourceType.presetAnimation, "", scanTime, firstScanResourceTypes.Contains(ResourceType.presetAnimation));
                RefreshLocalPresetManifest(ResourceType.presetBreastPhys, "", scanTime, firstScanResourceTypes.Contains(ResourceType.presetBreastPhys));
                RefreshLocalPresetManifest(ResourceType.presetClothing, "", scanTime, firstScanResourceTypes.Contains(ResourceType.presetClothing));
                RefreshLocalPresetManifest(ResourceType.presetGlutePhysics, "", scanTime, firstScanResourceTypes.Contains(ResourceType.presetGlutePhysics));
                RefreshLocalPresetManifest(ResourceType.presetHair, "", scanTime, firstScanResourceTypes.Contains(ResourceType.presetHair));
                RefreshLocalPresetManifest(ResourceType.presetMorphs, "", scanTime, firstScanResourceTypes.Contains(ResourceType.presetMorphs));
                RefreshLocalPresetManifest(ResourceType.presetPlugins, "", scanTime, firstScanResourceTypes.Contains(ResourceType.presetPlugins));
                RefreshLocalPresetManifest(ResourceType.presetPose, "", scanTime, firstScanResourceTypes.Contains(ResourceType.presetPose));
                RefreshLocalPresetManifest(ResourceType.presetSkin, "", scanTime, firstScanResourceTypes.Contains(ResourceType.presetSkin));
                RefreshLocalPresetManifest(ResourceType.presetGeneral, "", scanTime, firstScanResourceTypes.Contains(ResourceType.presetGeneral));

                RefreshLocalClothingManifest(ResourceType.femaleClothingItems, scanTime, firstScanResourceTypes.Contains(ResourceType.femaleClothingItems));
                RefreshLocalClothingManifest(ResourceType.maleClothingItems, scanTime, firstScanResourceTypes.Contains(ResourceType.maleClothingItems));

                RefreshLocalPresetManifest(ResourceType.presetSceneSessionPlugins,"", scanTime, firstScanResourceTypes.Contains(ResourceType.presetSceneSessionPlugins));

                RefreshLocalPresetManifest(ResourceType.clothingItemPresets, "Person", scanTime, firstScanResourceTypes.Contains(ResourceType.clothingItemPresets));

                foreach (var atomPresetType in Globals.nonPersonAtomTypes)
                {
                    RefreshLocalPresetManifest(ResourceType.presetAtom, atomPresetType, scanTime, firstScanResourceTypes.Contains(ResourceType.presetAtom));
                }

                if (FileManagerSecure.DirectoryExists("Custom\\Scripts"))
                {
                    DateTime startTime = DateTime.Now;
                    string[] patterns = new string[3] { "*.cs", "*.cslist", "*.dll" };
                    HashSet<string> localPluginFiles = Utils.GetFilesAtPathRecursive("Custom\\Scripts", patterns);

                    DateTime scriptFilesScanCompleteTime = DateTime.Now;
                    double findTime = (scriptFilesScanCompleteTime - startTime).TotalSeconds;

                    if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" Script files found in ", localPluginFiles.Count, findTime);

                    RefreshLocalResourceManifest(localPluginFiles, ResourceType.plugins, "", scanTime, firstScanResourceTypes.Contains(ResourceType.plugins));

                    DateTime scriptFilesProcessCompleteTime = DateTime.Now;
                    double processTime = (scriptFilesProcessCompleteTime - scriptFilesScanCompleteTime).TotalSeconds;

                    if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" Script files processed in ", localPluginFiles.Count, processTime);
                }
            }
            else
            {
                Globals.failureOnLocalMetaDataLoad = true;
            }           
        }

        public static void RefreshClothingUserPrefs()
        {
            foreach (var rvge in resourceVersionGroups.Where(o => o.resourceCategory == ResourceCategory.clothing))
            {
                rvge.RefreshUserPrefs();
            }
        }

        private static bool ReloadLocalDataFiles()
        {
            DateTime startTimeReloadLocalDataFiles = DateTime.Now;

            string[] localDataUserFiles;
            try
            {
                localDataUserFiles = FileManagerSecure.GetFiles(Const.LocalResourcesUserDataFolder, "*" + Const.UserDataFileExt);
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception in BA.ResourceManifest.ReloadLocalDataFiles whilst scanning for meta data files: " + e);
                Globals.failureOnLocalMetaDataLoad = true;
                return false;
            }

            HashSet<string> localDataUserFilesToSave = new HashSet<string>();
            HashSet<string> localDataUserFilesToDelete = new HashSet<string>();

            foreach (string localDataUserFile in localDataUserFiles)
            {
                try
                {
                    JSONClass localDataUserFileJC = SuperController.singleton.LoadJSON(localDataUserFile).AsObject;
                    JSONArray localResourcesJA = localDataUserFileJC["resources"].AsArray;

                    if (localResourcesJA.Count == 0) localDataUserFilesToDelete.Add(localDataUserFile);
                    else
                    {
                        foreach (JSONClass localResourceJC in localResourcesJA)
                        {
                            string resourceFilePath = localResourceJC["resourceFullFileName"];
                            try
                            {
                                if (FileManagerSecure.FileExists(resourceFilePath))
                                {
                                    int resourceType = EnumManifest.GetEnumValFromStoreVal(ResourceType.enumManifestName, localResourceJC["resourceType"]);
                                    int resourceCategory = ResourceType.GetResourceCategory(resourceType);

                                    if (resourceType != -1 && resourceCategory != -1)
                                    {
                                        string presetAtomType = GetPresetAtomType(resourceCategory, resourceType, resourceFilePath);

                                        if (localResourceJC["presetAtomType"] != null) presetAtomType = localResourceJC["presetAtomType"];  

                                        string resourceFileName = FileManagerSecure.GetFileName(resourceFilePath);

                                        if (resourceCategory != ResourceCategory.preset || (resourceCategory == ResourceCategory.preset && resourceFileName.StartsWith("Preset_")) || (resourceCategory == ResourceCategory.sceneOrSessionPluginPresets && resourceFileName.StartsWith("Plugins_")))
                                        {
                                            LocalRVGE rvge=null;

                                            if (resourceCategory == ResourceCategory.clothing)
                                            {
                                                try
                                                {
                                                    JSONClass clothingVAMJC = SuperController.singleton.LoadJSON(resourceFilePath).AsObject;
                                                    rvge = new LocalClothingRVGE(resourceFilePath, resourceType, clothingVAMJC);
                                                }
                                                catch (Exception e) { throw new Exception("BA.ResourceManifest.ReloadLocalDataFiles: exception caught in '" + resourceFilePath + "' - " + e); }


                                            }
                                            else
                                            {
                                                bool foundCorrectRootPath = false;
                                                foreach (string rootFolderPath in Const.GetResourceRootFolderNames(resourceType, presetAtomType))
                                                {
                                                    if (resourceFilePath.StartsWith(rootFolderPath))
                                                    {
                                                        foundCorrectRootPath = true;
                                                        rvge = new LocalRVGE(resourceFilePath, resourceType);

                                                        break;
                                                    }
                                                }
                                                if (!foundCorrectRootPath) localDataUserFilesToSave.Add(localDataUserFile);

                                            }

                                            if (rvge != null)
                                            {
                                                AddRVGE(resourceFilePath, rvge);
                                                DataFileManifest.RegisterExisitingDFUserLocalResource(rvge, localDataUserFile);
                                                rvge.LoadLocalUserData(localResourceJC);
                                            }                                        
                                        }
                                        else localDataUserFilesToSave.Add(localDataUserFile);
                                    }
                                    else
                                    {
                                        SuperController.LogError("Failure in BA.ResourceManifest.ReloadLocalDataFiles invalid resouce type or category for "+ resourceFilePath + " in File " + localDataUserFile);
                                        Globals.failureOnLocalMetaDataLoad = true;
                                    }
                                }
                                else localDataUserFilesToSave.Add(localDataUserFile);
                            }
                            catch (Exception e)
                            {
                                SuperController.LogError("Exception in BA.ResourceManifest.ReloadLocalDataFiles.1 for "+ resourceFilePath + " in File " + localDataUserFile + ": " + e);
                                Globals.failureOnLocalMetaDataLoad = true;
                            }

                        }
                    }                    
                }
                catch (Exception e)
                {
                    SuperController.LogError("Exception in BA.ResourceManifest.ReloadLocalDataFiles.2 for File " + localDataUserFile + ": " + e);
                    Globals.failureOnLocalMetaDataLoad = true;                   
                }
            }

            if (Globals.failureOnLocalMetaDataLoad) return false;

            foreach (string localDataUserFile in localDataUserFilesToSave)
            {
                DataFileManifest.RegisterLocalResourceUserSaveRequest(localDataUserFile);
            }

            foreach (string localDataUserFile in localDataUserFilesToDelete)
            {
                FileManagerSecure.DeleteFile(localDataUserFile);
            }

            DateTime endTimeReloadLocalDataFiles = DateTime.Now;
            double findTimeReload = (endTimeReloadLocalDataFiles - startTimeReloadLocalDataFiles).TotalSeconds;
            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" Local User Data files reloaded in ", localDataUserFiles.Count(), findTimeReload);

            return true;
        }

        private static void RefreshLocalCUAManifest(DateTime scanTime, bool forceNotNew)
        {
            if (FileManagerSecure.DirectoryExists("Custom\\Assets"))
            {
                DateTime startTime = DateTime.Now;

                string[] patterns = new string[2] { "*.assetbundle", "*.scene"};

                HashSet<string> localFiles = Utils.GetFilesAtPathRecursive("Custom\\Assets", patterns);

                DateTime cuaFilesScanCompleteTime = DateTime.Now;
                double findTime = (cuaFilesScanCompleteTime - startTime).TotalSeconds;
                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" CUA files found in ", localFiles.Count, findTime);

                RefreshLocalResourceManifest(localFiles, ResourceType.cua, "CustomUnityAsset", scanTime, forceNotNew);

                DateTime cuaFilesProcessCompleteTime = DateTime.Now;
                double processTime = (cuaFilesProcessCompleteTime - cuaFilesScanCompleteTime).TotalSeconds;

                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" CUA files processed in ", localFiles.Count, processTime);
            }
        }

        private static void RefreshLocalSubSceneManifest(DateTime scanTime, bool forceNotNew)
        {
            if (FileManagerSecure.DirectoryExists("Custom\\SubScene"))
            {
                DateTime startTime = DateTime.Now;

                HashSet<string> localFiles = Utils.GetFilesAtPathRecursive("Custom\\SubScene", "*.json");

                DateTime sceneFilesScanCompleteTime = DateTime.Now;
                double findTime = (sceneFilesScanCompleteTime - startTime).TotalSeconds;
                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" SubScene files found in ", localFiles.Count, findTime);

                RefreshLocalResourceManifest(localFiles, ResourceType.subScenes, "SubScene", scanTime, forceNotNew);

                DateTime sceneFilesProcessCompleteTime = DateTime.Now;
                double processTime = (sceneFilesProcessCompleteTime - sceneFilesScanCompleteTime).TotalSeconds;

                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" SubScene files processed in ", localFiles.Count, processTime);
            }
                

        }

        private static void RefreshLocalSceneManifest(DateTime scanTime, bool forceNotNew)
        {
            if (FileManagerSecure.DirectoryExists("Saves\\scene"))
            {
                DateTime startTime = DateTime.Now;

                HashSet<string> localFiles = Utils.GetFilesAtPathRecursive("Saves\\scene", "*.json");

                DateTime sceneFilesScanCompleteTime = DateTime.Now;
                double findTime = (sceneFilesScanCompleteTime - startTime).TotalSeconds;
                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" Scene files found in ", localFiles.Count, findTime);

                RefreshLocalResourceManifest(localFiles, ResourceType.scene, "", scanTime, forceNotNew);

                DateTime sceneFilesProcessCompleteTime = DateTime.Now;
                double processTime = (sceneFilesProcessCompleteTime - sceneFilesScanCompleteTime).TotalSeconds;

                if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" Scene files processed in ", localFiles.Count, processTime);
            }
                

        }
        private static void RefreshLocalClothingManifest(int resourceType, DateTime scanTime, bool forceNotNew)
        {
            DateTime startTime = DateTime.Now;

            HashSet<string> localFiles = new HashSet<string>();
            foreach (string resourcePath in Const.GetResourceRootFolderNames(resourceType, "Person"))
            {
                if (FileManagerSecure.DirectoryExists(resourcePath)) localFiles.UnionWith(Utils.GetFilesAtPathRecursive(resourcePath, "*.vam"));
            }

            DateTime presetFilesScanCompleteTime = DateTime.Now;
            double findTime = (presetFilesScanCompleteTime - startTime).TotalSeconds;
            string presetName = EnumManifest.GetDisplayVal(ResourceType.enumManifestName, resourceType);

            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" " + presetName + " files found in ", localFiles.Count, findTime);

            RefreshLocalResourceManifest(localFiles, resourceType, "Person", scanTime, forceNotNew);

            DateTime presetFilesProcessCompleteTime = DateTime.Now;
            double processTime = (presetFilesProcessCompleteTime - presetFilesScanCompleteTime).TotalSeconds;

            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" " + presetName + " files processed in ", localFiles.Count, processTime);
        }

        private static void RefreshLocalPresetManifest(int resourceType, string atomPresetResourceType,  DateTime scanTime, bool forceNotNew)
        {
            DateTime startTime = DateTime.Now;

            HashSet<string> localFiles = new HashSet<string>();
            foreach (string resourcePath in Const.GetResourceRootFolderNames(resourceType, atomPresetResourceType))
            {
                if (FileManagerSecure.DirectoryExists(resourcePath)) localFiles.UnionWith(Utils.GetFilesAtPathRecursive(resourcePath, "*.vap"));
            }

            DateTime presetFilesScanCompleteTime = DateTime.Now;
            double findTime = (presetFilesScanCompleteTime - startTime).TotalSeconds;
            string presetName = EnumManifest.GetDisplayVal(ResourceType.enumManifestName, resourceType);
            if (resourceType == ResourceType.presetAtom) presetName += "." + atomPresetResourceType;

            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" " + presetName + " files found in ", localFiles.Count, findTime);

            RefreshLocalResourceManifest(localFiles, resourceType, atomPresetResourceType, scanTime, forceNotNew);

            DateTime presetFilesProcessCompleteTime = DateTime.Now;
            double processTime = (presetFilesProcessCompleteTime - presetFilesScanCompleteTime).TotalSeconds;

            if (Globals.debugLoadTimingEnabled) Utils.LogProcessTime(" " + presetName + " files processed in ", localFiles.Count, processTime);
        }

        private static void RefreshLocalResourceManifest(HashSet<string> localFiles, int resourceType, string atomPresetResourceType, DateTime scanTime, bool forceNotNew)
        {
            foreach (string localFile in localFiles)
            {
                try
                {
                    RegisterLocalResourceOfType(localFile, resourceType, atomPresetResourceType, scanTime, forceNotNew);
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.RefreshLocalResourceManifest for " + localFile +" "+resourceType.ToString()+ ": " + e); }

            }
        }

        public static LocalRVGE RegisterLocalResourceOfType(string resourceFileNamePath, int resourceType, string atomPresetResourceType, DateTime scanTime, bool forceNotNew)
        {
            LocalRVGE rvge = null;
            if (!resourceVersionGroupsDict.ContainsKey(resourceFileNamePath))
            {
                int resourceCategory = ResourceType.GetResourceCategory(resourceType);
                string resourceFileName = FileManagerSecure.GetFileName(resourceFileNamePath);

                if (resourceCategory != ResourceCategory.preset || (resourceCategory == ResourceCategory.preset && resourceFileName.StartsWith("Preset_")) || (resourceCategory == ResourceCategory.sceneOrSessionPluginPresets && resourceFileName.StartsWith("Plugins_")))
                {
                    if (resourceCategory == ResourceCategory.clothing)
                    {
                        try
                        {
                            JSONClass clothingVAMJC = SuperController.singleton.LoadJSON(resourceFileNamePath).AsObject;
                            rvge = new LocalClothingRVGE(resourceFileNamePath, resourceType, clothingVAMJC);
                            (rvge as LocalRVGE).RefreshUserPrefs();
                        }
                        catch (Exception e) { throw new Exception("BA.ResourceManifest.RegisterLocalResourceOfType: exception caught in '"+ resourceFileNamePath+"' - " + e); }                        
                    }
                    else rvge = new LocalRVGE(resourceFileNamePath, resourceType);

                    if (rvge != null)
                    {
                        AddRVGE(resourceFileNamePath, rvge);
                        rvge.SetLocalFileScanTime(scanTime);
                        if (!forceNotNew) rvge.isNew = true;
                    }                    
                }
            }
            return rvge;
        }

        public static void ReloadResourceVARUserData(string[] varResourceUserFiles)
        {
            HashSet<string> varResourceUserFileNamesToSave = new HashSet<string>();
            HashSet<string> varResourceUserFileNamesToDelete = new HashSet<string>();

            foreach (string varResourceUserFileName in varResourceUserFiles)
            {
                try
                {
                    JSONClass jc = SuperController.singleton.LoadJSON(varResourceUserFileName).AsObject;

                    JSONArray resourceJA = jc["resources"].AsArray;
                    if (resourceJA.Count==0) varResourceUserFileNamesToDelete.Add(varResourceUserFileName);
                    else
                    {
                        foreach (JSONClass resourceJC in resourceJA)
                        {
                            string creatorName = resourceJC["creatorName"];
                            string packageName = resourceJC["packageName"];
                            string resourceFullFileName = resourceJC["resourceFullFileName"];

                            string rvgeKey = GetRVGKey(creatorName, packageName, resourceFullFileName);

                            if (resourceVersionGroupsDict.ContainsKey(rvgeKey))
                            {
                                PackagedRVGE rvge = resourceVersionGroupsDict[rvgeKey] as PackagedRVGE;
                                DataFileManifest.RegisterExisitingDFUserVARResource(rvge, varResourceUserFileName);

                                rvge.RestoreVARUserData(resourceJC);

                            }
                            else varResourceUserFileNamesToSave.Add(varResourceUserFileName);
                        }

                    }
                }
                catch (Exception e) { throw new Exception("Exception in BA.ResourceManifest.ReloadResourceVARUserData for " + varResourceUserFileName + ": " + e); }
            }
            foreach (string varResourceUserFileName in varResourceUserFileNamesToSave) DataFileManifest.RegisterVARResourceUserSaveRequest(varResourceUserFileName);
            foreach (string varResourceUserFileName in varResourceUserFileNamesToDelete) FileManagerSecure.DeleteFile(varResourceUserFileName);
        }

        public static void UpdateVARUserHideFlag(string rvgKey, int varVersion, bool hidden)
        {
            if (resourceVersionGroupsDict.ContainsKey(rvgKey))
            {
                ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                if (!rvge.localFileResource)
                {
                    (rvge as PackagedRVGE).SetHideUserVARVersion(varVersion, hidden);
                }

            }

        }
        public static void UpdateVARUserFavFlag(string rvgKey, int varVersion, bool favourite)
        {
            if (resourceVersionGroupsDict.ContainsKey(rvgKey))
            {
                ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                if (!rvge.localFileResource)
                {
                    (rvge as PackagedRVGE).SetFavUserVARVersion(varVersion, favourite);
                }

            }

        }

        public static IEnumerable<ResourceVersionGroupEntry> GetTagFilteredResourceGroups(ResourceBrowserFilterSettings filters)
        {
            IEnumerable<ResourceVersionGroupEntry> filteredRVGs = null;

            if (filters.resourceType == ResourceType.presetAtom)
            {
                filteredRVGs = resourceVersionGroups.Where(o =>
                        o.resourceType == filters.resourceType  &&
                        o.atomPresetResourceType == filters.atomPresetResourceType &&
                        o.resourcePresent &&
                        (filters.showHiddenJSB.val || (!o.isHidden && !o.isTagHidden && !o.isCreatorHidden)) &&
                        (!filters.onlyFavouritesJSB.val || o.isFavorite) &&
                        (!filters.onlyNewJSB.val || o.isNew) &&
                        !o.isMissing);
            }
            else
            {
                filteredRVGs = resourceVersionGroups.Where(o =>
                        o.resourceType == filters.resourceType &&
                        o.resourcePresent &&
                        (filters.showHiddenJSB.val || (!o.isHidden && !o.isTagHidden && !o.isCreatorHidden)) &&
                        (!filters.onlyFavouritesJSB.val || o.isFavorite) &&
                        (!filters.onlyNewJSB.val || o.isNew) &&
                        (Settings.enableCSLReferencedPluginsJSB.val || !o.isCSLReferenced) &&
                        !o.isMissing);
            }

            IEnumerable<ResourceVersionGroupEntry> tagFilteredRVGEs = filteredRVGs;
            if (filters.combiTagFilterEnabled)
            {
                tagFilteredRVGEs = filteredRVGs.Where(o => filters.combiTag.GetState(o));
            }
            else
            {
                if (filters.tagsFilterJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag)
                {
                    tagFilteredRVGEs = filteredRVGs.Where(o => o.IsTagged(TagManifest.GetTag(filters.resourceType, filters.atomPresetResourceType, filters.tagsFilterJSMultiEnum.mainVal)));
                }
                else
                {
                    if (filters.tagsFilterJSMultiEnum.valTopEnum == TagGroups.hiddenOnly) tagFilteredRVGEs = filteredRVGs.Where(o => o.isHidden || o.isTagHidden || o.isCreatorHidden);
                    if (filters.tagsFilterJSMultiEnum.valTopEnum == TagGroups.wornClothingOnly) tagFilteredRVGEs = filteredRVGs.Where(o => o.IsWornInScene());
                    if (filters.tagsFilterJSMultiEnum.valTopEnum == TagGroups.noTags) tagFilteredRVGEs = filteredRVGs.Where(o => o.IsNotTagged());
                    if (filters.tagsFilterJSMultiEnum.valTopEnum == TagGroups.noUserTags) tagFilteredRVGEs = filteredRVGs.Where(o => o.IsNotUserTagged());
                }
            }
            return tagFilteredRVGEs;
        }

        public static IEnumerable<ResourceVersionGroupEntry> GetCreatorFilteredResourceGroups(ResourceBrowserFilterSettings filters)
        {
            var tagFilteredRVGEs = GetTagFilteredResourceGroups(filters);

            IEnumerable<ResourceVersionGroupEntry> creatorFilteredRVGEs = tagFilteredRVGEs;
            if (filters.creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                if (filters.creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.allVARs)
                {
                    creatorFilteredRVGEs = tagFilteredRVGEs.Where(o =>
                        !o.localFileResource).ToList();
                }

                else if (filters.creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.localOnly)
                {
                    if (!filters.inludeSubFoldersJSB.val) creatorFilteredRVGEs = tagFilteredRVGEs.Where(o => o.localFileResource && (o.resourceCategory == ResourceCategory.clothing || o.containingLocalFolder == filters.selectedFolderFilter)).ToList();
                    else creatorFilteredRVGEs = tagFilteredRVGEs.Where(o => o.localFileResource && o.allContainingLocalFolders.Contains(filters.selectedFolderFilter)).ToList();
                }
            }
            else
            {
                creatorFilteredRVGEs = tagFilteredRVGEs.Where(o => o.creatorName == filters.creatorJSMultiEnum.displayVal).ToList();
            }
            return creatorFilteredRVGEs;
        }

        private static IEnumerable<ResourceVersionGroupEntry> GetLicenseFilteredResourceGroups(ResourceBrowserFilterSettings filters)
        {
            var creatorFilteredRVGEs = GetCreatorFilteredResourceGroups(filters);

            IEnumerable<ResourceVersionGroupEntry> licenseFilteredRVGEs = creatorFilteredRVGEs;

            if ( filters.creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag || filters.creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.allVARs)
            {
                if (filters.licenseFilterJSEnum.val != LicenseType.Any) licenseFilteredRVGEs = creatorFilteredRVGEs.Where(o => ((PackagedRVGE)o).varLicenseType == filters.licenseFilterJSEnum.val);
            }
            else if (filters.resourceCategory == ResourceCategory.clothing && filters.clothingCreatorJSME.valType== JSONStorableMultiEnumStringChooser.mainValFlag)
            {
                licenseFilteredRVGEs = creatorFilteredRVGEs.Where(o => o.clothingCreatorName == filters.clothingCreatorJSME.mainVal);
            }

            return licenseFilteredRVGEs;
        }

        public static List<int> GetLicenseTypesByInverseFilters(ResourceBrowserFilterSettings filters)
        {
            List<int> licenseTypesNotPresent = new List<int>();

            var creatorFilteredRVGEs = GetCreatorFilteredResourceGroups(filters);

            foreach (int licenseType in EnumManifest.GetEnumsList(LicenseType.enumManifestName))
            {
                if (creatorFilteredRVGEs.Where(o => ((PackagedRVGE)o).varLicenseType == licenseType).Count() == 0) licenseTypesNotPresent.Add(licenseType);
            }
            licenseTypesNotPresent.Remove(LicenseType.Any);
            return licenseTypesNotPresent;
        }



        public static List<ResourceVersionGroupEntry> GetResourceGroups(ResourceBrowserFilterSettings filters, int resourceNameSearchMode)
        {
            var licenseFilteredRVGEs = GetLicenseFilteredResourceGroups(filters);
            var globalFilterRVGs = licenseFilteredRVGEs.Where(o => filters.resourceNameFilterQuick == "" || (resourceNameSearchMode == ResourceNameSearchMode.StartsWith && o.resourceDisplayNameLowerCaps.StartsWith(filters.resourceNameFilterQuick)) || (resourceNameSearchMode == ResourceNameSearchMode.Contains && o.resourceDisplayNameLowerCaps.Contains(filters.resourceNameFilterQuick)) || (resourceNameSearchMode == ResourceNameSearchMode.ContainsFullPath && o.resourceFullPathReferenceWithAliasLowerCaps.Contains(filters.resourceNameFilterQuick)));

            if (filters.sortModeJSEnum.val == BrowserSortMode.AtoZ) globalFilterRVGs = globalFilterRVGs.OrderBy(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.ZtoA) globalFilterRVGs = globalFilterRVGs.OrderByDescending(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.newDateTime)
                globalFilterRVGs = globalFilterRVGs.OrderByDescending(o => o.firstCataloguedDateTime).ThenBy(o => o.creatorName + o.packageName + o.folderPath).ThenBy(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.reverseNewDateTime)
                globalFilterRVGs = globalFilterRVGs.OrderBy(o => o.firstCataloguedDateTime).ThenByDescending(o => o.creatorName + o.packageName + o.folderPath).ThenByDescending(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.updatedDateTime)
                globalFilterRVGs = globalFilterRVGs.OrderByDescending(o => o.lastCataloguedDateTime).ThenBy(o => o.creatorName + o.packageName + o.folderPath).ThenBy(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.reverseUpdatedDateTime)
                globalFilterRVGs = globalFilterRVGs.OrderBy(o => o.lastCataloguedDateTime).ThenByDescending(o => o.creatorName + o.packageName + o.folderPath).ThenByDescending(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.packageFolder) globalFilterRVGs = globalFilterRVGs.OrderBy(o => o.creatorName + o.packageName + o.folderPath).ThenBy(o => o.resourceDisplayName);

            return globalFilterRVGs.ToList();
        }
        public static HashSet<ResourceVersionGroupEntry> GetAllResourceGroups(int resourceType, string atomPresetResourceType)
        {
            HashSet<ResourceVersionGroupEntry> results;
            if (resourceType == ResourceType.presetAtom) results = new HashSet<ResourceVersionGroupEntry>(resourceVersionGroups.Where(o =>  o.atomPresetResourceType == atomPresetResourceType));
            else results = new HashSet<ResourceVersionGroupEntry>(resourceVersionGroups.Where(o => o.resourceType == resourceType));
            return results;
        }

        public static List<ResourceVersionGroupEntry> GetClothingItemPresets(string storePath, string clothingItemFileNameNoExt)
        {
            return new List<ResourceVersionGroupEntry>(resourceVersionGroups.Where(o =>(o.resourceType==ResourceType.clothingItemPresets && (o.folderPath == storePath ||o.folderPath.StartsWith(storePath+"\\")) && o.resourceFileNameNoExt.StartsWith(clothingItemFileNameNoExt+"_"))).OrderByDescending(o=>o.localFileResource).ThenBy(o=>o.resourceDisplayName));
        }
        public static string GetRVGKey(string creatorName, string packageName, string resourceFileName)
        {
            return creatorName + "." + packageName + ":" + resourceFileName;
        }
        public static string GetRVGKey(VARPackageManifestEntry vpme, string resourceFileName)
        {
            return GetRVGKey(vpme.creatorName, vpme.packageName, resourceFileName);
        }
        public static string GetRVGKey(VARPackageVersionGroupEntry vpvge, string resourceFileName)
        {
            return GetRVGKey(vpvge.creatorName, vpvge.packageName, resourceFileName);
        }
        public static string GetRVGKey(PackagedRVGE packagedRVGE)
        {
            return GetRVGKey(packagedRVGE.creatorName, packagedRVGE.packageName, packagedRVGE.resourceFolderPathFileName);
        }

        public static ResourceVersionGroupEntry GetRVGE(string rvgeKey)
        {
            if (!resourceVersionGroupsDict.ContainsKey(rvgeKey)) return null;
            else return resourceVersionGroupsDict[rvgeKey];
        }


        public static void RegisterNewVARResources(VARPackageManifestEntry vpme, HashSet<int> firstScanResourceTypes)
        {
            HashSet<string> contentFiles;
            HashSet<string> hideFiles = new HashSet<string>();
            HashSet<string> favFiles = new HashSet<string>();

            contentFiles = Utils.GetFilesAtPathRecursive(vpme.varFilePath + ":", "*.*");

            int length = vpme.varFilePath.Length + 2;
            foreach (string contentFile in contentFiles)
            {
                if (contentFile.EndsWith(".hide")) hideFiles.Add(contentFile);
                else if (contentFile.EndsWith(".fav")) favFiles.Add(contentFile);
                else
                {
                    string newResource = contentFile.Substring(length);
                    int resourceType = GetResourceType(newResource);
                    int resourceCategory = ResourceType.GetResourceCategory(resourceType);
                    bool notNew = firstScanResourceTypes.Contains(resourceType);

                    if (resourceType != ResourceType.unknown) RegisterNewVARResourceOfType(vpme, newResource, resourceType, notNew);
                }
            }

            HashSet<string> csListFiles = new HashSet<string>(contentFiles.Where(fileName => fileName.EndsWith(".cslist") && fileName.Contains(":\\Custom\\Scripts")));
            ProcessVARCSListReferences(vpme, csListFiles);

            RegisterCreatorHideFlags(vpme, hideFiles);
            RegisterCreatorFavFlags(vpme, favFiles);
        }

        public static bool RegisterVARResourceFromManifest(VARPackageVersionGroupEntry vpvge, JSONClass resourceJC, string varResourceMetaFileName)
        {
            
            string resourceFullFileName = resourceJC["resourceFullFileName"];
            string rvgKey = GetRVGKey(vpvge, resourceFullFileName);      

            if (resourceVersionGroupsDict.ContainsKey(rvgKey)) return true;
            else
            {
                JSONArray varVersionsJA = resourceJC["varVersions"].AsArray;
                if (vpvge.AnyExistingVARVersions(varVersionsJA))
                {
                    int resourceType = EnumManifest.GetEnumValFromStoreVal(ResourceType.enumManifestName, resourceJC["resourceType"]);
                    int resourceCategory = ResourceType.GetResourceCategory(resourceType);

                    PackagedRVGE rvge = null;

                    if (resourceCategory == ResourceCategory.clothing)
                    {
                        try
                        {
                            rvge = new PackagedClothingRVGE(resourceJC, vpvge);
                        }
                        catch(Exception e) { SuperController.LogError("BA.ResourceManifest.RegisterVARResourceFromManifest: Exception caught - "+e); }
                    }
                    else rvge = new PackagedRVGE(resourceJC, vpvge);

                    if (rvge != null)
                    {
                        AddRVGE(rvgKey, rvge);

                        DataFileManifest.RegisterExisitingDFCoreVARResource(rvge, varResourceMetaFileName);
                        vpvge.RegisterRVGE(rvge);

                        if (vpvge.AnyMismatchedVARVersions(varVersionsJA)) rvge.SetCoreDataPersist();
                        return true;
                    }
                    
                }

            }

            return false;
        }
        public static void RegisterCreatorHideFlags(VARPackageManifestEntry vpme, HashSet<string> hideFiles)
        {
            HashSet<ResourceVersionGroupEntry> hiddenRVGEs = new HashSet<ResourceVersionGroupEntry>();
            foreach (string hideFile in hideFiles)
            {
                string hideFileNoExt = hideFile.Substring(0, hideFile.Length - 5);
                string hideFilPath = hideFileNoExt.Substring(hideFileNoExt.IndexOf(":\\") + 2);

                string rvgKey = GetRVGKey(vpme, hideFilPath);

                if (resourceVersionGroupsDict.ContainsKey(rvgKey))
                {
                    ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                    hiddenRVGEs.Add(rvge);
                    (rvge as PackagedRVGE).SetHideCreatorVARVersion(vpme.varVersion,true);
                }
            }

            var vpvge = VARPackageManifest.varPackageVersionGroups[VARPackageManifest.GetVPVGEKey(vpme)];

            vpvge.UnhideResources(hiddenRVGEs, vpme.varVersion);
        }
        public static void RegisterCreatorFavFlags(VARPackageManifestEntry vpme, HashSet<string> favFiles)
        {
            HashSet<ResourceVersionGroupEntry> favRVGEs = new HashSet<ResourceVersionGroupEntry>();

            foreach (string favFile in favFiles)
            {
                string favFileNoExt = favFile.Substring(0, favFile.Length - 4);
                string favFilPath = favFileNoExt.Substring(favFileNoExt.IndexOf(":\\") + 2);

                string rvgKey = GetRVGKey(vpme, favFilPath);
                if (resourceVersionGroupsDict.ContainsKey(rvgKey))
                {
                    ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                    favRVGEs.Add(rvge);
                    (rvge as PackagedRVGE).SetFavCreatorVARVersion(vpme.varVersion,true);
                }
            }

            var vpvge = VARPackageManifest.varPackageVersionGroups[VARPackageManifest.GetVPVGEKey(vpme)];

            vpvge.UnfavResource(favRVGEs, vpme.varVersion);
        }

        public static void ProcessVARCSListReferences(VARPackageManifestEntry vpme, HashSet<string> csListFiles)
        {

            string[] stringSeparators = new string[] { ".cs", "\n", "\r\n" };
            string csListFileContents = null;

            foreach (string cslFilePath in csListFiles)
            {
                try
                {
                    csListFileContents = FileManagerSecure.ReadAllText(cslFilePath);
                }
                catch (Exception e)
                {
                    Utils.LogDebug("<ResourceManifest.ProcessCSListReferences> '" + cslFilePath + "' could not be read by VAM - possible malformed file");
                    break;
                }
                if (csListFileContents != null)
                {
                    string[] csListPluginFiles = csListFileContents.Split(stringSeparators, System.StringSplitOptions.RemoveEmptyEntries);

                    string cslPath = FileManagerSecure.GetDirectoryName(cslFilePath.Substring(cslFilePath.IndexOf(":\\") + 2));
                    foreach (string csFileRef in csListPluginFiles)
                    {
                        if (!(String.IsNullOrEmpty(csFileRef) || csFileRef.Trim().Length == 0) && !csFileRef.Contains(":"))
                        {
                            string rvgKey = GetRVGKey(vpme, cslPath + "\\" + csFileRef.Replace('/', '\\') + ".cs");
                            if (resourceVersionGroupsDict.ContainsKey(rvgKey))
                            {
                                var rvge = (PackagedRVGE)resourceVersionGroupsDict[rvgKey];
                                rvge.SetCSLFileReferenced(vpme.varVersion);
                            }


                        }

                    }
                }
            }

        }

        public static string GetPresetAtomType(int resourceCategory, int resourceType, string resourceFileName)
        {
            string presetAtomType = "";

            if (resourceType == ResourceType.presetAtom)
            {
                if (resourceFileName.StartsWith("Custom\\Atom\\") && !resourceFileName.StartsWith("Custom\\Atom\\Person\\") && resourceFileName.EndsWith(".vap"))
                {
                    string temp = resourceFileName.Substring(12);
                    if (temp.Contains('\\')) presetAtomType = temp.Substring(0, temp.IndexOf('\\'));
                }
            }
            else if (resourceCategory == ResourceCategory.preset || resourceType == ResourceType.clothingItemPresets) presetAtomType = "Person";
            else if (resourceType == ResourceType.subScenes) presetAtomType = "SubScene";
            else if (resourceType == ResourceType.cua) presetAtomType = "CustomUnityAsset";
            else if (resourceType == ResourceType.presetSceneSessionPlugins) presetAtomType = "CoreControl";
            return presetAtomType;
        }
        public static int GetResourceType(string resourceFileName)
        {
            if (resourceFileName.StartsWith("Saves\\scene") && resourceFileName.EndsWith(".json")) return ResourceType.scene;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\AnimationPresets") && resourceFileName.EndsWith(".vap")) return ResourceType.presetAnimation;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Appearance") && resourceFileName.EndsWith(".vap")) return ResourceType.presetAppearance;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\BreastPhysics") && resourceFileName.EndsWith(".vap")) return ResourceType.presetBreastPhys;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Clothing") && resourceFileName.EndsWith(".vap")) return ResourceType.presetClothing;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\General") && resourceFileName.EndsWith(".vap")) return ResourceType.presetGeneral;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\GlutePhysics") && resourceFileName.EndsWith(".vap")) return ResourceType.presetGlutePhysics;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Hair") && resourceFileName.EndsWith(".vap")) return ResourceType.presetHair;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Morphs") && resourceFileName.EndsWith(".vap")) return ResourceType.presetMorphs;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Plugins") && resourceFileName.EndsWith(".vap")) return ResourceType.presetPlugins;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Pose") && resourceFileName.EndsWith(".vap")) return ResourceType.presetPose;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Skin") && resourceFileName.EndsWith(".vap")) return ResourceType.presetSkin;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Skin") && resourceFileName.EndsWith(".vap")) return ResourceType.presetSkin;
            if(resourceFileName.StartsWith("Custom\\SubScene") && resourceFileName.EndsWith(".json")) return ResourceType.subScenes;
            if (resourceFileName.StartsWith("Custom\\Clothing\\Female") && resourceFileName.EndsWith(".vam")) return ResourceType.femaleClothingItems;
            if (resourceFileName.StartsWith("Custom\\Clothing\\Male") && resourceFileName.EndsWith(".vam")) return ResourceType.maleClothingItems;
            if (resourceFileName.StartsWith("Custom\\Assets") && (resourceFileName.EndsWith(".assetbundle") || resourceFileName.EndsWith(".scene"))) return ResourceType.cua;
            if ((resourceFileName.StartsWith("Custom\\Clothing\\Female") || resourceFileName.StartsWith("Custom\\Clothing\\Male")) && resourceFileName.EndsWith(".vap") ) return ResourceType.clothingItemPresets;

            //           if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Textures") && (resourceFileName.EndsWith(".jpg")|| resourceFileName.EndsWith(".png"))) return ResourceType.personTexture;
            //           if (resourceFileName.StartsWith("Custom\\Textures") && (resourceFileName.EndsWith(".jpg") || resourceFileName.EndsWith(".png"))) return ResourceType.atomTexture;

            if (resourceFileName.StartsWith("Custom\\Scripts") && (resourceFileName.EndsWith(".cs") || resourceFileName.EndsWith(".cslist") || resourceFileName.EndsWith(".dll"))) return ResourceType.plugins;

            if (resourceFileName.StartsWith("Custom\\Atom\\") && !resourceFileName.StartsWith("Custom\\Atom\\Person\\") && resourceFileName.EndsWith(".vap"))
            {
                string presetAtomType = GetPresetAtomType(ResourceCategory.preset,ResourceType.presetAtom, resourceFileName);
                if(Globals.nonPersonAtomTypes.Contains(presetAtomType)) return ResourceType.presetAtom;
            }

            if (resourceFileName.StartsWith("Custom\\PluginPresets") && resourceFileName.EndsWith(".vap"))
            {
                SuperController.LogMessage("TEST: "+resourceFileName);
                return ResourceType.presetSceneSessionPlugins;
            }

            return ResourceType.unknown;
        }

        public static void RegisterNewVARResourceOfType(VARPackageManifestEntry vpme, string resourceFilePath, int resourceType, bool forceNotNew)
        {
            string rvgKey = GetRVGKey(vpme, resourceFilePath);
          
            if (!resourceVersionGroupsDict.ContainsKey(rvgKey))
            {
                int resourceCategory = ResourceType.GetResourceCategory(resourceType);
                string resourceFileName = FileManagerSecure.GetFileName(resourceFilePath);

                if (resourceCategory != ResourceCategory.preset || (resourceCategory == ResourceCategory.preset && resourceFileName.StartsWith("Preset_")) || (resourceCategory == ResourceCategory.sceneOrSessionPluginPresets && resourceFileName.StartsWith("Plugins_")))
                {
                    PackagedRVGE rvge=null;
                    if (resourceCategory == ResourceCategory.clothing)
                    {
                        string vpmeKey = VARPackageManifest.GetVPMEKey(vpme);

                        try
                        {
                            JSONClass clothingVAMJC = SuperController.singleton.LoadJSON(vpmeKey+":\\"+ resourceFilePath).AsObject;
                            rvge = new PackagedClothingRVGE(vpme, resourceFilePath, resourceType, !forceNotNew, clothingVAMJC);
                        }
                        catch (Exception e) { SuperController.LogError("BA.ResourceManifest.RegisterNewVARResourceOfType: exception caught in '"+ vpmeKey + ":\\" + resourceFilePath+"' - " + e); }
                    }
                    else rvge = new PackagedRVGE(vpme, resourceFilePath, resourceType, !forceNotNew);

                    if (rvge != null)
                    {
                        AddRVGE(rvgKey, rvge);

                        DataFileManifest.RegisterNewDFCoreVARResource(rvge);
                        vpme.vpvge.RegisterRVGE(rvge);
                    }
                }                   
            }
            else
            {
                PackagedRVGE rvge = resourceVersionGroupsDict[rvgKey] as PackagedRVGE;
                rvge.AddVARVersion(vpme.varVersion, !forceNotNew);
                
            }
        }



        public static void ClearAllUserFilePrefs()
        {
            foreach (PackagedRVGE packegedRVGE in varResourceVersionGroups)
            {
                packegedRVGE.ClearAllUserFilePrefs();
            }
        }

        public static List<string> existingAtomPresetResourceTypes
        {
            get
            {
                List<string> choices = new List<string>();

                foreach (string nonPersonAtomType in Globals.nonPersonAtomTypes)
                {
                    int count = ResourceManifest.GetResourceTypeCount(ResourceType.presetAtom, nonPersonAtomType);
                    if (count > 0) choices.Add(nonPersonAtomType);

                }

                return choices;
            }
        }

        private static List<string> GetAtomPresetResourceChoices (bool useConceal)
        {
            List<string> choices = new List<string>();
            foreach (string nonPersonAtomType in Globals.nonPersonAtomTypes)
            {
                if (!Settings.IsResourceConcealed(ResourceType.presetAtom, nonPersonAtomType) || !useConceal)
                {
                    int count = GetResourceTypeCount(ResourceType.presetAtom, nonPersonAtomType);
                if (count > 0) choices.Add(nonPersonAtomType);
                }

            }

            return choices;
        }

        private static List<string> GetAtomPresetResourceDisplayChoices(bool useConceal)
        {
            List<string> choices = new List<string>();
            foreach (string nonPersonAtomType in Globals.nonPersonAtomTypes)
            {
                if (!Settings.IsResourceConcealed(ResourceType.presetAtom, nonPersonAtomType) || !useConceal)
                {
                    int count = GetResourceTypeCount(ResourceType.presetAtom, nonPersonAtomType);
                    if (count > 0) choices.Add("Preset: " + nonPersonAtomType);
                }

            }

            return choices;
        }


        public static void PopulateResourceTypeSelector(JSONStorableMultiEnumStringChooser resourceTypeJSEnum, bool useConceal=true)
        {
            List<int> resourceTypesToExclude = new List<int>();
            List<string> atomPresetChoices = GetAtomPresetResourceChoices(useConceal);
            List<string> atomPresetDisplayChoices = GetAtomPresetResourceDisplayChoices(useConceal);
            if (!PC.patreonContentEnabled)
            {
                resourceTypesToExclude = EnumManifest.GetEnumsList(ResourceType.enumManifestName);
                resourceTypesToExclude.Remove(ResourceType.scene);
                resourceTypesToExclude.Remove(ResourceType.presetAppearance);
                resourceTypesToExclude.Remove(ResourceType.femaleClothingItems);
                resourceTypesToExclude.Remove(ResourceType.maleClothingItems);
                atomPresetChoices = new List<string>();
                atomPresetDisplayChoices = new List<string>();
            }
            resourceTypesToExclude.Add(ResourceType.presetAtom);
            resourceTypesToExclude.Add(ResourceType.clothingItemPresets);

            if (useConceal)
            {
                foreach (int resourceType in EnumManifest.GetEnumsList(ResourceType.enumManifestName))
                {
                    if (resourceType != ResourceType.presetAtom && Settings.IsResourceConcealed(resourceType, "")) resourceTypesToExclude.Add(resourceType);

                }
            }
            
            resourceTypeJSEnum.SetTopEnum(ResourceType.enumManifestName, resourceTypesToExclude);
            resourceTypeJSEnum.SetMainChoices(atomPresetChoices, atomPresetDisplayChoices);
        }
    }


    public abstract class ResourceVersionGroupEntry
    {
        public string rvgeKey { get
            {
                if (localFileResource) return resourceFullPathReference;
                else
                {
                    var packagedRVGE = this as PackagedRVGE;
                    return ResourceManifest.GetRVGKey(packagedRVGE);
                }
            } }
        public virtual string clothingUID { get; protected set; }
        public virtual string clothingCreatorName { get; protected set; } = "";


        protected string primaryClothingItemPresetRVGKey = null;
        private ResourceVersionGroupEntry _primaryClothingItemPresetRVGE = null;
        public ResourceVersionGroupEntry primaryClothingItemPresetRVGE { get
            {
                if (_primaryClothingItemPresetRVGE == null)
                {
                    if (primaryClothingItemPresetRVGKey != null)
                    {
                        _primaryClothingItemPresetRVGE = ResourceManifest.GetRVGE(primaryClothingItemPresetRVGKey);
                        primaryClothingItemPresetRVGKey = null;
                    }
                }
                return _primaryClothingItemPresetRVGE;
            }
            set
            {
                _primaryClothingItemPresetRVGE = value;
                primaryClothingItemPresetRVGKey = null;
                SetUserDataPersist();
            }
        }

        public void CachePrimaryClothingItemPresetRVGE()
        {
            if (primaryClothingItemPresetRVGKey != null)
            {
                _primaryClothingItemPresetRVGE = ResourceManifest.GetRVGE(primaryClothingItemPresetRVGKey);
                if (_primaryClothingItemPresetRVGE == null) SuperController.LogError("BA.ResourceVersionGroupEntry.CachePrimaryClothingItemPresetRVGE: Unable to retrieve Primary Clothing Item Preset '"+ primaryClothingItemPresetRVGKey+"' for RVGE '"+resourceFullPathReference+"'");
                primaryClothingItemPresetRVGKey = null;
                
            }
            
        }

        protected int baFavorite = TriState.indeterminate;
        protected int baHidden = TriState.indeterminate;
        public virtual string creatorName { get; protected set; }
        public virtual string packageName { get; protected set; }
        public virtual bool resourcePresent { get { return true; } }

        // Path of Folder containing the Resource
        public string folderPath { get; protected set; }

        // Resource File Name with no path or file extension.
        public string resourceFileNameNoExt { get; private set; }

        public string resourceFileNameNoExtNoPrefix { get; private set; }

        public JSONStorableString resourceNameAliasJSS;
        public JSONStorableString notesJSS;

        private static HashSet<ResourceVersionGroupEntry> cuaLoadsTriggered = new HashSet<ResourceVersionGroupEntry>();
        private DateTime cuaLoadTiggeredTime = DateTime.MinValue;
        private CustomUnityAssetLoader cuaLoader = null;

        public static void Update()
        {
            HashSet<ResourceVersionGroupEntry> cuaLoadsTriggeredFinished = new HashSet<ResourceVersionGroupEntry>();
            foreach (var entry in cuaLoadsTriggered)
            {
                if (entry.cuaLoader != null && (DateTime.Now - entry.cuaLoadTiggeredTime).TotalSeconds < 45)
                {
                    var cuaAssetNameJSSC = entry.cuaLoader.GetStringChooserJSONParam("assetName");
                    if (cuaAssetNameJSSC.choices.Count>1)
                    {
                        cuaAssetNameJSSC.val = cuaAssetNameJSSC.choices[1];
                        cuaLoadsTriggeredFinished.Add(entry);
                    }
                }
                else cuaLoadsTriggeredFinished.Add(entry);
            }

            foreach (var entry in cuaLoadsTriggeredFinished) cuaLoadsTriggered.Remove(entry);
        }

        //Resource Filename (with no path or extension) with prefix removed for Presets
        public virtual string resourceDisplayName { get
            {
                if (resourceNameAliasJSS.val == "") return resourceFileNameNoExtNoPrefix;
                return resourceNameAliasJSS.val;
            }
        }

        // Resource File Name in lower case with no path. Includes file extension.
        public virtual string resourceDisplayNameLowerCaps { get; private set; }
        public virtual DateTime firstCataloguedDateTime { get; }
        public virtual DateTime lastCataloguedDateTime { get; }
        public string resourceFolderPathFileName { get; protected set; }

        public string resourceFolderPathFileNameNoExt { get; protected set; }
        public string resourceFullPathReference { get; protected set; }
        public string resourceFullPathReferenceWithAliasLowerCaps { get; protected set; }

        public string resourceFileNameExtension { get; protected set; }
        public int resourceType { get; protected set; }

        public string atomPresetResourceType { get; protected set; }

        private ResourceSettings resourceTypeSettings { get { return Settings.GetResourceSettings(resourceType, atomPresetResourceType); } }
        public int resourceCategory { get; protected set; }
        public int resourceTargetCategory { get; protected set; }
        public bool localFileResource { get; protected set; } = false;

        protected virtual bool vamUserHidden { get; set; } = false;
        protected virtual bool vamUserFavorite { get; set; } = false;

        public virtual bool isMissing
        {
            get
            {
                return false;
            }
        }

        public bool isTagHidden { get
            {
                return TagManifest.GetRVGETags(this).Where(a => a.hideTaggedResourcesJSB.val).Any();
            } }

        public bool isCreatorHidden
        {
            get
            {
                if (localFileResource) return false;
                return CreatorManifest.GetCreator(creatorName).IsHiddenResources(resourceType,atomPresetResourceType);
            }
        }

        public bool isTagPreferred
        {
            get
            {
                return TagManifest.GetRVGETags(this).Where(a => a.preferredTagJSB.val).Any();
            }
        }

        public bool isCreatorPreferred
        {
            get
            {
                if (localFileResource) return false;
                return CreatorManifest.GetCreator(creatorName).IsPreferredCreator(resourceType, atomPresetResourceType);
            }
        }


        public bool isRenameable
        {
            get
            {
                if (!localFileResource) return false;
                if (resourceCategory == ResourceCategory.scene && !Globals.scenesFolderAccess) return false;
                if (resourceCategory == ResourceCategory.plugins && !Globals.customFolderAccess) return false;
                if (resourceCategory == ResourceCategory.preset && !Globals.customFolderAccess) return false;
                if (resourceCategory == ResourceCategory.sceneOrSessionPluginPresets && !Globals.customFolderAccess) return false;
                if (resourceCategory == ResourceCategory.subScenes && !Globals.customFolderAccess) return false;
                if (resourceCategory == ResourceCategory.cua && !Globals.customFolderAccess) return false;
                if (resourceCategory == ResourceCategory.clothing) return false;
                return true;
            }
        }
        public  bool isDeletable { get
            {
                if (!localFileResource && Globals.addonPackagesAccess) return true;
                if (localFileResource)
                {
                    if (resourceCategory == ResourceCategory.scene && Globals.scenesFolderAccess) return true;
                    if ((resourceCategory == ResourceCategory.plugins || resourceCategory == ResourceCategory.preset || resourceCategory == ResourceCategory.subScenes || resourceCategory == ResourceCategory.clothing || resourceCategory== ResourceCategory.cua || resourceCategory ==ResourceCategory.clothingItemPresets || resourceCategory == ResourceCategory.sceneOrSessionPluginPresets) && Globals.customFolderAccess) return true;
                }
                return false;
            } }
        public virtual bool isHidden
        {
            get
            {
                return (vamUserHidden || baHidden == TriState.active);
            }
            set
            {
                if (value)
                {
                    if (baHidden != TriState.active)
                    {
                        baHidden = TriState.active;
                        SetUserDataPersist();
                    }

                }
                else
                {
                    if (baHidden != TriState.inactive)
                    {
                        baHidden = TriState.inactive;
                        SetUserDataPersist();
                    }
                }

                // TBD need to do something different for clothing?
                SetVAMUserHidden(value);
                foreach (CombiTag combiTag in TagManifest.GetCombiTags(resourceType,atomPresetResourceType))
                {
                    if (combiTag.usesHidden) SetTag(combiTag);
                }               
            }
        }
        public abstract bool isCSLReferenced { get; }

        public virtual bool isMultiVARVersion { get { return false; } }
        public virtual bool isFavorite
        {
            get
            {
                return (vamUserFavorite || baFavorite == TriState.active);
            }
            set
            {
                if (value)
                {
                    if (baFavorite != TriState.active)
                    {
                        baFavorite = TriState.active;
                        SetUserDataPersist();
                    }

                }
                else
                {
                    if (baFavorite != TriState.inactive)
                    {
                        baFavorite = TriState.inactive;
                        SetUserDataPersist();
                    }
                }
                if (resourceCategory!=ResourceCategory.clothing) SetVAMUserFavourite(value);
                foreach (CombiTag combiTag in TagManifest.GetCombiTags(resourceType, atomPresetResourceType))
                {
                    if (combiTag.usesFavorite) SetTag(combiTag);
                }
            }
        }

        protected bool _isNew = false;
        public bool isNew
        {
            get
            {
                return _isNew;
            }
            set
            {
                if (_isNew != value)
                {
                    _isNew = value;
                    SetUserDataPersist();
                }
                foreach (CombiTag combiTag in TagManifest.GetCombiTags(resourceType, atomPresetResourceType))
                {
                    if (combiTag.usesNew) SetTag(combiTag);
                }
            }
        }


        public Sprite resourceSprite = null;

        public abstract void SetUserDataPersist();

        protected abstract void SetVAMUserFavourite(bool value);
        protected abstract void SetVAMUserHidden(bool value);

        public virtual string resourceThumbnailReference { get; }

        protected UIDynamicButton queuedButtonLoad=null;
        protected ImageLoaderThreaded.QueuedImage queuedImage = null;

        public ResourceVersionGroupEntry() {
            resourceNameAliasJSS = new JSONStorableString("resourceNameAlias", "", delegate (string x)
            {
                resourceFullPathReferenceWithAliasLowerCaps = (resourceFullPathReference + resourceNameAliasJSS.val).ToLower();
                if (resourceCategory!=ResourceCategory.clothing) resourceDisplayNameLowerCaps = resourceDisplayName.ToLower();
                SetUserDataPersist();
                if (BrowserAssist.mainBrowserUI != null)
                {
                    BrowserAssist.mainBrowserUI.resourceBrowserUI. UpdateDisplayName(this);
                    BrowserAssist.mainBrowserUI.resourceSelectionsUI. UpdateDisplayName(this);
                }
                if (BrowserAssist.callbackBrowserUI != null)
                {
                    BrowserAssist.callbackBrowserUI.resourceBrowserUI.UpdateDisplayName(this);
                    BrowserAssist.callbackBrowserUI.resourceSelectionsUI.UpdateDisplayName(this);
                }
            });
            notesJSS = new JSONStorableString("notes", "", delegate (string x)
            {
                SetUserDataPersist();
            });
        }

        private Folder _containingLocalFolder = null;
        public Folder containingLocalFolder { get { return _containingLocalFolder; } protected set
            {
                if (_containingLocalFolder != null) _containingLocalFolder.UnRegisterLocalRVGE(this);
                _containingLocalFolder = value;
                if (_containingLocalFolder != null) _containingLocalFolder.RegisterLocalRVGE(this);
            } } 
        protected virtual void Init()
        {
            int fileExtensionIndex = resourceFolderPathFileName.LastIndexOf('.');
            int fileNameIndex = 0;
            resourceCategory = ResourceType.GetResourceCategory(resourceType);

            if (resourceFolderPathFileName.Contains("\\")) fileNameIndex= resourceFolderPathFileName.LastIndexOf('\\');
            folderPath = resourceFolderPathFileName.Substring(0, fileNameIndex);
            resourceFileNameNoExt = resourceFolderPathFileName.Substring(folderPath.Length + 1, fileExtensionIndex - fileNameIndex - 1);
            if (resourceCategory == ResourceCategory.preset) resourceFileNameNoExtNoPrefix= resourceFileNameNoExt.Substring(7);
            else if (resourceCategory == ResourceCategory.sceneOrSessionPluginPresets) resourceFileNameNoExtNoPrefix = resourceFileNameNoExt.Substring(8);
            else resourceFileNameNoExtNoPrefix = resourceFileNameNoExt;

            if (resourceCategory != ResourceCategory.clothing) resourceDisplayNameLowerCaps = resourceDisplayName.ToLower();
            resourceFileNameExtension = resourceFolderPathFileName.Substring(fileExtensionIndex + 1);
            resourceFolderPathFileNameNoExt = resourceFolderPathFileName.Substring(0, fileExtensionIndex);

            resourceTargetCategory = ResourceType.GetResourceTargetCategory(resourceType);
            atomPresetResourceType = ResourceManifest.GetPresetAtomType(resourceCategory, resourceType, resourceFolderPathFileName);

            if (localFileResource && containingLocalFolder==null)
            {
                var localFolder = FolderBrowser.GetFolder(folderPath, resourceType, atomPresetResourceType);
                if (localFolder!= null) containingLocalFolder = localFolder;
                else
                {
                    SuperController.LogError("BA.ResourceVersionGroupEntry.Init: Null folder for path "+folderPath+"\nPossibly related to change of case in folder name");
                    containingLocalFolder = null;
                }
            }

            
        }

        public virtual void RefreshUserPrefs()
        {
            throw new Exception("BA.ResourceVersionGroupEntry.RefreshUserPrefs: unexpected call for resource "+resourceFullPathReference);
        }
        public HashSet<Folder> allContainingLocalFolders { get
            {
                return containingLocalFolder.allContainingLocalFolders;
            } }

        public virtual void PersistCUP()
        {

        }

        public virtual JSONClass GetCoreJSON()
        {
            JSONClass jc = new JSONClass();

            jc["resourceFullFileName"] = resourceFolderPathFileName;
            jc["resourceType"] = EnumManifest.GetStoreVal(ResourceType.enumManifestName,resourceType);
            jc["presetAtomType"] = atomPresetResourceType;

            return (jc);
        }

        protected virtual void RestoreTags(JSONArray tagsArray)
        {
            foreach (JSONClass tagJC in tagsArray)
            {
                int tagCategory = EnumManifest.GetEnumValFromStoreVal(TagCategory.enumManifestName, tagJC["tagCategory"]);
                string tagName = tagJC["tagName"];

                if (tagCategory == TagCategory.userDefined) TagManifest.CreateTag(tagName, tagCategory, resourceType, atomPresetResourceType);
                if (tagCategory != TagCategory.combi) TagManifest.SetRVGETagged(this, tagName, true);
            }
        }
        protected virtual JSONNode GetTagsJSON()
        {
            JSONNode ja = new JSONArray();

            foreach(var tag in TagManifest.GetRVGETags(this))
            {
                JSONClass tagJC = new JSONClass();
                tagJC["tagName"] = tag.tagName;
                tagJC["tagCategory"] = tag.tagCategoryJSEnum.storeVal;

                ja[-1] = tagJC;
            }

            return (ja);
        }

        public static HashSet<DAZClothingItem> LockNonRealClothing(Atom atom)
        {
            var lockedDCIs = new HashSet<DAZClothingItem>();
#if VAM_GT_1_21
            foreach (var dci in AtomUtils.GetDAZClothingItems(atom))
            {
                if (!dci.isRealItem && !dci.locked)
                {
                    dci.SetLocked(true);
                    lockedDCIs.Add(dci);
                }
            }
#endif
            return lockedDCIs;
        }
        public static void UnlockNonRealClothing(HashSet<DAZClothingItem> lockedDCIs)
        {
#if VAM_GT_1_21
            foreach (var dci in lockedDCIs) dci.SetLocked(false);
#endif
        }

        public static void RemoveNonRealClothing(Atom atom)
        {

            JSONStorable receiver = atom.GetStorableByID("geometry");
#if VAM_GT_1_21
            foreach (var dci in AtomUtils.GetDAZClothingItems(atom))
            {
                if (!dci.isRealItem)
                {
                    if (receiver != null)
                    {
                        JSONStorableBool active = receiver.GetBoolJSONParam("clothing:" + dci.uid);
                        if (active != null) active.val = false;
                    }
                }
            }
#endif
        }

        public static JSONClass GetNonRealClothingPreset(string currentActionSelectedFile)
        {
            JSONClass appPresetJC = (JSONClass)SuperController.singleton.LoadJSON(SuperController.singleton.NormalizePath(currentActionSelectedFile));

            return (GetNonRealClothingPreset(appPresetJC));
        }

        public static JSONClass GetNonRealClothingPreset(JSONClass appPresetJC)
        {
            JSONClass nonRealClothingPresetJC = new JSONClass();
            nonRealClothingPresetJC["setUnlistedParamsToDefault"].AsBool = true;

            JSONArray nonRealClothingPresetStorablesJA = new JSONArray();
            nonRealClothingPresetJC["storables"] = nonRealClothingPresetStorablesJA;

            JSONClass nonRealClothingPresetGeometryJC = new JSONClass();
            nonRealClothingPresetGeometryJC["id"] = "geometry";
            JSONArray nonRealClothingPresetClothingGeometryJA = new JSONArray();
            nonRealClothingPresetGeometryJC["clothing"] = nonRealClothingPresetClothingGeometryJA;
            nonRealClothingPresetStorablesJA.Add(nonRealClothingPresetGeometryJC);

            

            if (appPresetJC["storables"] == null) return nonRealClothingPresetJC;

            var appPresetStorableJCs = new Dictionary<string, JSONClass>();

            foreach (JSONClass jc in appPresetJC["storables"].AsArray)
            {
                if (jc["id"] != null)
                {
                    string storableID = jc["id"].Value;
                    appPresetStorableJCs.Add(storableID, jc);
                }
            }
            HashSet<string> enabledNonRealClothingGeometryIDs = new HashSet<string>();

            if (appPresetStorableJCs.ContainsKey("geometry"))
            {
                var geometryJC = appPresetStorableJCs["geometry"];
                if (geometryJC["clothing"] != null)
                {
                    JSONArray clothingGeometryJA = geometryJC["clothing"].AsArray;
                    foreach (JSONClass clothingGeometryItemJC in clothingGeometryJA)
                    {
                        string internalID;
                        if (clothingGeometryItemJC["internalId"] != null) internalID = clothingGeometryItemJC["internalId"];
                        else internalID = FileManagerSecure.GetFileName(clothingGeometryItemJC["id"]);

                        if (clothingGeometryItemJC["enabled"] == null || clothingGeometryItemJC["enabled"].AsBool)
                        {
                            string itemControlStorableID = internalID + "ItemControl";
                            if (appPresetStorableJCs.ContainsKey(itemControlStorableID) && appPresetStorableJCs[itemControlStorableID]["isRealClothingItem"] != null)
                            {
                                bool isRealClothingItem = appPresetStorableJCs[itemControlStorableID]["isRealClothingItem"].AsBool;
                                if (!isRealClothingItem)
                                {
                                    enabledNonRealClothingGeometryIDs.Add(internalID);
                                    nonRealClothingPresetClothingGeometryJA.Add(clothingGeometryItemJC);
                                }
                            }

                        }
                    }
                }
            }

            foreach (var kvp in appPresetStorableJCs)
            {
                foreach (string enabledNonRealClothingGeometryID in enabledNonRealClothingGeometryIDs)
                {
                    if (kvp.Key.StartsWith(enabledNonRealClothingGeometryID + "WrapControl") || kvp.Key.StartsWith(enabledNonRealClothingGeometryID + "Sim") || kvp.Key.StartsWith(enabledNonRealClothingGeometryID + "ItemControl") || kvp.Key.StartsWith(enabledNonRealClothingGeometryID + "Material"))
                    {
                        nonRealClothingPresetStorablesJA.Add(kvp.Value);
                    }
                }
            }

            if (nonRealClothingPresetClothingGeometryJA.Count == 0) return null;

            return nonRealClothingPresetJC;
        }

        protected virtual void ActivateClothing(Atom atom)
        {

        }

        private IEnumerator _Activate(BrowserUI browserUI, ResourcePrimaryActionUI resourceActions, bool createNewAtom = false, string targetAtomName = null, Action<string> onAtomCreatedCallback = null)
        {
            ResourceBrowserUI resourceBrowserUI = browserUI.resourceBrowserUI;

            if (resourceCategory!=ResourceCategory.clothing || createNewAtom) resourceBrowserUI.SetButtonLoading(this);

            if (isNew)
            {
                isNew = false;
                resourceBrowserUI.SetNewIcon(this);
            }

            if (resourceTypeSettings.closeBrowserUIonResourceLoadJSB.val) browserUI.SetInactive();

            if (resourceType == ResourceType.scene)
            {
                if (browserUI.isVREnabled) browserUI.SetInactive();
                if (resourceActions.mergeLoadJSB.val) SuperController.singleton.LoadMerge(FileManagerSecure.NormalizePath(resourceFullPathReference));
                else SuperController.singleton.Load(FileManagerSecure.NormalizePath(resourceFullPathReference));
            }
            else if (resourceCategory == ResourceCategory.preset || resourceType == ResourceType.subScenes || resourceCategory == ResourceCategory.clothing || resourceType == ResourceType.cua)
            {               
                bool mergeLoad = resourceActions.mergeLoadJSB.val && !createNewAtom;

                if (createNewAtom)
                {
                    yield return SuperController.singleton.AddAtomByType(atomPresetResourceType, targetAtomName, true);
                }

                yield return new WaitForEndOfFrame();

                if (AtomUtils.GetAtomNames().Contains(targetAtomName))
                {
                    if (createNewAtom) resourceActions.targetAtomJSMultiEnum.mainVal = targetAtomName;
                    if (createNewAtom)
                    {
                        AtomUtils.TeleportAtomToSpawnPoint(targetAtomName, ResourcePositionActionUI.newAtomPosition.GetPosition(), ResourcePositionActionUI.newAtomPosition.GetRotation());

                    }
                   
                    Atom atom = SuperController.singleton.GetAtomByUid(targetAtomName);

                    if (resourceType==ResourceType.maleClothingItems && createNewAtom)
                    {
                        var geometryJS = atom.GetStorableByID("geometry");
                        if (geometryJS != null)
                        {
                            var characterJSC = geometryJS.GetStringChooserJSONParam("characterSelection");
                            if (characterJSC != null)
                            {
                                characterJSC.val = "Male Dummy";

                            }
                        }
                    }

                    if (resourceType == ResourceType.cua)
                    {
                        if (resourceActions.cuaLoader != null)
                        {
                            var assetURLJSS = resourceActions.cuaLoader.GetUrlJSONParam ("assetUrl");
                            assetURLJSS.val = SuperController.singleton.NormalizePath(resourceFullPathReference);
                            cuaLoadsTriggered.Add(this);
                            cuaLoadTiggeredTime = DateTime.Now;
                            cuaLoader = resourceActions.cuaLoader;
                        }
                    }
                    else if(resourceType == ResourceType.subScenes)
                    {
                        JSONStorable js = atom.GetStorableByID("SubScene");

                        JSONStorableUrl subScenePathJSON = js.GetUrlJSONParam("browsePath");
                        subScenePathJSON.val = SuperController.singleton.NormalizePath(resourceFullPathReference);
                    }
                    else if (resourceCategory == ResourceCategory.clothing)
                    {
                        ActivateClothing(atom);
                    }
                    else if (resourceCategory == ResourceCategory.preset)
                    {
                        LoadPreset(resourceFullPathReference, atom, resourceType, mergeLoad);
                    }

                    if (createNewAtom)
                    {
                        resourceActions.RefreshTargetAtomPopup();
                        resourceActions.targetAtomJSMultiEnum.mainVal = targetAtomName;
                    }
                    if (Settings.selectAtomOnLoadJSB.val && resourceActions.isAtomTargetable)
                    {
                        FreeControllerV3 fcV3 = atom.freeControllers.First(fc => fc.name == "control");
                        SuperController.singleton.SelectController(fcV3);
                    }

                    if (createNewAtom && onAtomCreatedCallback != null) onAtomCreatedCallback(targetAtomName);
                }
            }
            else if (resourceCategory == ResourceCategory.plugins)
            {
                Atom atom = null;
                if (resourceActions.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
                {
                    if (AtomUtils.GetAtomNames(includeNonPersonAtoms: true).Contains(resourceActions.targetAtomJSMultiEnum.mainVal))
                    {
                        atom = SuperController.singleton.GetAtomByUid(resourceActions.targetAtomJSMultiEnum.mainVal);

                    }
                }
                else if (resourceActions.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag)
                {
                    if (resourceActions.targetAtomJSMultiEnum.valTopEnum == PluginTargetTypes.sessionPlugins) atom = Globals.mvrScript.containingAtom;
                    else atom = SuperController.singleton.GetAtomByUid("CoreControl");
                }

                PC.LoadPluginAsPreset(atom, resourceFullPathReference, Settings.openPluginUIJSB.val);

                if (Settings.selectAtomOnLoadJSB.val && resourceActions.isAtomTargetable)
                {
                    FreeControllerV3 fcV3 = atom.freeControllers.First(fc => fc.name == "control");
                    SuperController.singleton.SelectController(fcV3);
                }
            }
            

            yield return new WaitForSeconds(1f);

            resourceActions.parentBrowser.resourceBrowserUI.UnsetButtonLoading(this);
        }

        public static void LoadPreset(string resourceFullPathReference, Atom targetAtom, int resourceType, bool mergeLoad, JSONClass presetJC=null)
        {
            PresetLockStore atomPresetLockStore = new PresetLockStore();
            if (targetAtom.type == "Person") atomPresetLockStore.StorePresetLocks(targetAtom, true, Settings.suppressClothingLoadJSB.val && resourceType == ResourceType.presetAppearance && !mergeLoad);


            if (resourceType == ResourceType.presetAppearance && Settings.suppressClothingLoadJSB.val && Settings.onlySuppressRealClothingJSB.val && !mergeLoad && Globals.isRealClothingAvailable)
            {
                RemoveNonRealClothing(targetAtom);
                JSONClass nonRealClothingPresetJC;
                if (presetJC == null) nonRealClothingPresetJC= GetNonRealClothingPreset(resourceFullPathReference);
                else nonRealClothingPresetJC=GetNonRealClothingPreset(presetJC);

                if (nonRealClothingPresetJC != null)
                {
                    var clothingPMC = targetAtom.presetManagerControls.First(x => x.name == "ClothingPresets");

                    bool clothingLockPreState = clothingPMC.lockParams;
                    clothingPMC.lockParams = false;
                    JSONStorable presetJS = targetAtom.GetStorableByID("ClothingPresets");
                    PresetManager pm = presetJS.GetComponentInChildren<PresetManager>();

                    targetAtom.SetLastRestoredData(nonRealClothingPresetJC, true, true);
                    pm.LoadPresetFromJSON(nonRealClothingPresetJC, true);
                    
                    clothingPMC.lockParams = clothingLockPreState;
                }
            }

            HashSet<DAZClothingItem> lockedDCIs = null;
            if (resourceType == ResourceType.presetClothing && !mergeLoad && Settings.onlyReplaceRealClothingJSB.val && Globals.isRealClothingAvailable && Globals.isRealClothingAvailable)
            {
                lockedDCIs = LockNonRealClothing(targetAtom);
            }

            if (resourceType == ResourceType.presetAppearance && Settings.suppressScaleChangeJSB.val && PC.patreonContentEnabled)
            {
                if (presetJC == null) PC.AppPresetSuppressScaleLoad(targetAtom, resourceFullPathReference, mergeLoad);
                else PC.AppPresetSuppressScaleLoad(targetAtom, presetJC, mergeLoad);
            }
            else
            {
                bool deletePresetPostLoad = false;
                JSONStorable js = null;
                if (resourceType == ResourceType.presetAppearance) js = targetAtom.GetStorableByID("AppearancePresets");
                else if (PC.patreonContentEnabled) js = PC.GetPresetStorable(targetAtom, resourceType);

                if (js != null)
                {
                    JSONStorableBool loadOnSelectJSB = js.GetBoolJSONParam("loadPresetOnSelect");

                    bool loadOnSelectPreState = loadOnSelectJSB.val;
                    loadOnSelectJSB.val = false;

                    bool preState1 = false;
                    bool preState2 = false;
                    bool preState3 = false;

                    JSONStorableUrl presetPathJSON = js.GetUrlJSONParam("presetBrowsePath");

                    string intialPresetPath = presetPathJSON.val;

                        
                    if (presetJC != null)
                    {
                        string presetFolder = Const.GetResourceRootFolderNames(resourceType, targetAtom.type).First();
                        string presetName = Utils.GetUniquePresetName(presetFolder, "BATempPreset");
                        resourceFullPathReference = presetFolder + "\\Preset_" + presetName + ".vap";

                        SuperController.singleton.SaveJSON(presetJC, resourceFullPathReference);

                        deletePresetPostLoad = true;
                    }


                    if (resourceType == ResourceType.presetMorphs)
                    {
                        JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                        preState1 = includePhysical.val;
                        includePhysical.val = Settings.morphPresetIncludePhysicalJSB.val;
                        JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                        preState2 = includeAppearance.val;
                        includeAppearance.val = Settings.morphPresetIncludeAppJSB.val;
                    }
                    if (resourceType == ResourceType.presetGeneral)
                    {
                        JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                        preState1 = includePhysical.val;
                        includePhysical.val = Settings.generalPresetIncludePhysicalJSB.val;

                        JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                        preState2 = includeAppearance.val;
                        includeAppearance.val = Settings.generalPresetIncludeAppJSB.val;

                        JSONStorableBool includePose = js.GetBoolJSONParam("includeOptional");
                        preState3 = includePose.val;
                        includePose.val = Settings.generalPresetIncludePoseJSB.val;
                    }

                    
                    presetPathJSON.val = SuperController.singleton.NormalizePath(resourceFullPathReference);

                    if (mergeLoad) js.CallAction("MergeLoadPreset");
                    else js.CallAction("LoadPreset");

                    if (deletePresetPostLoad)
                    {
                        PresetManager pm = js.GetComponentInChildren<PresetManager>();
                        pm.DeletePreset();

                        presetPathJSON.val = intialPresetPath;
                    }

                    loadOnSelectJSB.val = loadOnSelectPreState;

                    if (resourceType == ResourceType.presetMorphs)
                    {
                        JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                        includePhysical.val = preState1;
                        JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                        includeAppearance.val = preState2;
                    }
                    if (resourceType == ResourceType.presetGeneral)
                    {
                        JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                        includePhysical.val = preState1;

                        JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                        includeAppearance.val = preState2;

                        JSONStorableBool includePose = js.GetBoolJSONParam("includeOptional");
                        includePose.val = preState3;
                    }
                }
            }

            if (lockedDCIs != null) UnlockNonRealClothing(lockedDCIs);

            atomPresetLockStore.RestorePresetLocks(targetAtom);
        }


        public IEnumerator Activate(BrowserUI browserUI, ResourcePrimaryActionUI resourceActions, bool createNewAtom = false, string targetAtomName = null, Action<string> onAtomCreatedCallback=null)
        {
            try
            {
                return _Activate(browserUI, resourceActions, createNewAtom, targetAtomName, onAtomCreatedCallback);
            }
            catch (Exception e) {
                SuperController.LogError("BA.ResourceVersionGroupEntry.Activate: Exception caught:\n" + e);
                return null;
            }
        }

        private void QueuedImageLoadCallback(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi == queuedImage)
            {
                if (resourceSprite != null) GameObject.Destroy( resourceSprite.texture);
                resourceSprite = Sprite.Create(qi.tex, new Rect(0.0f, 0.0f, qi.tex.width, qi.tex.height), new Vector2(0.5f, 0.5f), 100.0f);               
                LoadResourceTexture(queuedButtonLoad);
            }
            else GameObject.Destroy(qi.tex);
            
        }

        private void LoadResourceTexture(UIDynamicButton button)
        {
            if (button != null)
            {
                Transform newImageTrans = button.button.transform.Find("ButtonSubImage");
                UnityEngine.UI.Image image = newImageTrans.GetComponent<UnityEngine.UI.Image>();
                image.sprite = resourceSprite;
            }
        }

        public void CancelButtonImageUpdate()
        {
            queuedButtonLoad = null;
        }

        private void QueuedImageReloadCallback(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi.tex != null)
            {
                resourceSprite = Sprite.Create(qi.tex, new Rect(0.0f, 0.0f, qi.tex.width, qi.tex.height), new Vector2(0.5f, 0.5f), 100.0f);
            }
            

            BrowserAssist.mainBrowserUI.resourceBrowserUI.ReloadSprite(this);
            if (BrowserAssist.callbackBrowserUI!= null) BrowserAssist.callbackBrowserUI.resourceBrowserUI.ReloadSprite(this);
            BrowserAssist.mainBrowserUI.resourceActionsUI.RefreshUI();
        }

        public void ReloadSprite()
        {
            if (resourceSprite != null)
            {
                GameObject.Destroy(resourceSprite.texture);
                GameObject.Destroy(resourceSprite);
            }
            
            resourceSprite = null;

            string url = FileManagerSecure.NormalizePath(resourceThumbnailReference);

            if (!FileManagerSecure.FileExists(url)) url = url.Substring(0, url.Length - 3) + "JPG";

            ImageUtils.QueueLoadTexture(url, QueuedImageReloadCallback, true);
        }

        public void LoadSprite(UIDynamicButton button)
        {
            string url = FileManagerSecure.NormalizePath( resourceThumbnailReference);

            Transform newImageTrans = button.button.transform.Find("ButtonSubImage");
            GameObject newImageGO = newImageTrans.gameObject;           

            UnityEngine.UI.Image image = newImageGO.GetComponent<UnityEngine.UI.Image>();

            if (!FileManagerSecure.FileExists(url)) url = url.Substring(0, url.Length - 3)+"JPG";

            if (string.IsNullOrEmpty(url) || !FileManagerSecure.FileExists(url))
            {
                resourceSprite = null;
                image.sprite = null;
                return;
            }
            if (resourceSprite != null)
            {
                LoadResourceTexture(button);
            }
            else 
            {
                image.sprite = null;
                queuedButtonLoad = button;
                queuedImage = ImageUtils.QueueLoadTexture(url, QueuedImageLoadCallback,true);
            }
           
        }

        public virtual bool IsTagged(Tag tag)
        {
            return TagManifest.IsRVGETagged(this, tag);
        }

        public virtual bool IsNotTagged()
        {
            return TagManifest.IsRVGENotTagged(this);
        }

        public virtual bool IsNotUserTagged()
        {
            return TagManifest.IsRVGENotUserTagged(this);
        }

        public virtual bool IsWornInScene()
        {
            return false;
        }

        public virtual void SetTag(CombiTag combiTag)
        {
            SetTag(combiTag, combiTag.GetState(this));
        }

        public virtual void SetTag(Tag tag, bool value)
        {
            bool currentTagValue = TagManifest.IsRVGETagged(this, tag);

            if (value != currentTagValue)
            {
                TagManifest.SetRVGETagged(this, tag, value);
                foreach (CombiTag combiTag in tag.usedByCombiTags) SetTag(combiTag);

                if (tag.hideTaggedResourcesJSB.val)
                {
                    BrowserAssist.mainBrowserUI.resourceBrowserUI.SetHideFavColorTags(this);
                    if (!BrowserAssist.mainBrowserUI.browserFiltersUI.showHiddenJSB.val) BrowserAssist.mainBrowserUI.RefreshFilterResults();
                    if (BrowserAssist.callbackBrowserUI != null)
                    {
                        BrowserAssist.callbackBrowserUI.resourceBrowserUI.SetHideFavColorTags(this);
                        if (!BrowserAssist.callbackBrowserUI.browserFiltersUI.showHiddenJSB.val) BrowserAssist.callbackBrowserUI.RefreshFilterResults();
                    }
                }
                SetUserDataPersist();
            }
            
        }
    }

    public class LocalRVGE :ResourceVersionGroupEntry
    {
        public override string creatorName { get { return ""; } }
        public override string packageName { get { return ""; } }

        private LocalResourceUserDataFile _userDataFile = null;
        public LocalResourceUserDataFile userDataFile
        {
            get
            {
                if (_userDataFile == null) DataFileManifest.RegisterNewDFUserLocalResource(this);

                return _userDataFile;
            }
            private set {
                    _userDataFile = value;
            }

        }

        public override bool isCSLReferenced
        {
            get
            {
                return csListReferenced;
            }
        }
        public override DateTime firstCataloguedDateTime
        {
            get {
                if (Globals.isFileCreateModifiedDateAvailable) return localFileResourceCreatedDate;
                else return localFileResourceFirstScan;
            }
        }

        // For Local resources, first and last dates are the same (unfortunately)
        public override DateTime lastCataloguedDateTime
        {
            get {
                if (Globals.isFileCreateModifiedDateAvailable) return localFileResourceModifiedDate;
                else return localFileResourceFirstScan; }
        }
        public override string resourceThumbnailReference
        {
            get
            {
                return resourceFolderPathFileNameNoExt + ".jpg";
            }
        }
        private DateTime localFileResourceFirstScan;
        private DateTime localFileResourceCreatedDate;
        private DateTime localFileResourceModifiedDate;

        private bool isLocalResourceSymLinkAvaliable
        {
            get
            {
                if (resourceCategory == ResourceCategory.preset && Globals.customFolderAccess) return true;
                if (resourceCategory == ResourceCategory.sceneOrSessionPluginPresets && Globals.customFolderAccess) return true;
                if (resourceCategory == ResourceCategory.scene && Globals.scenesFolderAccess) return true;
                return false;
            }
        }

        public LocalRVGE(string resourceFileNamePath, int _resourceType) : base()
        {
            resourceType = _resourceType;
            resourceFolderPathFileName = resourceFileNamePath;
            localFileResource = true;
            UpdateLocalFileCreateUpdateDates();

            Init();
            RefreshLocalHideFavFlags();
        }

        public void UpdateLocalFileCreateUpdateDates()
        {
            if (Globals.isFileCreateModifiedDateAvailable)
            {
                localFileResourceCreatedDate = Utils.FileCreationTime(resourceFolderPathFileName);
                localFileResourceModifiedDate = Utils.FileLastWriteTime(resourceFolderPathFileName);
            }
        }

        protected override void Init()
        {
            base.Init();

            resourceFullPathReference = resourceFolderPathFileName;
            resourceFullPathReferenceWithAliasLowerCaps = (resourceFullPathReference + resourceNameAliasJSS.val).ToLower();
        }

        public void ResetFolderPath()
        {
            resourceFolderPathFileName = containingLocalFolder.GetFolderPath() + "\\" + resourceFileNameNoExt+"."+resourceFileNameExtension;
            Init();
            SetUserDataPersist();
        }

        public void MoveFolderPath(Folder targetFolder)
        {
            containingLocalFolder = targetFolder;
            ResetFolderPath();
        }

        public void MoveRenameFolderPath(string newFullPathFileName, Folder targetFolder)
        {
            containingLocalFolder = targetFolder;
            resourceFolderPathFileName = targetFolder.GetFolderPath()+"\\"+newFullPathFileName;
            Init();
            ResetFolderPath();
        }

        public void RenameRVGEFilename(string newFullPathFileName)
        {
            resourceFolderPathFileName = newFullPathFileName;
            Init();
            SetUserDataPersist();
        }

        protected override void SetVAMUserHidden(bool value)
        {
            bool success = SetVAMUserHideFav(value, ".hide");
            if (success) vamUserHidden = value;
        }
        protected override void SetVAMUserFavourite(bool value)
        {
            bool success = SetVAMUserHideFav(value, ".fav");
            if (success) vamUserFavorite = value;
        }

        private bool csListReferenced = false;

        private bool SetVAMUserHideFav(bool value, string fileExt)
        {
            if (isLocalResourceSymLinkAvaliable)
            {
                string path = Const.SymLinksFolder + "\\" + folderPath;
                string tagFileName = path + "\\" + resourceFileNameNoExt + "." + resourceFileNameExtension + fileExt;
                if (!FileManagerSecure.DirectoryExists(path) && value) FileManagerSecure.CreateDirectory(path);
                if (FileManagerSecure.DirectoryExists(path))
                {
                    if (value) FileManagerSecure.WriteAllBytes(tagFileName, new byte[] { 0 });
                    else FileManagerSecure.DeleteFile(tagFileName);
                    return true;
                }
            }
            return false;
        }

        public void SetLocalFileScanTime(DateTime scanTime)
        {
            localFileResourceFirstScan = scanTime;
            SetUserDataPersist();
        }

        public override void SetUserDataPersist()
        {
            DataFileManifest.RegisterLocalResourceUserSaveRequest(userDataFile);
        }

        public virtual void LoadLocalUserData(JSONClass resourceJC)
        {

            if (resourceJC["localFileFirstScanDateTime"] != null) localFileResourceFirstScan = DateTime.Parse(resourceJC["localFileFirstScanDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);

            if (resourceJC["baFavourite"] != null) baFavorite = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baFavourite"]);
            if (resourceJC["baHidden"] != null) baHidden = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baHidden"]);
            if (resourceJC["baNew"] != null) _isNew = resourceJC["baNew"].AsBool;
            else _isNew = false;


            if (resourceJC["Tags"] != null) RestoreTags(resourceJC["Tags"].AsArray);

            if (resourceJC[resourceNameAliasJSS.name] != null) resourceNameAliasJSS.valNoCallback = resourceJC[resourceNameAliasJSS.name];
            if (resourceJC[notesJSS.name] != null) notesJSS.valNoCallback = resourceJC[notesJSS.name];

            if (resourceCategory == ResourceCategory.clothing && primaryClothingItemPresetRVGE != null)
            {
                if (resourceJC["primaryClothingPreset"] != null)
                {
                    primaryClothingItemPresetRVGKey = resourceJC["primaryClothingPreset"];
                    ResourceManifest.RegisterPrimaryClothingItemPresetRVGE(this);
                }
            }

            Init();
        }

        public override JSONClass GetCoreJSON()
        {
            JSONClass resourceJC = base.GetCoreJSON();

            resourceJC["localFileFirstScanDateTime"] = localFileResourceFirstScan.ToString("O");
            if (baFavorite != TriState.indeterminate) resourceJC["baFavourite"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baFavorite);
            if (baHidden != TriState.indeterminate) resourceJC["baHidden"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baHidden);
            if (isNew != false) resourceJC["baNew"].AsBool = isNew;

            if (TagManifest.GetRVGETags(this).Count>0) resourceJC["Tags"] = GetTagsJSON();

            if (resourceCategory==ResourceCategory.clothing && primaryClothingItemPresetRVGE != null)
            {
                resourceJC["primaryClothingPreset"] = primaryClothingItemPresetRVGE.rvgeKey;
            }

            resourceNameAliasJSS.StoreJSON(resourceJC);
            notesJSS.StoreJSON(resourceJC);

            return resourceJC;
        }

        private void RefreshLocalHideFavFlags()
        {
            if (FileManagerSecure.FileExists(resourceFolderPathFileName + ".hide")) vamUserHidden = true;
            else vamUserHidden = false;
            if (FileManagerSecure.FileExists(resourceFolderPathFileName + ".fav")) vamUserFavorite = true;
            else vamUserFavorite = false;
        }

        public void RegisterUserDataFile(LocalResourceUserDataFile userDataFile)
        {
            this.userDataFile = userDataFile;
        }

        public void DeRegisterUserDataFile()
        {
            this.userDataFile = null;
        }
    }

    public class PackagedRVGE : ResourceVersionGroupEntry
    {
        public VARPackageVersionGroupEntry vpvge { get; private set; }

        private HashSet<int> varVersionsCreatorHidden = new HashSet<int>();
        private HashSet<int> varVersionsUserHidden = new HashSet<int>();
        private HashSet<int> varVersionsCreatorFav = new HashSet<int>();
        private HashSet<int> varVersionsUserFav = new HashSet<int>();
        private HashSet<int> varVersions = new HashSet<int>();
        private HashSet<int> varVersionsCSListReferenced = new HashSet<int>();

        private VARResourcesUserDataFile _userDataFile = null;

        public VARResourcesCoreDataFile coreDataFile { get; private set; }
        public VARResourcesUserDataFile userDataFile
        {
            get
            {
                if (_userDataFile == null) DataFileManifest.RegisterNewDFUserVARResource(this);
                return _userDataFile;
            }
            private set { _userDataFile = value; }
        }

        public int highestVARVersion { get; private set; }
        private bool packageCreatorHidden { get; set; } = false;
        private bool packageCreatorFav { get; set; } = false;

        public int varLicenseType { get { return vpvge.varVersionLicenses[highestVARVersion]; } }

        public override string creatorName
        {
            get {return vpvge.creatorName; }
        }
        public override string packageName
        {
            get { return vpvge.packageName; }
        }

        public virtual string packageUID
        {
            get { return vpvge.packageUID; }
        }

        public virtual string packageVersionUID
        {
            get { return vpvge.packageUID+"."+highestVARVersion.ToString(); }
        }

        protected override bool vamUserHidden { get
            {
                if (varVersionsUserHidden.Count == 0) return false;
                if (varVersionsUserHidden.Max() == highestVARVersion) return true;
                return false;
            }
        }
        protected override bool vamUserFavorite { get
            {
                if (varVersionsUserFav.Count == 0) return false;
                if (varVersionsUserFav.Max() == highestVARVersion) return true;
                return false;
            }
        }

        public string customThumbnailFolder { get {
            return Const.VARResourceThumbnails + "\\" + vpvge.varGroupName + "\\" + folderPath;
            } }

        public string customThumbnailFile
        {
            get
            {
                return customThumbnailFolder+"\\"+resourceFileNameNoExt+".jpg";
            }
        }

        public override bool isCSLReferenced
        {
            get
            {
                if (varVersionsCSListReferenced.Contains(highestVARVersion)) return true;
                return false;
            }
        }

        public override DateTime firstCataloguedDateTime
        {
            get {
               return vpvge.firstCataloguedDateTime; }
        }

        public override DateTime lastCataloguedDateTime
        {
            get
            {
                return vpvge.lastCataloguedDateTime;
            }
        }

        public override bool resourcePresent
        {
            get
            {
                if (vpvge.highestVARVersion == highestVARVersion) return true;
                return false;
            }
        }

        public override bool isMultiVARVersion { get { return (varVersions.Count>1); } }
        public int versionsCount { get { return varVersions.Count(); } }

        public override bool isHidden
        {
            get
            {
                return (vamUserHidden || packageCreatorHidden);
            }
        }

        public override bool isFavorite
        {
            get
            {
                return (baFavorite==TriState.active || vamUserFavorite || (packageCreatorFav && baFavorite!=TriState.inactive) );
            }
        }

        public override bool isMissing
        {
            get
            {
                return vpvge.IsVARMissing(highestVARVersion);
            }
        }

        public override void SetUserDataPersist()
        {
            DataFileManifest.RegisterVARResourceUserSaveRequest(userDataFile);
        }
        public void SetCoreDataPersist()
        {
            DataFileManifest.RegisterVARResourceCoreSaveRequest(coreDataFile);
        }

        
        protected override void SetVAMUserFavourite(bool value)
        {
            bool success=SetVAMUserHideFav(value, ".fav");
            if (success) {
                if (value) varVersionsUserFav.Add(highestVARVersion);
                else varVersionsUserFav.Remove(highestVARVersion);
            }
        }
        protected override void SetVAMUserHidden(bool value)
        {
            bool success = SetVAMUserHideFav(value, ".hide");
            if (success)
            {
                if (value) varVersionsUserHidden.Add(highestVARVersion);
                else varVersionsUserHidden.Remove(highestVARVersion);
            }
        }

        private bool SetVAMUserHideFav(bool value, string fileExt)
        {
            if (Globals.addonPackageFilePrefsFolderAccess)
            {
                string path = Const.SymLinksFolder + "\\AddonPackagesFilePrefs\\" + packageVersionUID + "\\" + folderPath;
                string tagFileName = path + "\\" + resourceFileNameNoExt + "." + resourceFileNameExtension + fileExt;
                if (!FileManagerSecure.DirectoryExists(path) && value) FileManagerSecure.CreateDirectory(path);
                if (FileManagerSecure.DirectoryExists(path))
                {
                    if (value) FileManagerSecure.WriteAllBytes(tagFileName, new byte[] { 0 });
                    else FileManagerSecure.DeleteFile(tagFileName);
                }
                return true;
            }
            return false;
        }

        protected override void Init()
        {
            base.Init();

            resourceFullPathReference = packageVersionUID + ":\\" + resourceFolderPathFileName;
            resourceFullPathReferenceWithAliasLowerCaps = (resourceFullPathReference+resourceNameAliasJSS.val).ToLower();
            
        }
        public override string resourceThumbnailReference
        {
            get
            {
                if (FileManagerSecure.FileExists(customThumbnailFile)) return customThumbnailFile;

                string thumbailFileRef= packageVersionUID + ":\\" + resourceFolderPathFileNameNoExt + ".jpg";
 //               if (!FileManagerSecure.FileExists(thumbailFileRef)) thumbailFileRef = packageVersionUID + ":\\" + resourceFolderPathFileNameNoExt + ".jpg";

                if (resourceType == ResourceType.plugins && !FileManagerSecure.FileExists(thumbailFileRef))
                {
                    string pluginThumnailsVARRef = Utils.GetLatestVARPath(Const.resourceThumbailsVAR + ".latest:\\Saves\\PluginData\\JayJayWon\\BrowserAssist\\VARPluginThumbnails\\" + packageUID +".jpg");
                    
                    if (FileManagerSecure.FileExists(pluginThumnailsVARRef)) return pluginThumnailsVARRef;
                }

                return thumbailFileRef;
            }
        }

        public PackagedRVGE(VARPackageManifestEntry vpme, string resourceFileName, int _resourceType, bool setNew) : base()
        {
            
            vpvge = vpme.vpvge;
            
            resourceType = _resourceType;
            resourceFolderPathFileName = resourceFileName;        

            highestVARVersion = vpme.varVersion;
            varVersions.Add(vpme.varVersion);

            Init();

            if (setNew) isNew = true;
        }

        public PackagedRVGE(JSONClass jc, VARPackageVersionGroupEntry _vpvge) : base()
        {
            vpvge = _vpvge;

            if (jc["resourceFullFileName"] != null) resourceFolderPathFileName = jc["resourceFullFileName"];
            if (jc["resourceType"] != null) resourceType = EnumManifest.GetEnumValFromStoreVal(ResourceType.enumManifestName, jc["resourceType"]);


            if (jc["varVersions"] != null)
            {
                JSONArray varVersionsJA = jc["varVersions"].AsArray;
                foreach (JSONNode jn in varVersionsJA)
                {
                    int varVersion = jn.AsInt;
                    if (vpvge.varVersionLicenses.ContainsKey(varVersion)) InternalAddVARVersion(varVersion,false);

                }
            }

            if (jc["varCreatorHidVersions"]!=null)
            {
                JSONArray varVersionsCreatorHidJA = jc["varCreatorHidVersions"].AsArray;
                foreach (JSONNode jn in varVersionsCreatorHidJA) varVersionsCreatorHidden.Add(jn.AsInt);
            }

            if (jc["varCreatorFavVersions"] != null)
            {
                JSONArray varVersionsCreatorFavJA = jc["varCreatorFavVersions"].AsArray;
                foreach (JSONNode jn in varVersionsCreatorFavJA) varVersionsCreatorFav.Add(jn.AsInt);
            }               

            if (jc["varVersionsCSLReferenced"] != null)
            {
                JSONArray varVersionsCSLReferencedJA = jc["varVersionsCSLReferenced"].AsArray;
                foreach (JSONNode jn in varVersionsCSLReferencedJA) varVersionsCSListReferenced.Add(jn.AsInt);
            }           

            Init();
       
        }

        public void RegisterCoreDataFile(VARResourcesCoreDataFile coreDataFile)
        {
            this.coreDataFile = coreDataFile;
        }
        public void RegisterUserDataFile(VARResourcesUserDataFile userDataFile)
        {
            this.userDataFile = userDataFile;
        }

        public virtual void RestoreVARUserData(JSONClass resourceJC)
        {

            if (resourceJC["baFavourite"] != null) baFavorite = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baFavourite"]);
            if (resourceJC["baHidden"] != null) baHidden = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baHidden"]);
            if (resourceJC["baNew"] != null) _isNew = resourceJC["baNew"].AsBool;

            if (resourceJC["Tags"] != null) RestoreTags(resourceJC["Tags"].AsArray);

            resourceNameAliasJSS.RestoreFromJSON(resourceJC);        

            if (resourceCategory == ResourceCategory.clothing)
            {
                if (resourceJC["primaryClothingPreset"] != null)
                {
                    primaryClothingItemPresetRVGKey = resourceJC["primaryClothingPreset"];
                    ResourceManifest.RegisterPrimaryClothingItemPresetRVGE(this);
                }
            }
        }


        private bool isUserDataDefault { get
            {
                if (baFavorite != TriState.indeterminate) return false;
                if (baHidden != TriState.indeterminate) return false;
                if (isNew) return false;
                if (TagManifest.GetRVGETags(this).Count > 0) return false;

                return true;
            }
        }
        public virtual JSONClass GetVARUserDataJSON()
        {
//            if (isUserDataDefault) return null;

            JSONClass resourceJC = new JSONClass();

            resourceJC["creatorName"] = creatorName;
            resourceJC["packageName"] = packageName;
            resourceJC["resourceFullFileName"] = resourceFolderPathFileName;

            if (baFavorite != TriState.indeterminate)  resourceJC["baFavourite"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baFavorite);
            if (baHidden != TriState.indeterminate) resourceJC["baHidden"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baHidden);
            if (isNew) resourceJC["baNew"].AsBool = isNew;

            if (TagManifest.GetRVGETags(this).Count > 0) resourceJC["Tags"] = GetTagsJSON();

            resourceNameAliasJSS.StoreJSON(resourceJC);

            if (resourceCategory == ResourceCategory.clothing && primaryClothingItemPresetRVGE != null)
            {
                resourceJC["primaryClothingPreset"] = primaryClothingItemPresetRVGE.rvgeKey; ;
            }

            return resourceJC;
        }

        public override JSONClass GetCoreJSON()
        {
            JSONClass jc = base.GetCoreJSON();

            jc["creatorName"] = creatorName;
            jc["packageName"] = packageName;

            JSONArray varVersionsJA = new JSONArray();
            jc["varVersions"] = varVersionsJA;

            foreach (int varVersion in varVersions)  varVersionsJA[-1].AsInt = varVersion;

            if (varVersionsCreatorHidden.Count > 0)
            {
                JSONArray varVersionsCreatorHidJA = new JSONArray();
                jc["varCreatorHidVersions"] = varVersionsCreatorHidJA;
                foreach (int varVersion in varVersionsCreatorHidden) varVersionsCreatorHidJA[-1].AsInt = varVersion;
            }
            
            if (varVersionsCreatorFav.Count > 0)
            {
                JSONArray varVersionsCreatorFavJA = new JSONArray();
                jc["varCreatorFavVersions"] = varVersionsCreatorFavJA;
                foreach (int varVersion in varVersionsCreatorFav) varVersionsCreatorFavJA[-1].AsInt = varVersion;
            }
            
            if (varVersionsCSListReferenced.Count > 0)
            {
                JSONArray varVersionsCSLReferencedJA = new JSONArray();
                jc["varVersionsCSLReferenced"] = varVersionsCSLReferencedJA;
                foreach (int varVersion in varVersionsCSListReferenced) varVersionsCSLReferencedJA[-1].AsInt = varVersion;
            }            

            return (jc);
        }
        public virtual void PurgeVARVersion(int varVersion)
        {
            varVersions.Remove(varVersion);
            varVersionsCreatorFav.Remove(varVersion);
            varVersionsCreatorHidden.Remove(varVersion);
            varVersionsUserFav.Remove(varVersion);
            varVersionsUserHidden.Remove(varVersion);
            varVersionsCSListReferenced.Remove(varVersion);

            if (highestVARVersion == varVersion)
            {
                highestVARVersion = -1;
                foreach(int remainingVARVersion in varVersions)
                {
                    if (remainingVARVersion > highestVARVersion) highestVARVersion = remainingVARVersion;
                }
                if (varVersionsCreatorFav.Contains(highestVARVersion)) packageCreatorFav = true;
                else packageCreatorFav = false;
                if (varVersionsCreatorHidden.Contains(highestVARVersion)) packageCreatorHidden = true;
                else packageCreatorHidden = false;

                Init();
            }
        }

        private void InternalAddVARVersion(int varVersion, bool setNew)
        {
            varVersions.Add(varVersion);

            if (varVersion > highestVARVersion)
            {
                highestVARVersion = varVersion;
                if (setNew) isNew = true;
                if (varVersionsCreatorFav.Contains(varVersion)) packageCreatorFav = true;
                else packageCreatorFav = false;
                if (varVersionsCreatorHidden.Contains(varVersion)) packageCreatorHidden = true;
                else packageCreatorHidden = false;
                Init();
            }
        }
        public virtual void AddVARVersion(int varVersion, bool setNew)
        {
            InternalAddVARVersion(varVersion, setNew);
            SetCoreDataPersist();
        }

        public void NewVARVersions(HashSet<int> newVARVersions)
        {            
            foreach (int varVersion in newVARVersions.Union(varVersions))
            {
                if (baHidden == TriState.active) SetVAMUserHidden(true);
                if (baFavorite == TriState.active && resourceCategory!=ResourceCategory.clothing) SetVAMUserFavourite(true);
            }
        }

        public void SetHideCreatorVARVersion(int varVersion, bool hide)
        {
            if (hide) varVersionsCreatorHidden.Add(varVersion);
            else varVersionsCreatorHidden.Remove(varVersion);

            packageCreatorHidden = (varVersionsCreatorHidden.Contains(highestVARVersion));
        }

        public void SetFavCreatorVARVersion(int varVersion, bool fav)
        {
            if (fav) varVersionsCreatorFav.Add(varVersion);
            else varVersionsCreatorFav.Remove(varVersion);

            packageCreatorFav = (varVersionsCreatorFav.Contains(highestVARVersion));
        }

        public void SetHideUserVARVersion(int varVersion, bool hide)
        {
            if (hide) varVersionsUserHidden.Add(varVersion);
            else varVersionsUserHidden.Remove(varVersion);
            if (varVersion == highestVARVersion)
            {
                if (hide)
                {
                    if (baHidden != TriState.active)
                    {
                        baHidden = TriState.active;
                        vpvge.SetPersistVARGroupUserData();
                    }
                    
                }
                else
                {
                    if (baHidden != TriState.inactive)
                    {
                        baHidden = TriState.inactive;
                        vpvge.SetPersistVARGroupUserData();
                    }
                }
            }
        }
        
        public void SetFavUserVARVersion(int varVersion, bool fav)
        {
            if (fav) varVersionsUserFav.Add(varVersion);
            else varVersionsUserFav.Remove(varVersion);
            if (varVersion == highestVARVersion)
            {
                if (fav)
                {
                    if (baFavorite != TriState.active)
                    {
                        baFavorite = TriState.active;
                        vpvge.SetPersistVARGroupUserData();
                    }
                    
                }
                else
                {
                    if (baFavorite != TriState.inactive)
                    {
                        baFavorite = TriState.inactive;
                        vpvge.SetPersistVARGroupUserData();
                    }
                }
            }
        }

        public void ClearAllUserFilePrefs()
        {
            varVersionsUserHidden.Clear();
            varVersionsUserFav.Clear();
        }

        public void SetCSLFileReferenced(int varVersion)
        {
            varVersionsCSListReferenced.Add(varVersion);
        }

        public bool IsInVersion (int varVersion)
        {
            if (varVersions.Contains(varVersion)) return true;
            return false;
        }
    }

    public class LocalClothingRVGE : LocalRVGE
    {
        protected string clothingItemDisplayName = "";
        protected bool clothingIsRealItemCreator = true;
        protected List<string> vamBuiltInTags = new List<string>();

        protected JSONClass clothingVAMJC =null;

        protected ClothingUserPrefs vamCUP = null;
        protected ClothingUserPrefs baCUP = null;
        public override string resourceDisplayName
        {
            get
            {
                if (resourceNameAliasJSS.val != "") return resourceNameAliasJSS.val;
                if (clothingItemDisplayName != "") return clothingItemDisplayName;
                return resourceFileNameNoExtNoPrefix;
            }
        }
        public override string resourceDisplayNameLowerCaps
        {
            get
            {
                return resourceDisplayName.ToLower();
            }
        }

        public LocalClothingRVGE(string resourceFileNamePath, int _resourceType, JSONClass clothingVAMJC) : base(resourceFileNamePath, _resourceType)
        {
            SetClothing(clothingVAMJC);
        }

        private void SetClothing(JSONClass clothingVAMJC)
        {

            clothingItemDisplayName = clothingVAMJC["displayName"];
            clothingUID = clothingVAMJC["uid"];
            clothingCreatorName = clothingVAMJC["creatorName"];
            CreatorManifest.RegisterClothingCreator(clothingCreatorName);

            if (clothingVAMJC["isRealItem"] != null) clothingIsRealItemCreator = clothingVAMJC["isRealItem"].AsBool;

            if (clothingVAMJC["tags"] != null)
            {
                string tags = clothingVAMJC["tags"];
                if (!string.IsNullOrEmpty(tags) && tags != "") vamBuiltInTags = tags.Split(',').ToList();
            }

            this.clothingVAMJC = clothingVAMJC;
        }

        protected override void ActivateClothing(Atom atom)
        {
            var activeJSB = AtomUtils.GetClothingActive(atom, this);

            if (activeJSB != null) activeJSB.val = !activeJSB.val;

        }

        private void ApplyClothingTags(List<string> tagNames, bool value)
        {
            foreach (string rawTagName in tagNames)
            {
                string tagName = rawTagName.ToLower().Trim();

                if (tagName != "")
                {
                    int tagCategory = TagCategory.userDefined;
                    if (Const.BodyRegionTagNames.Contains(tagName)) tagCategory = TagCategory.bodyRegion;

                    if (resourceType == ResourceType.femaleClothingItems && Const.FemaleClothingTypeTagNames.Contains(tagName)) tagCategory = TagCategory.clothingType;

                    if (resourceType == ResourceType.maleClothingItems && Const.MaleClothingTypeTagNames.Contains(tagName)) tagCategory = TagCategory.clothingType;

                    if (tagCategory == TagCategory.userDefined) TagManifest.CreateTag(tagName, tagCategory, resourceType, atomPresetResourceType);


                    TagManifest.SetRVGETagged(this, tagName, value);
                    SetUserDataPersist();
                }
            }
        }

        private void PurgeUnusedBuiltInTags()
        {
            foreach (string rawTagName in vamBuiltInTags)
            {
                string tagName = rawTagName.ToLower().Trim();

                if (tagName != "")
                {
                    Tag tag = TagManifest.GetTag(resourceType, atomPresetResourceType, tagName);
                    if (tag != null && tag.tagCategory == TagCategory.userDefined && TagManifest.GetTaggedResourcesCount(tag) == 0)
                    {
                        TagManifest.DeleteTag(tag);
                    }

                }
            }
        }

        public override void RefreshUserPrefs()
        {
            if (vamCUP == null && baCUP == null)
            {
                if (FileManagerSecure.FileExists(resourceFolderPathFileNameNoExt + ".prefs"))
                {
                    vamCUP = new ClothingUserPrefs(resourceFolderPathFileNameNoExt + ".prefs");
                    if (vamCUP.vamReplaceBuiltInTags) ApplyClothingTags(vamBuiltInTags, false);
                    ApplyClothingTags(vamCUP.userTagNames, true);

                    SetUserDataPersist();
                }
            }
            else
            {
                if (vamCUP != null && !FileManagerSecure.FileExists(resourceFolderPathFileNameNoExt + ".prefs"))
                {
                    ApplyClothingTags(vamCUP.userTagNames, false);
                    if (vamCUP.vamReplaceBuiltInTags) ApplyClothingTags(vamBuiltInTags, true);
                    vamCUP = null;

                    //TBD should this be create BACUP?
                    baCUP = null;

                    SetUserDataPersist();
                }
            }
            
            if (FileManagerSecure.FileExists(resourceFolderPathFileNameNoExt + ".prefs"))
            {
                var cup = new ClothingUserPrefs(resourceFolderPathFileNameNoExt + ".prefs");

                if (baCUP == null || cup.lastModifiedDateTime != baCUP.lastModifiedDateTime)
                {
                    if (vamCUP != null)
                    {
                        //Must have been modified by VAM
                        if (cup.rawTagNames != vamCUP.rawTagNames)
                        {
                            // foreach new user tag set to true
                            ApplyClothingTags(cup.userTagNames.Except(vamCUP.userTagNames).ToList(), true);

                            // foreach removed user tag set to false
                            ApplyClothingTags(vamCUP.userTagNames.Except(cup.userTagNames).ToList(), false);
                        }
                        if (cup.vamReplaceBuiltInTags != vamCUP.vamReplaceBuiltInTags)
                        {
                            if (cup.vamReplaceBuiltInTags) ApplyClothingTags(vamBuiltInTags.Except(cup.userTagNames).ToList(), false);
                            else ApplyClothingTags(vamBuiltInTags, true);
                        }
                    }
                    else
                    {
                        ApplyClothingTags(cup.userTagNames, true);
                        if (cup.vamReplaceBuiltInTags) ApplyClothingTags(vamBuiltInTags.Except(cup.userTagNames).ToList(), false);
                        else ApplyClothingTags(vamBuiltInTags, true);
                    }
                    vamCUP = cup;
                }
            }
            
            PurgeUnusedBuiltInTags();
        }

        public override JSONClass GetCoreJSON()
        {
            JSONClass jc = base.GetCoreJSON();

            if (baCUP != null) jc["baClothingUserPrefs"] = baCUP.GetJSON();
            if (vamCUP != null) jc["vamClothingUserPrefs"] = vamCUP.GetJSON();

            return jc;

        }


        public override void LoadLocalUserData(JSONClass resourceJC)
        {
            base.LoadLocalUserData(resourceJC);

            if (resourceJC["baClothingUserPrefs"] != null) baCUP = new ClothingUserPrefs(resourceJC["baClothingUserPrefs"].AsObject);
            if (resourceJC["vamClothingUserPrefs"] != null) vamCUP = new ClothingUserPrefs(resourceJC["vamClothingUserPrefs"].AsObject);

            RefreshUserPrefs();
        }

        public override bool IsWornInScene()
        {
            if (resourceCategory != ResourceCategory.clothing) return false;

            bool isFemale = (resourceType == ResourceType.femaleClothingItems);

            var personAtoms = AtomUtils.GetPersonAtoms(isFemale, !isFemale);

            foreach (var personAtom in personAtoms)
            {
                var dcis = ResourcePrimaryActionUI.GetCachedDCIs(personAtom);

                if (dcis.Where(dci => dci.active && dci.internalUid == clothingUID).Any()) return true;

            }

            return false;
        }

        public override void PersistCUP()
        {
            if (Globals.customFolderAccess)
            {
                var jc = new JSONClass();

                bool replaceBuiltInTags = false;
                List<string> keepBuiltInTagNames = new List<string>();
                List<Tag> keepBuiltInTags = new List<Tag>();
                foreach (var rawTagName in vamBuiltInTags)
                {
                    if (rawTagName != "")
                    {
                        Tag tag = TagManifest.GetTag(resourceType, atomPresetResourceType, rawTagName.ToLower().Trim());
                        if (!IsTagged(tag)) replaceBuiltInTags = true;
                        else
                        {
                            keepBuiltInTagNames.Add(rawTagName);
                            keepBuiltInTags.Add(tag);
                        }
                    }

                }

                string rawTagsList = "";
                bool firstTag = true;

                jc["replaceBuiltInTags"].AsBool = replaceBuiltInTags;

                foreach (Tag tag in TagManifest.GetRVGETags(this))
                {
                    if (!keepBuiltInTags.Contains(tag))
                    {
                        if (firstTag) rawTagsList += tag.tagName;
                        else rawTagsList += "," + tag.tagName;
                        firstTag = false;
                    }
                }
                if (replaceBuiltInTags)
                {
                    foreach (string rawTagName in keepBuiltInTagNames)
                    {
                        if (firstTag) rawTagsList += rawTagName;
                        else rawTagsList += "," + rawTagName;
                        firstTag = false;
                    }
                }
                if (rawTagsList != "") jc["userTags"] = rawTagsList;

                string path = Const.SymLinksFolder + "\\" + resourceFolderPathFileNameNoExt + ".prefs";
                string folder = FileManagerSecure.GetDirectoryName(path);
                if (!FileManagerSecure.DirectoryExists(folder)) FileManagerSecure.CreateDirectory(folder);
                SuperController.singleton.SaveJSON(jc, Const.SymLinksFolder + "\\" + resourceFolderPathFileNameNoExt + ".prefs");
                baCUP = new ClothingUserPrefs(resourceFolderPathFileNameNoExt + ".prefs");
                SetUserDataPersist();
            }
        }

        public override void SetTag(Tag tag, bool value)
        {
            base.SetTag(tag, value);
            ResourceManifest.RegisterCUPUpdate(this);
        }
    }
    public class PackagedClothingRVGE : PackagedRVGE
    {
        protected string clothingItemDisplayName { get { return clothingItemDisplayNamesByVersion[highestVARVersion]; } }
        protected bool clothingIsRealItemCreator { get {
                if (!clothingIsRealItemCreatorByVersion.ContainsKey(highestVARVersion)) return true;
                return clothingIsRealItemCreatorByVersion[highestVARVersion];
            } }
        protected List<string> vamBuiltInTags { get {
                if (!vamBuiltInTagsByVersion.ContainsKey(highestVARVersion)) return new List<string>();
                return vamBuiltInTagsByVersion[highestVARVersion];
            } }
        public override string clothingUID
        {
            get { return clothingUIDsByVersion[highestVARVersion]; }
            protected set { throw new Exception("BA.PackagedClothingRVGE.clothingUID: Invalid set"); }
        }
        public override string clothingCreatorName
        {
            get
            {
                while (clothingCreatorNamesByVersion.Count > 0)
                {
                    if (clothingCreatorNamesByVersion.ContainsKey(highestVARVersion)) return clothingCreatorNamesByVersion[highestVARVersion];
                    else PurgeVARVersion(highestVARVersion);
                }
                return "";
                
            }
            protected set
            {
                throw new Exception("BA.PackagedClothingRVGE.clothingCreatorName: Invalid set");
            }
        }

        protected ClothingUserPrefs vamCUP =null;
        protected ClothingUserPrefs baCUP =null;

        protected Dictionary<int,string> clothingCreatorNamesByVersion= new Dictionary<int,string>();
        protected Dictionary<int, string> clothingItemDisplayNamesByVersion = new Dictionary<int, string>();
        protected Dictionary<int, bool> clothingIsRealItemCreatorByVersion = new Dictionary<int, bool>();
        protected Dictionary<int, string> clothingUIDsByVersion = new Dictionary<int, string>();
        protected Dictionary<int, List<string>> vamBuiltInTagsByVersion = new Dictionary<int, List<string>>();
        protected Dictionary<int, JSONClass> clothingVAMJCByVersion = new Dictionary<int, JSONClass>();

        public override string resourceDisplayName
        {
            get
            {
                if (resourceNameAliasJSS.val != "") return resourceNameAliasJSS.val;
                if (clothingItemDisplayName != "") return clothingItemDisplayName;
                return resourceFileNameNoExtNoPrefix;
            }
        }

        public override string resourceDisplayNameLowerCaps
        {
            get
            {
                return resourceDisplayName.ToLower();
            }
        }
        public PackagedClothingRVGE(VARPackageManifestEntry vpme, string resourceFileName, int _resourceType, bool setNew, JSONClass clothingVAMJC) : base(vpme, resourceFileName, _resourceType, setNew)
        {
            
            SetClothing(clothingVAMJC, vpme.varVersion);

            if (FileManagerSecure.FileExists(resourceFolderPathFileNameNoExt + ".prefs")) vamCUP = new ClothingUserPrefs(resourceFolderPathFileNameNoExt + ".prefs");

            if (vamCUP == null || !vamCUP.vamReplaceBuiltInTags) ApplyClothingTags(vamBuiltInTags,true);
            if (vamCUP != null) ApplyClothingTags(vamCUP.userTagNames, true);

            SetUserDataPersist();
        }

        public override void AddVARVersion(int varVersion, bool setNew)
        {
            JSONClass clothingVAMJC=null;
            try
            {
                clothingVAMJC = SuperController.singleton.LoadJSON(vpvge.packageUID + "." + varVersion.ToString() + ":\\" + resourceFolderPathFileName).AsObject;
            }
            catch(Exception ex) {
                SuperController.LogError("BA.PackagedClothingRVGE.AddVARVersion: Exception caught for '"+resourceFullPathReference+"' - "+ex);
                return;
            }
            
            base.AddVARVersion(varVersion, setNew);
            SetClothing(clothingVAMJC, varVersion);
        }

        public override void PurgeVARVersion(int varVersion)
        {
            clothingCreatorNamesByVersion.Remove(varVersion);
            clothingItemDisplayNamesByVersion.Remove(varVersion);
            clothingIsRealItemCreatorByVersion.Remove(varVersion);
            clothingUIDsByVersion.Remove(varVersion);
            vamBuiltInTagsByVersion.Remove(varVersion);
            clothingVAMJCByVersion.Remove(varVersion);
            base.PurgeVARVersion(varVersion);
        }

        private void SetClothing (JSONClass clothingVAMJC, int varVersion)
        {
            if (clothingItemDisplayNamesByVersion.ContainsKey(varVersion)) clothingItemDisplayNamesByVersion[varVersion] = clothingVAMJC["displayName"];
            else clothingItemDisplayNamesByVersion.Add(varVersion, clothingVAMJC["displayName"]);

            if (clothingUIDsByVersion.ContainsKey(varVersion)) clothingUIDsByVersion[varVersion] = clothingVAMJC["uid"];
            else clothingUIDsByVersion.Add(varVersion, clothingVAMJC["uid"]);

            if (clothingCreatorNamesByVersion.ContainsKey(varVersion)) clothingCreatorNamesByVersion[varVersion] = clothingVAMJC["creatorName"];
            else clothingCreatorNamesByVersion.Add(varVersion, clothingVAMJC["creatorName"]);

            CreatorManifest.RegisterClothingCreator(clothingCreatorNamesByVersion[varVersion]);

            if (clothingVAMJC["isRealItem"] != null)
            {
                if (clothingIsRealItemCreatorByVersion.ContainsKey(varVersion)) clothingIsRealItemCreatorByVersion[varVersion] = clothingVAMJC["isRealItem"].AsBool;
                else clothingIsRealItemCreatorByVersion.Add(varVersion, clothingVAMJC["isRealItem"].AsBool);
            }
            else clothingIsRealItemCreatorByVersion.Remove(varVersion);

            if (clothingVAMJC["tags"] != null && !string.IsNullOrEmpty(clothingVAMJC["tags"]) && clothingVAMJC["tags"] != "")
            {
                string tags = clothingVAMJC["tags"];
                List<string> vamBuiltInTags = tags.Split(',').ToList();
                if (vamBuiltInTagsByVersion.ContainsKey(varVersion)) vamBuiltInTagsByVersion[varVersion] = vamBuiltInTags;
                else vamBuiltInTagsByVersion.Add(varVersion, vamBuiltInTags);
            }

            if (clothingVAMJCByVersion.ContainsKey(varVersion)) clothingVAMJCByVersion[varVersion] = clothingVAMJC;
            else clothingVAMJCByVersion.Add(varVersion, clothingVAMJC);
        }

        public PackagedClothingRVGE(JSONClass jc, VARPackageVersionGroupEntry _vpvge) : base(jc, _vpvge)
        {

            if (jc["clothingVersions"] != null)
            {
                JSONArray clothingByVersionJA = jc["clothingVersions"].AsArray;

                if (clothingByVersionJA.Count == 0) throw new Exception("BA.PackagedClothingRVGE.const: no clothing versions registed");

                foreach (JSONClass clothingVersionJC in clothingByVersionJA)
                {
                    SetClothing(clothingVersionJC["clothing"].AsObject, clothingVersionJC["varVersion"].AsInt);
                }
            }
            
        }

        protected override void ActivateClothing(Atom atom)
        {
            var activeJSB = AtomUtils.GetClothingActive(atom, this);

            if (activeJSB!= null) activeJSB.val = !activeJSB.val; 
            
        }

        private void ApplyClothingTags(List<string> tagNames, bool value)
        {
            foreach (string rawTagName in tagNames)
            {
                string tagName = rawTagName.ToLower().Trim();

                if (tagName != "")
                {
                    int tagCategory = TagCategory.userDefined;
                    if (Const.BodyRegionTagNames.Contains(tagName)) tagCategory = TagCategory.bodyRegion;
                    if (resourceType == ResourceType.femaleClothingItems && Const.FemaleClothingTypeTagNames.Contains(tagName)) tagCategory = TagCategory.clothingType;

                    if (resourceType == ResourceType.maleClothingItems && Const.MaleClothingTypeTagNames.Contains(tagName)) tagCategory = TagCategory.clothingType;

                    if (tagCategory == TagCategory.userDefined) TagManifest.CreateTag(tagName, tagCategory, resourceType, atomPresetResourceType);

                    TagManifest.SetRVGETagged(this, tagName, value);

                    SetUserDataPersist();
                }
            }
        }

        private void PurgeUnusedBuiltInTags()
        {
            foreach (string rawTagName in vamBuiltInTags)
            {
                string tagName = rawTagName.ToLower().Trim();

                if (tagName != "")
                {
                    Tag tag = TagManifest.GetTag(resourceType, atomPresetResourceType, tagName);
                    if (tag != null && tag.tagCategory== TagCategory.userDefined && TagManifest.GetTaggedResourcesCount(tag) == 0)
                    {
                        TagManifest.DeleteTag(tag);
                    }

                }
            }
        }

        public override void RefreshUserPrefs()
        {
            if (vamCUP == null && baCUP == null)
            {
                if (FileManagerSecure.FileExists(resourceFolderPathFileNameNoExt + ".prefs"))
                {
                    vamCUP = new ClothingUserPrefs(resourceFolderPathFileNameNoExt + ".prefs");
                    if (vamCUP.vamReplaceBuiltInTags) ApplyClothingTags(vamBuiltInTags, false);
                    ApplyClothingTags(vamCUP.userTagNames, true);

                    SetUserDataPersist();
                }

            }
            else
            {

                if (vamCUP != null && !FileManagerSecure.FileExists(resourceFolderPathFileNameNoExt + ".prefs"))
                {
                    ApplyClothingTags(vamCUP.userTagNames, false);
                    if (vamCUP.vamReplaceBuiltInTags) ApplyClothingTags(vamBuiltInTags, true);
                    vamCUP = null;

                    //TBD should this be create BACUP?
                    baCUP = null;

                    SetUserDataPersist();
                }
            }


            if (FileManagerSecure.FileExists(resourceFolderPathFileNameNoExt + ".prefs"))
            {
                var cup = new ClothingUserPrefs(resourceFolderPathFileNameNoExt + ".prefs");

                if (baCUP==null || cup.lastModifiedDateTime != baCUP.lastModifiedDateTime)
                {
                    if (vamCUP != null)
                    {
                        //Must have been modified by VAM
                        if (cup.rawTagNames != vamCUP.rawTagNames)
                        {
                            // foreach new user tag set to true
                            ApplyClothingTags(cup.userTagNames.Except(vamCUP.userTagNames).ToList(), true);

                            // foreach removed user tag set to false
                            ApplyClothingTags(vamCUP.userTagNames.Except(cup.userTagNames).ToList(), false);
                        }
                        if (cup.vamReplaceBuiltInTags != vamCUP.vamReplaceBuiltInTags)
                        {
                            if (cup.vamReplaceBuiltInTags) ApplyClothingTags(vamBuiltInTags.Except(cup.userTagNames).ToList(), false);
                            else ApplyClothingTags(vamBuiltInTags, true);
                        }
                        
                    }
                    else
                    {
                        ApplyClothingTags(cup.userTagNames, true);
                        if (cup.vamReplaceBuiltInTags) ApplyClothingTags(vamBuiltInTags.Except(cup.userTagNames).ToList(), false);
                        else ApplyClothingTags(vamBuiltInTags, true);
                    }
                    vamCUP = cup;


                }
            }
            PurgeUnusedBuiltInTags();
        }

        public override JSONClass GetCoreJSON()
        {
            JSONClass jc = base.GetCoreJSON();

            JSONArray clothingByVersionJA = new JSONArray();

            foreach (var varVersionKVP in clothingVAMJCByVersion)
            {
                JSONClass varVersionClothingJC = new JSONClass();
                varVersionClothingJC["varVersion"].AsInt = varVersionKVP.Key;
                varVersionClothingJC["clothing"] = varVersionKVP.Value;
                clothingByVersionJA[-1] = varVersionClothingJC;
            }

            jc["clothingVersions"] = clothingByVersionJA;

            return jc;

        }

        public override JSONClass GetVARUserDataJSON()
        {
            JSONClass jc = base.GetVARUserDataJSON();

            if (baCUP != null) jc["baClothingUserPrefs"] = baCUP.GetJSON();
            if (vamCUP != null) jc["vamClothingUserPrefs"] = vamCUP.GetJSON();


            return jc;
        }

        public override void RestoreVARUserData(JSONClass resourceJC)
        {
            base.RestoreVARUserData(resourceJC);

            if (resourceJC["baClothingUserPrefs"] != null) baCUP = new ClothingUserPrefs(resourceJC["baClothingUserPrefs"].AsObject);
            if (resourceJC["vamClothingUserPrefs"] != null) vamCUP = new ClothingUserPrefs(resourceJC["vamClothingUserPrefs"].AsObject);

            RefreshUserPrefs();
        }

        public override bool IsWornInScene()
        {
            if (resourceCategory!= ResourceCategory.clothing) return false;

            bool isFemale = (resourceType == ResourceType.femaleClothingItems);

            var personAtoms = AtomUtils.GetPersonAtoms(isFemale, !isFemale);

            foreach (var personAtom in personAtoms)
            {
                var dcis = ResourcePrimaryActionUI.GetCachedDCIs(personAtom);

                if (dcis.Where(dci => dci.active && dci.internalUid == clothingUID && dci.packageUid == packageVersionUID).Any()) return true;
                
            }

            return false;
        }

        public override void PersistCUP()
        {
            if (Globals.customFolderAccess)
            {
                var jc = new JSONClass();

                bool replaceBuiltInTags = false;
                List<string> keepBuiltInTagNames = new List<string>();
                List<Tag> keepBuiltInTags = new List<Tag>();
                foreach (var rawTagName in vamBuiltInTags)
                {
                    if (rawTagName != "")
                    {
                        Tag tag = TagManifest.GetTag(resourceType, atomPresetResourceType, rawTagName.ToLower().Trim());
                        if (!IsTagged(tag)) replaceBuiltInTags = true;
                        else
                        {
                            keepBuiltInTagNames.Add(rawTagName);
                            keepBuiltInTags.Add(tag);
                        }
                    }
                    
                }

                string rawTagsList = "";
                bool firstTag = true;

                jc["replaceBuiltInTags"].AsBool = replaceBuiltInTags;

                foreach(Tag tag in TagManifest.GetRVGETags(this))
                {
                    if (!keepBuiltInTags.Contains(tag))
                    {
                        if (firstTag) rawTagsList += tag.tagName;
                        else rawTagsList += ","+tag.tagName;
                        firstTag = false;
                    }
                }
                if (replaceBuiltInTags)
                {
                    foreach (string rawTagName in keepBuiltInTagNames)
                    {
                        if (firstTag) rawTagsList += rawTagName;
                        else rawTagsList += "," + rawTagName;
                        firstTag = false;
                    }
                }
                if (rawTagsList!="") jc["userTags"] = rawTagsList;

                string path = Const.SymLinksFolder + "\\" + resourceFolderPathFileNameNoExt + ".prefs";
                string folder = FileManagerSecure.GetDirectoryName(path);
                if (!FileManagerSecure.DirectoryExists(folder))FileManagerSecure.CreateDirectory(folder);
                SuperController.singleton.SaveJSON(jc, Const.SymLinksFolder + "\\" + resourceFolderPathFileNameNoExt + ".prefs");
                baCUP = new ClothingUserPrefs(resourceFolderPathFileNameNoExt + ".prefs");
                SetUserDataPersist();
            }
        }

        public override void SetTag(Tag tag, bool value)
        {
            base.SetTag(tag, value);
            ResourceManifest.RegisterCUPUpdate(this);
        }
    }

    public class ClothingUserPrefs
    {

        public bool vamReplaceBuiltInTags { get; protected set; } = false;

        public List<string> userTagNames = new List<string>();

        public string rawTagNames { get; protected set; } = "";

        public DateTime lastModifiedDateTime { get; protected set; } = DateTime.MinValue;

        public ClothingUserPrefs(string filePath)
        {
            if (!FileManagerSecure.FileExists(filePath)) { throw new Exception("BA.ClothingUserPrefs.Constuctor: path '"+filePath+"' does not exist."); }

            JSONClass cupJC = SuperController.singleton.LoadJSON(filePath).AsObject;

            if (cupJC["replaceBuiltInTags"] != null) vamReplaceBuiltInTags = cupJC["replaceBuiltInTags"].AsBool;
            if (cupJC["userTags"] != null) rawTagNames = cupJC["userTags"];

            if (rawTagNames != null) userTagNames = rawTagNames.Split(',').ToList();

            lastModifiedDateTime = Utils.FileLastWriteTime(filePath);
        }

        public ClothingUserPrefs(JSONClass jc)
        {
            vamReplaceBuiltInTags = jc["vamReplaceBuiltInTags"].AsBool;
            rawTagNames = jc["userTags"];
            userTagNames = rawTagNames.Split(',').ToList();

            lastModifiedDateTime = DateTime.Parse(jc["lastModifiedDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);
        }

        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();

            jc["vamReplaceBuiltInTags"].AsBool = vamReplaceBuiltInTags;
            jc["userTags"] = rawTagNames;
            jc["lastModifiedDateTime"] = lastModifiedDateTime.ToString("O");

            return jc;
        }
    }

    public class CreatorManifest
    {
        private static Dictionary<string, CreatorManifestEntry> creatorManifestDict = new Dictionary<string, CreatorManifestEntry>();

        private static Dictionary<string, CreatorManifestEntry> clothingCreatorManifestDict = new Dictionary<string, CreatorManifestEntry>();

        public static CreatorManifestEntry RegisterCreator (string creatorName)
        {
            if (creatorManifestDict.ContainsKey(creatorName)) return creatorManifestDict[creatorName];
            CreatorManifestEntry cme = new CreatorManifestEntry(creatorName);
            creatorManifestDict.Add(creatorName,cme);
            return cme;
        }

        public static CreatorManifestEntry RegisterClothingCreator(string creatorName)
        {
            if (clothingCreatorManifestDict.ContainsKey(creatorName)) return clothingCreatorManifestDict[creatorName];
            CreatorManifestEntry cme = new CreatorManifestEntry(creatorName);
            clothingCreatorManifestDict.Add(creatorName, cme);
            return cme;
        }

        public static JSONArray GetJSON()
        {
            JSONArray creatorsJA = new JSONArray();

            foreach (var cmeKVP in creatorManifestDict)
            {
                JSONClass cmeJC = new JSONClass();
                cmeJC["creatorName"] = cmeKVP.Key;
                var hiddenJA = cmeKVP.Value.GetHiddenResourceTypes();
                if (hiddenJA != null) cmeJC["hiddenResourceTypes"] = hiddenJA;
                var hiddenAtomPresetJA = cmeKVP.Value.GetHiddenAtomPresetResourceTypes();
                if (hiddenAtomPresetJA != null) cmeJC["hiddenAtomPresetResourceTypes"] = hiddenAtomPresetJA;

                var prefJA = cmeKVP.Value.GetPreferredResourceTypes();
                if (prefJA != null) cmeJC["preferredResourceTypes"] = prefJA;
                var prefAtomPresetJA = cmeKVP.Value.GetPreferredAtomPresetResourceTypes();
                if (prefAtomPresetJA != null) cmeJC["preferredAtomPresetResourceTypes"] = prefAtomPresetJA;

                if (prefJA!=null || hiddenJA!=null || prefAtomPresetJA !=null || hiddenAtomPresetJA!=null) creatorsJA[-1] = cmeJC;
            }

            return creatorsJA;
        }

        public static void RestoreJSON(JSONArray creatorsJA)
        {
            foreach(JSONClass cmeJC in creatorsJA)
            {
                string creatorName = cmeJC["creatorName"];
                if (creatorManifestDict.ContainsKey(creatorName))
                {
                    if (cmeJC["hiddenResourceTypes"]!=null) creatorManifestDict[creatorName].RestoreHiddenResourceTypes(cmeJC["hiddenResourceTypes"].AsArray);
                    if (cmeJC["preferredResourceTypes"] != null) creatorManifestDict[creatorName].RestorePreferredResourceTypes(cmeJC["preferredResourceTypes"].AsArray);

                    if (cmeJC["hiddenAtomPresetResourceTypes"] != null) creatorManifestDict[creatorName].RestoreHiddenAtomPresetResourceTypes(cmeJC["hiddenAtomPresetResourceTypes"].AsArray);
                    if (cmeJC["preferredAtomPresetResourceTypes"] != null) creatorManifestDict[creatorName].RestorePreferredAtomPresetResourceTypes(cmeJC["preferredAtomPresetResourceTypes"].AsArray);
                }
            }
        }

        public static void RegisterResource (ResourceVersionGroupEntry rvge)
        {
            creatorManifestDict[rvge.creatorName].RegisterResource(rvge);
        }

        public static void RegisterClothingResource(ResourceVersionGroupEntry rvge)
        {           
            clothingCreatorManifestDict[rvge.clothingCreatorName].RegisterResource(rvge);            
        }

        public static void RemoveRVGE(ResourceVersionGroupEntry rvge)
        {
            creatorManifestDict[rvge.creatorName].RemoveRVGE(rvge);
        }

        public static void RemoveClothingRVGE(ResourceVersionGroupEntry rvge)
        {
            if (!rvge.localFileResource)
            {
                var packagedRVGE = rvge as PackagedRVGE;
                if (packagedRVGE.versionsCount == 0) return;
            }
            if (clothingCreatorManifestDict.ContainsKey(rvge.clothingCreatorName)) clothingCreatorManifestDict[rvge.clothingCreatorName].RemoveRVGE(rvge);
        }

        public static CreatorManifestEntry GetCreator(string creatorName)
        {
            if (!creatorManifestDict.ContainsKey(creatorName)) return null;

            return creatorManifestDict[creatorName];
        }

        public static HashSet<string> GetAllClothingCreatorNames()
        {
            return new HashSet<string>(clothingCreatorManifestDict.Keys);
        }

        public static List<string> GetAllCreatorNames()
        {
            return new List<string>(creatorManifestDict.Keys);
        }

        public static List<string> GetAllClothingCreatorNamesList()
        {
            return new List<string>(clothingCreatorManifestDict.Keys);
        }

        public static List<string> GetAllCreatorNames(int resourceType, string atomPresetResourceType)
        {
            List<string> creatorNames = new List<string>();

            foreach (var creatorKVP in creatorManifestDict)
            {
                if (creatorKVP.Value.HasResourceType(resourceType, atomPresetResourceType))
                {
                    creatorNames.Add(creatorKVP.Key);
                }
            }
                return creatorNames;
        }

        public static HashSet<CreatorManifestEntry> GetAllCreators(int resourceType, string atomPresetResourceType)
        {
            HashSet<CreatorManifestEntry> creators = new HashSet<CreatorManifestEntry>();

            foreach (var creatorKVP in creatorManifestDict)
            {
                if (creatorKVP.Value.HasResourceType(resourceType, atomPresetResourceType))
                {
                    creators.Add(creatorKVP.Value);
                }
            }
            return creators;
        }


        public static List<string> GetAllClothingCreatorNamesList(int resourceType, string atomPresetResourceType)
        {
            List<string> creatorNames = new List<string>();

            foreach (var creatorKVP in clothingCreatorManifestDict)
            {
                if (creatorKVP.Value.HasResourceType(resourceType, atomPresetResourceType))
                {
                    creatorNames.Add(creatorKVP.Key);
                }
            }
            return creatorNames;
        }


        public static void UpdateCreatorChoices(JSONStorableMultiEnumStringChooser creatorJSMultiEnum, HashSet<ResourceVersionGroupEntry> filteredRVGEs, int resourceType, string atomPresetResourceType)
        {
            HashSet<JSSCDisplayChoice> jSSCDisplayChoices = new HashSet<JSSCDisplayChoice>();
            HashSet<JSSCDisplayChoice> jSSCPreferredDisplayChoices = new HashSet<JSSCDisplayChoice>();

            foreach (var creatorKVP in creatorManifestDict)
            {
                var creator = creatorKVP.Value;
                int rvgeCount = creator.GetResourceCount(filteredRVGEs);
                if (rvgeCount> 0)
                {
                    var jsscDisplayChoice = new JSSCDisplayChoice();
                    bool isPreferred = creator.IsPreferredCreator(resourceType, atomPresetResourceType);
                    jsscDisplayChoice.choice = creator.creatorName;
                    if (isPreferred) jsscDisplayChoice.displayChoice = "<i>* "+creator.creatorName + " *</i>" + " (" + rvgeCount.ToString() + ")";
                    else jsscDisplayChoice.displayChoice = creator.creatorName + " (" + rvgeCount.ToString() + ")";
                    jsscDisplayChoice.count = rvgeCount;

                    if (isPreferred) jSSCPreferredDisplayChoices.Add(jsscDisplayChoice);
                    else jSSCDisplayChoices.Add(jsscDisplayChoice);
                }
            }

            var creatorFilterOrderJSE = Settings.GetCurrentCreatorFilterOrder(resourceType, atomPresetResourceType);
            List<string> creatorChoices = GetSortedCreatorChoices(jSSCPreferredDisplayChoices, creatorFilterOrderJSE) ;
            List<string> creatorDisplayChoices = GetSortedCreatorDisplayChoices(jSSCPreferredDisplayChoices, creatorFilterOrderJSE);
            creatorChoices.AddRange(GetSortedCreatorChoices(jSSCDisplayChoices, creatorFilterOrderJSE));
            creatorDisplayChoices.AddRange(GetSortedCreatorDisplayChoices(jSSCDisplayChoices, creatorFilterOrderJSE));

            creatorJSMultiEnum.SetMainChoices(creatorChoices, creatorDisplayChoices);

            if (creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
            {
                if (!creatorChoices.Contains(creatorJSMultiEnum.mainVal)) creatorJSMultiEnum.valTopEnum = CreatorFilterChoices.allVARs;
            }
        }

        private static List<string> GetSortedCreatorChoices(HashSet<JSSCDisplayChoice> jSSCDisplayChoices, JSONStorableEnumStringChooser creatorFilterOrderJSE)
        {
            if (creatorFilterOrderJSE.val == CreatorFilterSortOrder.alphabetically) return jSSCDisplayChoices.OrderBy(o => o.choice).Select(o => o.choice).ToList();
            else return jSSCDisplayChoices.OrderByDescending(o => o.count).ThenBy(o => o.choice).Select(o => o.choice).ToList();
        }

        private static List<string> GetSortedCreatorDisplayChoices(HashSet<JSSCDisplayChoice> jSSCDisplayChoices, JSONStorableEnumStringChooser creatorFilterOrderJSE)
        {
            if (creatorFilterOrderJSE.val == CreatorFilterSortOrder.alphabetically) return jSSCDisplayChoices.OrderBy(o => o.choice).Select(o => o.displayChoice).ToList();
            else return jSSCDisplayChoices.OrderByDescending(o => o.count).ThenBy(o => o.choice).Select(o => o.displayChoice).ToList();
        }

        public static void UpdateClothingCreatorChoices(JSONStorableMultiEnumStringChooser clothingCreatorJSME, HashSet<ResourceVersionGroupEntry> filteredRVGEs, int resourceType, string atomPresetResourceType)
        {
            HashSet<JSSCDisplayChoice> jSSCDisplayChoices = new HashSet<JSSCDisplayChoice>();

            foreach (var creatorKVP in clothingCreatorManifestDict)
            {
                var creator = creatorKVP.Value;
                int rvgeCount = creator.GetResourceCount(filteredRVGEs);
                if (rvgeCount > 0)
                {
                    var jsscDisplayChoice = new JSSCDisplayChoice();
                    jsscDisplayChoice.choice = creator.creatorName;
                    jsscDisplayChoice.displayChoice = creator.creatorName + " (" + rvgeCount.ToString() + ")";
                    jsscDisplayChoice.count = rvgeCount;
                    jSSCDisplayChoices.Add(jsscDisplayChoice);
                }
            }

            List<string> creatorChoices;
            List<string> creatorDisplayChoices;
            var creatorFilterOrderJSE = Settings.GetCurrentCreatorFilterOrder(resourceType, atomPresetResourceType);

            if (creatorFilterOrderJSE.val == CreatorFilterSortOrder.alphabetically)
            {
                creatorChoices = jSSCDisplayChoices.OrderBy(o => o.choice).Select(o => o.choice).ToList();
                creatorDisplayChoices = jSSCDisplayChoices.OrderBy(o => o.choice).Select(o => o.displayChoice).ToList();

            }
            else
            {
                creatorChoices = jSSCDisplayChoices.OrderByDescending(o => o.count).ThenBy(o => o.choice).Select(o => o.choice).ToList();
                creatorDisplayChoices = jSSCDisplayChoices.OrderByDescending(o => o.count).ThenBy(o => o.choice).Select(o => o.displayChoice).ToList();
            }

            clothingCreatorJSME.SetMainChoices(creatorChoices, creatorDisplayChoices);

            if (clothingCreatorJSME.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
            {
                if (!creatorChoices.Contains(clothingCreatorJSME.mainVal)) clothingCreatorJSME.valTopEnum = ClothingCreatorFilterChoices.all;
            }

        }

    }

    public class JSSCDisplayChoice
    {
        public string choice { get; set; }
        public string displayChoice { get; set; }
        public int count { get; set; } = 0;
    }
    public class CreatorManifestEntry
    {
        public string creatorName { get; private set; }

        private HashSet<int> resourceTypes { get; set; } = new HashSet<int>();
        private HashSet<string> atomPresetResourceTypes { get; set; } = new HashSet<string>();

        private HashSet<int> preferredByResourceTypes { get; set; } = new HashSet<int>();

        private HashSet<string> preferredByAtomPresetResourceTypes { get; set; } = new HashSet<string>();

        private HashSet<int> hiddenByResourceTypes { get; set; } = new HashSet<int>();

        private HashSet<string> hiddenByAtomPresetResourceTypes { get; set; } = new HashSet<string>();

        private Dictionary<int, HashSet<ResourceVersionGroupEntry>> resourcesByTypeDict = new Dictionary<int, HashSet<ResourceVersionGroupEntry>>();

        private Dictionary<string, HashSet<ResourceVersionGroupEntry>> resourcesByAtomPresetTypeDict = new Dictionary<string, HashSet<ResourceVersionGroupEntry>>();
        public CreatorManifestEntry(string creatorName)
        {
            this.creatorName = creatorName;
        }

        public JSONArray GetPreferredResourceTypes()
        {
            JSONArray ja = null;

            if (preferredByResourceTypes.Count > 0)
            {
                ja = new JSONArray();

                foreach (int resourceType in preferredByResourceTypes) ja[-1].AsInt = resourceType;
            }

            return ja;
        }

        public JSONArray GetPreferredAtomPresetResourceTypes()
        {
            JSONArray ja = null;

            if (preferredByAtomPresetResourceTypes.Count > 0)
            {
                ja = new JSONArray();

                foreach (string atomPresetResourceType in preferredByAtomPresetResourceTypes) ja[-1] = atomPresetResourceType;
            }

            return ja;
        }


        public void RestorePreferredResourceTypes(JSONArray preferredResourcesJA)
        {
            preferredByResourceTypes.Clear();

            foreach (JSONNode jn in preferredResourcesJA) preferredByResourceTypes.Add(jn.AsInt);
        }

        public void RestorePreferredAtomPresetResourceTypes(JSONArray preferredAtomPresetResourcesJA)
        {
            preferredByAtomPresetResourceTypes.Clear();

            foreach (JSONNode jn in preferredAtomPresetResourcesJA) preferredByAtomPresetResourceTypes.Add(jn);
        }


        public void RestoreHiddenResourceTypes(JSONArray preferredResourcesJA)
        {
            hiddenByResourceTypes.Clear();

            foreach (JSONNode jn in preferredResourcesJA) hiddenByResourceTypes.Add(jn.AsInt);
        }

        public void RestoreHiddenAtomPresetResourceTypes(JSONArray preferredAtomPresetResourcesJA)
        {
            hiddenByAtomPresetResourceTypes.Clear();

            foreach (JSONNode jn in preferredAtomPresetResourcesJA) hiddenByAtomPresetResourceTypes.Add(jn);
        }

        public JSONArray GetHiddenResourceTypes()
        {
            JSONArray ja = null;

            if (hiddenByResourceTypes.Count > 0)
            {
                ja = new JSONArray();

                foreach (int resourceType in hiddenByResourceTypes) ja[-1].AsInt = resourceType;
            }

            return ja;
        }

        public JSONArray GetHiddenAtomPresetResourceTypes()
        {
            JSONArray ja = null;

            if (hiddenByAtomPresetResourceTypes.Count > 0)
            {
                ja = new JSONArray();

                foreach (string atomPresetResourceType in hiddenByAtomPresetResourceTypes) ja[-1] = atomPresetResourceType;
            }

            return ja;
        }


        public bool IsHiddenResources (int resourceType, string atomPresetResourceType)
        {
            if (resourceType == ResourceType.presetAtom) return hiddenByAtomPresetResourceTypes.Contains(atomPresetResourceType);
            return hiddenByResourceTypes.Contains(resourceType);
        }

        public void SetHiddenResources(int resourceType, string atomPresetResourceType, bool hidden)
        {
            if (resourceType == ResourceType.presetAtom)
            {
                if (hidden) hiddenByAtomPresetResourceTypes.Add(atomPresetResourceType);
                else hiddenByAtomPresetResourceTypes.Remove(atomPresetResourceType);
            }
            else
            {
                if (hidden) hiddenByResourceTypes.Add(resourceType);
                else hiddenByResourceTypes.Remove(resourceType);
            }


            Settings.SaveSettings();

            var mainBrowserFilter = BrowserAssist.mainBrowserUI.browserFiltersUI.currentFilterSettings;

            if (mainBrowserFilter.resourceType== resourceType && (resourceType!=ResourceType.presetAtom || mainBrowserFilter.atomPresetResourceType == atomPresetResourceType))  BrowserAssist.mainBrowserUI.RefreshFilterResults();
            if (BrowserAssist.callbackBrowserUI.browserFiltersUI.currentFilterSettings != null)
            {
                var callbackBrowserFilter = BrowserAssist.callbackBrowserUI.browserFiltersUI.currentFilterSettings;
                if (callbackBrowserFilter.resourceType == resourceType && (resourceType != ResourceType.presetAtom || callbackBrowserFilter.atomPresetResourceType == atomPresetResourceType)) BrowserAssist.callbackBrowserUI.RefreshFilterResults();
            }
        }

        public void SetPreferredCreator(int resourceType, string atomPresetResourceType, bool preferred)
        {
            if (resourceType == ResourceType.presetAtom)
            {
                if (preferred) preferredByAtomPresetResourceTypes.Add(atomPresetResourceType);
                else preferredByAtomPresetResourceTypes.Remove(atomPresetResourceType);
            }
            else
            {
                if (preferred) preferredByResourceTypes.Add(resourceType);
                else preferredByResourceTypes.Remove(resourceType);
            }
            Settings.SaveSettings();
        }

        public bool IsPreferredCreator(int resourceType, string atomPresetResourceType)
        {
            if (resourceType == ResourceType.presetAtom) return preferredByAtomPresetResourceTypes.Contains(atomPresetResourceType);
            return preferredByResourceTypes.Contains(resourceType);
        }

        public bool HasResourceType(int resourceType, string atomPresetResourceType)
        {
            bool hasResourceType = false;
            if (resourceType != ResourceType.presetAtom && resourceTypes.Contains(resourceType)) hasResourceType = true;

            if (resourceType == ResourceType.presetAtom && resourceTypes.Contains(resourceType) && atomPresetResourceTypes.Contains(atomPresetResourceType)) hasResourceType = true;

            return hasResourceType;
        }

        public int GetResourceCount(HashSet<ResourceVersionGroupEntry> rvges) 
        {
            if (rvges.Count() == 0 || !resourcesByTypeDict.ContainsKey(rvges.First().resourceType)) return 0;

            return rvges.Intersect(resourcesByTypeDict[rvges.First().resourceType]).Count();
        }

        public void RegisterResource (ResourceVersionGroupEntry rvge)
        {
            resourceTypes.Add(rvge.resourceType);

            if (!resourcesByTypeDict.ContainsKey(rvge.resourceType)) resourcesByTypeDict.Add(rvge.resourceType, new HashSet<ResourceVersionGroupEntry>());
            resourcesByTypeDict[rvge.resourceType].Add(rvge);

            if (rvge.resourceType == ResourceType.presetAtom)
            {
                atomPresetResourceTypes.Add(rvge.atomPresetResourceType);

                if (!resourcesByAtomPresetTypeDict.ContainsKey(rvge.atomPresetResourceType)) resourcesByAtomPresetTypeDict.Add(rvge.atomPresetResourceType, new HashSet<ResourceVersionGroupEntry>());
                resourcesByAtomPresetTypeDict[rvge.atomPresetResourceType].Add(rvge);
            }
        }

        public void RemoveRVGE(ResourceVersionGroupEntry rvge)
        {
            resourcesByTypeDict[rvge.resourceType].Remove(rvge);
            if (resourcesByTypeDict[rvge.resourceType].Count ==0) resourceTypes.Remove(rvge.resourceType);

            if (rvge.resourceType == ResourceType.presetAtom)
            {
                resourcesByAtomPresetTypeDict[rvge.atomPresetResourceType].Remove(rvge);
                if (resourcesByAtomPresetTypeDict[rvge.atomPresetResourceType].Count == 0) resourcesByAtomPresetTypeDict.Remove(rvge.atomPresetResourceType);
            }
        }
    }
}
;
